/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QDir>
#include <QTimer>
#include <QtDebug>
#include <QSqlQuery>
#include <QSqlDatabase>
#include <QProgressDialog>

#include "dmisc.h"
#include "dooble.h"
#include "dcookies.h"

dcookies::dcookies(void):QNetworkCookieJar()
{
  /*
  ** settingsWindow/cookiesShouldBe
  ** 0 - deleted upon exit
  ** 1 - kept forever (that's a long time)
  ** 2 - kept until they expire
  */

  m_timer = new QTimer(this);
  connect(m_timer,
	  SIGNAL(timeout(void)),
	  this,
	  SLOT(slotRemoveDomains(void)));
  connect(this,
	  SIGNAL(cookieReceived(const QString &,
				const QUrl &,
				const QDateTime &)),
	  dooble::s_cookiesBlockWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));

  int interval = computedTimerInterval();

  if(interval > 0)
    m_timer->start(interval);

  createCookieDatabase();
}

dcookies::~dcookies()
{
  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    deleteExpiredCookies();
  else
    purge();
}

QList<QNetworkCookie> dcookies::allCookies(void) const
{
  /*
  ** If cookies are disabled, should this method return an empty
  ** list or the current list? For now, it will return the
  ** current list of cookies.
  */

  return QNetworkCookieJar::allCookies();
}

QList<QNetworkCookie> dcookies::cookiesForUrl(const QUrl &url) const
{
  /*
  ** If cookies are disabled, should this method return an empty
  ** list or the current list? For now, it will return the
  ** current list of cookies minus the expired cookies.
  */

  QList<QNetworkCookie> list(QNetworkCookieJar::cookiesForUrl(url));

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      QDateTime now(QDateTime::currentDateTime());

      /*
      ** Remove expired cookies.
      */

      for(int i = list.size() - 1; i >= 0; i--)
	if(list.at(i).isSessionCookie())
	  continue;
	else if(list.at(i).expirationDate().toLocalTime() <= now)
	  list.removeAt(i);
    }

  return list;
}

bool dcookies::setCookiesFromUrl(const QList<QNetworkCookie> &cookieList,
				 const QUrl &url)
{
  if(!dooble::s_settings.value("settingsWindow/cookiesEnabled",
			       true).toBool())
    return false;
  else if(cookieList.isEmpty())
    return false;

  QDateTime now(QDateTime::currentDateTime());
  QList<QNetworkCookie> list(cookieList);

  for(int i = list.size() - 1; i >= 0; i--)
    {
      QString domain(list.at(i).domain());

      if(domain == "." || domain.isEmpty())
	domain = url.host();

      /*
      ** Other exceptions models are only populated if some
      ** action has been denied. With cookies, knowledge of the
      ** cookie may be beneficial.
      */

      emit cookieReceived(domain, url, now);

      if(dooble::s_cookiesBlockWindow->allowed(domain))
	/*
	** Exempt domains must be blocked.
	*/

	list.removeAt(i);
    }

  if(!QNetworkCookieJar::setCookiesFromUrl(list, url))
    {
      QList<QNetworkCookie> all(allCookies());

      all.append(list);
      setAllCookies(all);
      all.clear();

      /*
      ** These cookies may have originated from a
      ** different url.
      */

      for(int i = 0; i < list.size(); i++)
	saveCookie(list.at(i));

      emit cookiesChanged(list);
    }

  /*
  ** Save the cookies for the current url.
  ** Session cookies will not be saved.
  */

  list = cookiesForUrl(url);

  for(int i = 0; i < list.size(); i++)
    saveCookie(list.at(i));

  emit cookiesChanged(list);

  /*
  ** We're going to remove all of the expired cookies.
  */

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      list = allCookies();

      QDateTime now(QDateTime::currentDateTime());

      /*
      ** Remove expired cookies.
      */

      for(int i = list.size() - 1; i >= 0; i--)
	if(list.at(i).isSessionCookie())
	  continue;
	else if(list.at(i).expirationDate().toLocalTime() <= now)
	  list.removeAt(i);

      setAllCookies(list);
    }

  /*
  ** Now, submit a request to the Cookies window to remove any
  ** entries that have been removed for url.
  */

  emit cookiesChangedForUrl(allCookies(), url);
  return true;
}

void dcookies::removeCookie(const QString &domain)
{
  QList<QNetworkCookie> all(allCookies());

  for(int i = all.size() - 1; i >= 0; i--)
    if(all.at(i).domain() == domain)
      /*
      ** Remove all cookies from the specified domain.
      */

      all.removeAt(i);

  setAllCookies(all);
  all.clear();
  m_favorites.remove(domain);

  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.prepare("DELETE FROM cookies WHERE HEX(domain) = HEX(?)");
	    query.bindValue
	      (0, dmisc::encodedString(domain.toAscii(),
				       true).toBase64().data());
	    query.exec();
	    query.prepare("DELETE FROM favorites WHERE HEX(domain) = HEX(?)");
	    query.bindValue
	      (0, dmisc::encodedString(domain.toAscii(),
				       true).toBase64().data());
	    query.exec();
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::removeCookie(const QNetworkCookie &cookie)
{
  removeCookie(QVariant(cookie.name()).toString(),
	       cookie.domain(),
	       cookie.path());
}

void dcookies::removeCookie(const QString &name, const QString &domain,
			    const QString &path)
{
  QList<QNetworkCookie> all(allCookies());

  for(int i = all.size() - 1; i >= 0; i--)
    if(all.at(i).domain() == domain &&
       QVariant(all.at(i).name()).toString() == name &&
       all.at(i).path() == path)
      {
	/*
	** Remove exactly one cookie.
	*/

	all.removeAt(i);
	break;
      }

  setAllCookies(all);
  all.clear();

  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.prepare
	      ("DELETE FROM cookies WHERE "
	       "HEX(domain) = HEX(?) AND "
	       "HEX(name) = HEX(?) AND "
	       "HEX(path) = HEX(?)");
	    query.bindValue
	      (0, dmisc::encodedString(domain.toAscii(),
				       true).toBase64().data());
	    query.bindValue
	      (1, dmisc::encodedString(name.toAscii(),
				       true).toBase64().data());
	    query.bindValue
	      (2, dmisc::encodedString(path.toAscii(),
				       true).toBase64().data());
	    query.exec();
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::createCookieDatabase(void)
{
  if(!dooble::s_settings.value("settingsWindow/cookiesEnabled",
			       true).toBool())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QSqlQuery query(db);

	query.exec("CREATE TABLE IF NOT EXISTS cookies ("
		   "domain TEXT NOT NULL, "
		   "name BLOB NOT NULL, "
		   "path TEXT NOT NULL, "
		   "expiration_date TEXT NOT NULL, "
		   "http TEXT NOT NULL, "
		   "secure TEXT NOT NULL, "
		   "value BLOB, "
		   "passphrase_hash TEXT NOT NULL, "
		   "PRIMARY KEY (domain, name, path))");
	query.exec("CREATE TABLE IF NOT EXISTS favorites ("
		   "domain TEXT PRIMARY KEY NOT NULL, "
		   "passphrase_hash TEXT NOT NULL)");
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

void dcookies::allowDomain(const QString &domain,
			   const bool allowed)
{
  if(allowed)
    m_favorites[domain] = true;
  else
    m_favorites.remove(domain);

  if(dmisc::passphraseWasAuthenticated())
    {
      createCookieDatabase();

      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QSqlQuery query(db);

	    if(allowed)
	      {
		query.exec("PRAGMA synchronous = OFF");
		query.prepare("INSERT OR REPLACE INTO favorites "
			      "(domain, passphrase_hash)"
			      "VALUES (?, ?)");
		query.bindValue
		  (0, dmisc::encodedString(domain.toAscii(),
					   true).toBase64());

		QString hashType(dooble::s_settings.value
				 ("settingsWindow/hashType").
				 toString());

		query.bindValue(1, dmisc::passphraseHash(hashType));
		query.exec();
	      }
	    else
	      {
		query.exec("PRAGMA synchronous = OFF");
		query.prepare("DELETE FROM favorites WHERE HEX(domain) = "
			      "HEX(?)");
		query.bindValue
		  (0, dmisc::encodedString(domain.toAscii(),
					   true).toBase64().data());
		query.exec();
	      }
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}

void dcookies::populate(void)
{
  m_favorites.clear();
  deleteExpiredCookies();
  setAllCookies(QList<QNetworkCookie> ());
  emit allCookiesRemoved();

  if(!dmisc::passphraseWasAuthenticated())
    return;

  QList<QNetworkCookie> favorites;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);
	QList<QNetworkCookie> all;

	if(query.exec(QString("SELECT domain, name, path, expiration_date, "
			      "http, secure, value "
			      "FROM cookies WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QNetworkCookie cookie;

	      cookie.setDomain
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(0).toString().toAscii())).data());

	      if(!QUrl::fromEncoded(cookie.domain().toAscii(),
				    QUrl::StrictMode).isValid())
		{
		  QSqlQuery deleteQuery(db);

		  deleteQuery.exec("PRAGMA synchronous = OFF");
		  deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					   "HEX(domain) = HEX('%1')").
				   arg(query.value(0).toString()));
		  continue;
		}

	      QByteArray bytes;

	      bytes = dmisc::decodedString
		(QByteArray::fromBase64(query.value(1).toByteArray()));
	      cookie.setName(bytes.data());
	      cookie.setPath
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(2).toString().toAscii())).data());
	      cookie.setExpirationDate
		(QDateTime::fromString
		 (dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(3).toString().toAscii())).data(),
		  Qt::ISODate));
	      cookie.setHttpOnly
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(4).toString().toAscii())).toInt());
	      cookie.setSecure
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(5).toString().toAscii())).toInt());
	      bytes = dmisc::decodedString
		(QByteArray::fromBase64(query.value(6).toByteArray()));
	      cookie.setValue(bytes.data());
	      all.append(cookie);
	    }

	setAllCookies(all);
	all.clear();

	if(query.exec(QString("SELECT domain FROM favorites WHERE "
			      "passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QNetworkCookie cookie;

	      cookie.setDomain
		(dmisc::decodedString
		 (QByteArray::fromBase64
		  (query.value(0).toString().toAscii())).data());

	      if(!QUrl::fromEncoded(cookie.domain().toAscii(),
				    QUrl::StrictMode).isValid())
		{
		  QSqlQuery deleteQuery(db);

		  deleteQuery.exec("PRAGMA synchronous = OFF");
		  deleteQuery.exec(QString("DELETE FROM favorites WHERE "
					   "HEX(domain) = HEX('%1')").
				   arg(query.value(0).toString()));
		  continue;
		}

	      m_favorites[cookie.domain()] = true;
	      favorites.append(cookie);
	    }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
  emit cookiesChanged(favorites);
  emit cookiesChanged(allCookies());
}

void dcookies::saveCookie(const QNetworkCookie &cookie)
{
  if(cookie.isSessionCookie())
    return;
  else if(!dmisc::passphraseWasAuthenticated())
    return;

  if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
			      0).toInt() == 2)
    {
      QDateTime now(QDateTime::currentDateTime());

      if(cookie.expirationDate().toLocalTime() <= now)
	return;
    }

  createCookieDatabase();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.prepare("INSERT OR REPLACE INTO cookies "
		      "(domain, name, path, "
		      "expiration_date, http, secure, "
		      "value, passphrase_hash) "
		      "VALUES (?, ?, ?, ?, ?, ?, ?, ?)");
	query.bindValue
	  (0,
	   dmisc::encodedString(cookie.domain().toAscii(), true).
	   toBase64());
	query.bindValue
	  (1,
	   dmisc::encodedString(cookie.name(), true).toBase64());
	query.bindValue
	  (2,
	   dmisc::encodedString(cookie.path().toAscii(), true).
	   toBase64());
	query.bindValue
	  (3,
	   dmisc::encodedString(cookie.expirationDate().
				toLocalTime().
				toString(Qt::ISODate).toAscii(),
				true).toBase64());
	query.bindValue
	  (4,
	   dmisc::encodedString
	   (QString::number(cookie.isHttpOnly()).toAscii(),
	    true).toBase64());
	query.bindValue
	  (5,
	   dmisc::encodedString
	   (QString::number(cookie.isSecure()).toAscii(),
	    true).toBase64());
	query.bindValue
	  (6,
	   dmisc::encodedString(cookie.value(), true).toBase64());

	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());

	query.bindValue(7, dmisc::passphraseHash(hashType));
	query.exec();
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

void dcookies::deleteExpiredCookies(void)
{
  if(!dmisc::passphraseWasAuthenticated())
    return;
  else if(dooble::s_settings.value("settingsWindow/cookiesShouldBe",
				   0).toInt() != 2)
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "cookies.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	if(query.exec(QString("SELECT expiration_date FROM cookies "
			      "WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  {
	    QDateTime now(QDateTime::currentDateTime());
	    QSqlQuery deleteQuery(db);

	    deleteQuery.exec("PRAGMA synchronous = OFF");

	    while(query.next())
	      {
		QDateTime dateTime
		  (QDateTime::fromString
		   (dmisc::decodedString
		    (QByteArray::fromBase64
		     (query.value(0).toString().toAscii())).data(),
		    Qt::ISODate));

		if(dateTime.isValid())
		  {
		    if(dateTime.toLocalTime() <= now)
		      deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					       "HEX(expiration_date) = "
					       "HEX('%1')").arg
				       (query.value(0).toString()));
		  }
		else
		  deleteQuery.exec(QString("DELETE FROM cookies WHERE "
					   "HEX(expiration_date) = "
					   "HEX('%1')").arg
				   (query.value(0).toString()));
	      }
	  }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("cookies");
}

int dcookies::computedTimerInterval(void) const
{
  int interval = 0;

  if(dooble::s_settings.value("settingsWindow/cookiesEnabled",
			      true).toBool() &&
     dooble::s_settings.value("settingsWindow/cookieTimerEnabled",
			      false).toBool())
    {
      int unit = dooble::s_settings.value
	("settingsWindow/cookieTimerUnit", 0).toInt();

      if(!(unit == 0 || unit == 1))
	unit = 0;

      interval = dooble::s_settings.value
	("settingsWindow/cookieTimerInterval", 1).toInt();

      if(unit == 0)
	{
	  if(interval < 1 || interval > 60)
	    interval = 1;

	  interval = 3600000 * interval;
	}
      else
	{
	  if(interval < 1 || interval > 3600)
	    interval = 1;

	  interval = 60000 * interval;
	}
    }

  return interval;
}

void dcookies::slotCookieTimerChanged(void)
{
  int interval = computedTimerInterval();

  if(interval > 0)
    {
      if(interval != m_timer->interval())
	m_timer->start(interval);
    }
  else
    m_timer->stop();
}

bool dcookies::isFavorite(const QString &domain) const
{
  if(m_favorites.contains(domain))
    return m_favorites.value(domain);
  else
    return false;
}

void dcookies::slotRemoveDomains(void)
{
  QStringList domains;
  QList<QNetworkCookie> all(allCookies());

  /*
  ** Locate the domains that are not in the
  ** m_favorites container.
  */

  for(int i = 0; i < all.size(); i++)
    if(m_favorites.contains(all.at(i).domain()))
      {
	if(!m_favorites.value(all.at(i).domain()))
	  if(!domains.contains(all.at(i).domain()))
	    domains.append(all.at(i).domain());
      }
    else if(!domains.contains(all.at(i).domain()))
      domains.append(all.at(i).domain());

  all.clear();

  /*
  ** Remove all cookies that are located in the
  ** domains container.
  */

  for(int i = 0; i < domains.size(); i++)
    removeCookie(domains.at(i));

  emit domainsRemoved(domains);
}

int dcookies::count(void) const
{
  return allCookies().size();
}

void dcookies::reencode(QProgressDialog *progress)
{
  /*
  ** This method assumes that the respective container
  ** is as current as possible. It will use the container's
  ** contents to create new database.
  ** Obsolete data will be deleted via the purge method.
  */

  if(progress)
    {
      progress->reset();
      progress->setWindowTitle(tr("Dooble: Progress Dialog"));
      progress->setLabelText(tr("Re-encoding cookie information..."));
      progress->show();
    }

  QHash<QString, bool> favorites(m_favorites);
  QList<QNetworkCookie> all(allCookies());

  if(progress)
    {
      progress->setMaximum(all.size());
      progress->update();
    }

  for(int i = 0; i < all.size(); i++)
    {
      saveCookie(all.at(i));

      if(progress)
	progress->setValue(i + 1);
    }

  all.clear();
  m_favorites.clear();

  if(progress)
    {
      progress->setMaximum(favorites.size());
      progress->update();
    }

  for(int i = 0; i < favorites.keys().size(); i++)
    {
      allowDomain(favorites.keys().at(i),
		  favorites[favorites.keys().at(i)]);

      if(progress)
	progress->setValue(i + 1);
    }

  favorites.clear();
  purge();

  if(progress)
    progress->setVisible(false);
}

void dcookies::purge(void)
{
  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "cookies");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "cookies.db");

	if(db.open())
	  {
	    QString hashType(dooble::s_settings.value
			     ("settingsWindow/hashType").
			     toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.exec
	      (QString("DELETE FROM cookies WHERE "
		       "passphrase_hash <> '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	    query.exec
	      (QString("DELETE FROM favorites WHERE "
		       "passphrase_hash <> '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("cookies");
    }
}
