/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QDir>
#include <QDateTime>
#include <QKeyEvent>
#include <QSettings>
#include <QSqlIndex>
#include <QSqlQuery>
#include <QSqlDatabase>
#include <QProgressDialog>

#include "dmisc.h"
#include "dooble.h"
#include "ddownloadwindow.h"

ddownloadwindow::ddownloadwindow(void):QMainWindow()
{
  ui.setupUi(this);
#ifdef Q_WS_MAC
  setAttribute(Qt::WA_MacMetalStyle, true);
#endif
#if QT_VERSION >= 0x040700
  ui.searchLineEdit->setPlaceholderText(tr("Search Downloads"));
#endif
  connect(ui.closePushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotClose(void)));
  connect(ui.clearListPushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotClearList(void)));
  connect(ui.enterUrlPushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotEnterUrl(void)));
  connect(ui.cancelPushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotCancelDownloadUrl(void)));
  connect(ui.downloadPushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotDownloadUrl(void)));
  connect(ui.clearPushButton, SIGNAL(clicked(void)), this,
	  SLOT(slotClear(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  ui.searchLineEdit,
	  SLOT(slotSetIcons(void)));
  connect(ui.searchLineEdit, SIGNAL(textEdited(const QString &)),
	  this, SLOT(slotTextChanged(const QString &)));
  slotSetIcons();
  createDownloadDatabase();
}

ddownloadwindow::~ddownloadwindow()
{
  saveState();
  purge();
}

void ddownloadwindow::purge(void)
{
  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "downloads");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "downloads.db");

	if(db.open())
	  {
	    QString hashType(dooble::s_settings.value
			     ("settingsWindow/hashType").
			     toString());
	    QSqlQuery query(db);

	    query.exec("PRAGMA synchronous = OFF");
	    query.exec
	      (QString("DELETE FROM downloads WHERE "
		       "passphrase_hash <> '%1'").
	       arg(dmisc::passphraseHash(hashType)));
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("downloads");
    }
}

void ddownloadwindow::populate(void)
{
  for(int i = ui.table->rowCount() - 1; i >= 0; i--)
    {
      ddownloadwindowitem *item = qobject_cast<ddownloadwindowitem *>
	(ui.table->cellWidget(i, 0));

      if(item && !item->isDownloading())
	{
	  /*
	  ** If the user authenticates after a download has been
	  ** initiated for a session-based process, they may
	  ** wish to continue the download.
	  */

	  item->deleteLater();
	  ui.table->removeRow(i);
	}
    }

  if(!dmisc::passphraseWasAuthenticated())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "downloads");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "downloads.db");

    if(db.open())
      {
	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());
	QSqlQuery query(db);

	if(query.exec(QString("SELECT filename, url, "
			      "sortby FROM downloads "
			      "WHERE passphrase_hash = '%1'").
		      arg(dmisc::passphraseHash(hashType))))
	  while(query.next())
	    {
	      QUrl url;
	      QString sortby("");
	      QString fileName("");

	      url = QUrl::fromEncoded
		(dmisc::decodedString
		 (QByteArray::fromBase64(query.value(1).toString().
					 toAscii())).data(),
		 QUrl::StrictMode);
	      sortby = dmisc::decodedString
		(QByteArray::fromBase64(query.value(2).toString().
					toAscii())).data();
	      fileName = dmisc::decodedString
		(QByteArray::fromBase64(query.value(0).toString().
					toAscii())).data();

	      /*
	      ** These if-statements should take care of
	      ** decoding mishaps.
	      */

	      if(QFileInfo(url.toLocalFile()).exists())
		addFileItem(url.toLocalFile(), fileName, false);
	      else if(url.isValid() &&
		      (url.scheme().toLower() == "file" ||
		       url.scheme().toLower() == "ftp" ||
		       url.scheme().toLower() == "http" ||
		       url.scheme().toLower() == "https" ||
		       url.scheme().toLower() == "qrc"))
		addUrlItem(url, fileName, false);
	    }
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("downloads");
}

void ddownloadwindow::addFileItem(const QString &srcFileName,
				  const QString &dstFileName,
				  const bool isNew)
{
  QString l_dstFileName;

  if(isNew)
    l_dstFileName = dmisc::findUniqueFileName(dstFileName);
  else
    l_dstFileName = dstFileName;

  ui.stackedWidget->setCurrentIndex(0);

  ddownloadwindowitem *item = new ddownloadwindowitem(ui.table);

  connect(this,
	  SIGNAL(iconsChanged(void)),
	  item,
	  SLOT(slotSetIcons(void)));
  connect(item,
	  SIGNAL(recordDownload(const QString &,
				const QUrl &)),
	  this,
	  SLOT(slotRecordDownload(const QString &,
				  const QUrl &)));
  connect(item,
	  SIGNAL(downloadFinished(void)),
	  this,
	  SLOT(slotDownloadFinished(void)));
  item->downloadFile(srcFileName, l_dstFileName, isNew);
  ui.table->insertRow(0);
  ui.table->setCellWidget(0, 0, item);
  ui.table->resizeRowToContents(0);
}

void ddownloadwindow::addUrlItem(const QUrl &url, const QString &dstFileName,
				 const bool isNew)
{
  QString l_dstFileName;

  if(isNew)
    l_dstFileName = dmisc::findUniqueFileName(dstFileName);
  else
    l_dstFileName = dstFileName;

  ui.stackedWidget->setCurrentIndex(0);

  ddownloadwindowitem *item = new ddownloadwindowitem(ui.table);

  connect(this,
	  SIGNAL(iconsChanged(void)),
	  item,
	  SLOT(slotSetIcons(void)));
  connect(item,
	  SIGNAL(recordDownload(const QString &,
				const QUrl &)),
	  this,
	  SLOT(slotRecordDownload(const QString &,
				  const QUrl &)));
  connect(item,
	  SIGNAL(downloadFinished(void)),
	  this,
	  SLOT(slotDownloadFinished(void)));
  item->downloadUrl(url, l_dstFileName, isNew);
  ui.table->insertRow(0);
  ui.table->setCellWidget(0, 0, item);
  ui.table->resizeRowToContents(0);
}

void ddownloadwindow::addHtmlItem(const QString &html,
				  const QString &dstFileName)
{
  QString l_dstFileName(dmisc::findUniqueFileName(dstFileName));

  ui.stackedWidget->setCurrentIndex(0);

  ddownloadwindowitem *item = new ddownloadwindowitem(ui.table);

  connect(this,
	  SIGNAL(iconsChanged(void)),
	  item,
	  SLOT(slotSetIcons(void)));
  connect(item,
	  SIGNAL(recordDownload(const QString &,
				const QUrl &)),
	  this,
	  SLOT(slotRecordDownload(const QString &,
				  const QUrl &)));
  item->downloadHtml(html, l_dstFileName);
  ui.table->insertRow(0);
  ui.table->setCellWidget(0, 0, item);
  ui.table->resizeRowToContents(0);
}

void ddownloadwindow::slotClose(void)
{
  close();
}

void ddownloadwindow::closeEvent(QCloseEvent *event)
{
  saveState();
  QMainWindow::closeEvent(event);
}

void ddownloadwindow::show(void)
{
  if(!isVisible())
   {
     /*
     ** Don't annoy the user.
     */

     if(dooble::s_settings.contains("downloadWindow/geometry"))
       setGeometry(dooble::s_settings.value("downloadWindow/geometry",
					    "").toRect());
   }

  showNormal();
  raise();

  if(ui.stackedWidget->currentIndex() == 1)
    ui.urlLineEdit->setFocus();
  else
    ui.closePushButton->setFocus();
}

bool ddownloadwindow::isActive(void) const
{
  for(int i = ui.table->rowCount() - 1; i >= 0; i--)
    {
      ddownloadwindowitem *item =
	qobject_cast<ddownloadwindowitem *> (ui.table->cellWidget(i, 0));

      if(item && item->isDownloading())
	return true;
    }

  return false;
}

void ddownloadwindow::slotClearList(void)
{
  if(dmisc::passphraseWasAuthenticated())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "downloads");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "downloads.db");

	if(db.open())
	  {
	    bool state = false;
	    QSqlQuery query(db);
	    ddownloadwindowitem *item = 0;

	    query.exec("PRAGMA synchronous = OFF");

	    for(int i = ui.table->rowCount() - 1; i >= 0; i--)
	      if(!ui.table->isRowHidden(i))
		{
		  state = false;
		  item = qobject_cast<ddownloadwindowitem *>
		    (ui.table->cellWidget(i, 0));

		  if(item && !item->isDownloading())
		    state = query.exec
		      (QString("DELETE FROM downloads WHERE "
			       "HEX(filename) = HEX('%1')").
		       arg(dmisc::encodedString
			   (item->primaryKey().
			    toAscii(), true).toBase64().data()));

		  if(state)
		    {
		      item->deleteLater();
		      ui.table->removeRow(i);
		    }
		}
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("downloads");
    }
  else
    {
      for(int i = ui.table->rowCount() - 1; i >= 0; i--)
	if(!ui.table->isRowHidden(i))
	  {
	    ddownloadwindowitem *item = qobject_cast<ddownloadwindowitem *>
	      (ui.table->cellWidget(i, 0));

	    if(item && !item->isDownloading())
	      {
		item->deleteLater();
		ui.table->removeRow(i);
	      }
	  }
    }
}

void ddownloadwindow::slotEnterUrl(void)
{
  ui.stackedWidget->setCurrentIndex(1);
  ui.urlLineEdit->setFocus();
}

void ddownloadwindow::slotCancelDownloadUrl(void)
{
  ui.stackedWidget->setCurrentIndex(0);
}

void ddownloadwindow::slotDownloadUrl(void)
{
  QUrl url(QUrl::fromUserInput(ui.urlLineEdit->text().trimmed()));

  if(url.isValid())
    emit saveUrl(url);
}

void ddownloadwindow::slotSetIcons(void)
{
  QSettings settings
    (dooble::s_settings.value("iconSet").toString(), QSettings::IniFormat);

  setWindowIcon
    (QIcon(settings.value("downloadWindow/windowIcon").toString()));
  ui.closePushButton->setIcon
    (QIcon(settings.value("cancelButtonIcon").toString()));
  ui.cancelPushButton->setIcon
    (QIcon(settings.value("cancelButtonIcon").toString()));
  ui.clearPushButton->setIcon
    (QIcon(settings.value("downloadWindow/clearListButtonIcon").toString()));
  ui.clearListPushButton->setIcon
    (QIcon(settings.value("downloadWindow/clearListButtonIcon").toString()));
  ui.downloadPushButton->setIcon
    (QIcon(settings.value("downloadWindow/downloadButtonIcon").toString()));
  ui.enterUrlPushButton->setIcon
    (QIcon(settings.value("downloadWindow/enterUrlButtonIcon").
	   toString()));
  emit iconsChanged();
}

void ddownloadwindow::slotClear(void)
{
  ui.urlLineEdit->clear();
  ui.urlLineEdit->setFocus();
}

void ddownloadwindow::createDownloadDatabase(void)
{
  if(!dooble::s_settings.value("settingsWindow/rememberDownloads",
			       true).toBool())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "downloads");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "downloads.db");

    if(db.open())
      {
	QSqlQuery query(db);

#if DOOBLE_VERSION == 0x012500
	QSqlRecord record(db.record("downloads"));

	if(record.contains("path"))
	  query.exec("DROP TABLE downloads");
#endif

	query.exec("CREATE TABLE IF NOT EXISTS downloads ("
		   "filename TEXT PRIMARY KEY NOT NULL, "
		   "url TEXT NOT NULL, "
		   "sortby TEXT NOT NULL, "
		   "passphrase_hash TEXT NOT NULL)");
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("downloads");
}

void ddownloadwindow::slotRecordDownload(const QString &fileName,
					 const QUrl &url)
{
  if(!dmisc::passphraseWasAuthenticated())
    return;

  static quint64 artificialTime = 0;

  if(!dooble::s_settings.value("settingsWindow/rememberDownloads",
			       true).toBool())
    return;

  createDownloadDatabase();

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "downloads");

    db.setDatabaseName(dooble::s_homePath +
		       QDir::separator() + "downloads.db");

    if(db.open())
      {
	QSqlQuery query(db);

	artificialTime += 1; /*
			     ** slotRecordDownload() is called by
			     ** reencode(). Time is used for sorting
			     ** the download information.
			     */
	query.exec("PRAGMA synchronous = OFF");
	query.prepare
	  ("INSERT OR REPLACE INTO downloads "
	   "(filename, url, sortby, "
	   "passphrase_hash) "
	   "VALUES (?, ?, ?, ?)");
	query.bindValue
	  (0,
	   dmisc::encodedString(fileName.toAscii(),
				true).toBase64());
	query.bindValue
	  (1,
	   dmisc::encodedString
	   (url.toEncoded(QUrl::StripTrailingSlash), true).toBase64());
#if QT_VERSION >= 0x040700
	query.bindValue
	  (2,
	   dmisc::encodedString
	   (QString::number(QDateTime::currentMSecsSinceEpoch() +
			    artificialTime).
	    toAscii(), true).toBase64());
#else
	query.bindValue
	  (2,
	   dmisc::encodedString
	   (QString::number(QDateTime::currentDateTime().
			    toTime_t() + artificialTime).toAscii(),
	    true).toBase64());
#endif

	QString hashType(dooble::s_settings.value("settingsWindow/hashType").
			 toString());

	query.bindValue(3, dmisc::passphraseHash(hashType));
	query.exec();
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("downloads");
}

void ddownloadwindow::saveState(void)
{
  /*
  ** geometry() may return (0, 0) coordinates if the window is
  ** not visible.
  */

  if(!isVisible())
    return;

  dooble::s_settings["downloadWindow/geometry"] = geometry();

  QSettings settings;

  settings.setValue("downloadWindow/geometry", geometry());
}

void ddownloadwindow::slotDownloadFinished(void)
{
  if(!dooble::s_settings.value("settingsWindow/closeDownloads",
			       false).toBool())
    return;

  ddownloadwindowitem *item = qobject_cast<ddownloadwindowitem *> (sender());

  if(item && item->abortedByUser())
    /*
    ** If the item that emitted the downloadFinished() signal did so due to
    ** user interaction, the Downloads window should not be closed.
    ** Note, if the user aborts a download and another download completes
    ** within some minute amount of time, the Downloads window will be closed.
    */

    return;

  foreach(ddownloadwindowitem *item, findChildren<ddownloadwindowitem *> ())
    if(item->isDownloading())
      return;

  close();
}

void ddownloadwindow::keyPressEvent(QKeyEvent *event)
{
  if(event && event->key() == Qt::Key_Escape)
    close();
  else if(event && event->key() == Qt::Key_Delete &&
	  ui.stackedWidget->currentIndex() == 0 &&
	  ui.table->currentRow() > -1)
    {
      ddownloadwindowitem *item = qobject_cast<ddownloadwindowitem *>
	(ui.table->cellWidget(ui.table->currentRow(), 0));

      if(item && !item->isDownloading())
	{
	  ui.table->removeRow(ui.table->currentRow());

	  if(dmisc::passphraseWasAuthenticated())
	    {
	      {
		QSqlDatabase db = QSqlDatabase::addDatabase
		  ("QSQLITE", "downloads");

		db.setDatabaseName(dooble::s_homePath +
				   QDir::separator() + "downloads.db");

		if(db.open())
		  {
		    QSqlQuery query(db);

		    query.exec("PRAGMA synchronous = OFF");
		    query.exec
		      (QString("DELETE FROM downloads WHERE "
			       "HEX(filename) = HEX('%1')").
		       arg(dmisc::encodedString
			   (item->primaryKey().
			    toAscii(), true).toBase64().data()));
		  }

		db.close();
	      }

	      QSqlDatabase::removeDatabase("downloads");
	    }

	  item->deleteLater();
	}
    }

  QMainWindow::keyPressEvent(event);
}

void ddownloadwindow::slotTextChanged(const QString &text)
{
  for(int i = 0; i < ui.table->rowCount(); i++)
    {
      ddownloadwindowitem *item =
	qobject_cast<ddownloadwindowitem *> (ui.table->cellWidget(i, 0));

      if(item)
	{
	  if(item->text().toLower().contains(text.trimmed().toLower()))
	    ui.table->showRow(i);
	  else
	    ui.table->hideRow(i);
	}
      else
	ui.table->hideRow(i);
    }
}

void ddownloadwindow::reencode(QProgressDialog *progress)
{
  /*
  ** This method assumes that the respective container
  ** is as current as possible. It will use the container's
  ** contents to create new database.
  ** Obsolete data will be deleted via the purge method.
  */

  if(progress)
    {
      progress->reset();
      progress->setWindowTitle(tr("Dooble: Progress Dialog"));
      progress->setLabelText(tr("Re-encoding download information..."));
      progress->setMaximum(ui.table->rowCount());
      progress->show();
      progress->update();
    }

  for(int i = ui.table->rowCount() - 1, j = 1; i >= 0; i--, j++)
    {
      ddownloadwindowitem *item =
	qobject_cast<ddownloadwindowitem *> (ui.table->cellWidget(i, 0));

      if(item)
	slotRecordDownload(item->fileName(),
			   item->url());

      if(progress)
	progress->setValue(j);
    }

  purge();

  if(progress)
    progress->setVisible(false);
}
