/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <QDir>
#include <QBuffer>
#include <QSettings>
#include <QSqlQuery>
#include <QSqlDatabase>
#include <QtCore/qmath.h>
#include <QFileIconProvider>

#include "dmisc.h"
#include "dooble.h"

/*
** OK, so you know. One of the below proxy methods (a static one)
** returns a list of proxies. How does one detect which proxy is valid?
*/

/*
** The method systemProxyForQuery() may take several
** seconds to execute on Windows systems. Therefore,
** some intelligence may need to be introduced to
** enhance Dooble's response times.
*/

extern "C"
{
#include <errno.h>
#include <pthread.h>

GCRY_THREAD_OPTION_PTHREAD_IMPL;
}

int dmisc::s_algo = -1;
bool dmisc::s_cryptOK = true;
bool dmisc::s_passphraseWasAuthenticated = false;

/*
** The following variables are used
** by libgcrypt.
*/

static char *s_cipherPassphrase = 0;
static size_t s_cipherLength = 0;
static gcry_cipher_hd_t s_cipherCtx;

void dmisc::destroyCrypt(void)
{
  if(s_cipherPassphrase)
    gcry_free(s_cipherPassphrase);

  gcry_cipher_close(s_cipherCtx);
}

void dmisc::initializeCrypt(void)
{
  gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread, 0);

  if(!gcry_check_version(GCRYPT_VERSION))
    {
      s_cryptOK = false;
      qDebug() << "gcry_check_version() failure.";
    }
  else
    {
      gcry_control(GCRYCTL_ENABLE_M_GUARD);
      gcry_control(GCRYCTL_SUSPEND_SECMEM_WARN);

      if(gcry_control(GCRYCTL_INIT_SECMEM, 16384, 0) != 0)
	s_cryptOK = false;

      gcry_control(GCRYCTL_RESUME_SECMEM_WARN);
      gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);

      if(s_cryptOK)
	{
	  /*
	  ** Remember, initializeCrypt() is called before
	  ** a QApplication object is created. As a result,
	  ** a QSettings object may not be in the correct
	  ** context.
	  */

	  int algo = gcry_cipher_map_name
	    (findStrongestCipherType().toAscii().
	     constData());

	  if(gcry_cipher_open(&s_cipherCtx, algo,
			      GCRY_CIPHER_MODE_CBC,
			      GCRY_CIPHER_SECURE |
			      GCRY_CIPHER_CBC_CTS) != 0)
	    s_cryptOK = false;
	  else
	    s_algo = algo;
	}
    }
}

void dmisc::setCipherPassphrase(const QString &passphrase)
{
  if(s_cryptOK)
    {
      if(s_cipherPassphrase)
	gcry_free(s_cipherPassphrase);

      if(passphrase.isEmpty())
	s_cipherPassphrase = (char *) gcry_calloc_secure
	  (256, sizeof(char));
      else
	s_cipherPassphrase = (char *) gcry_calloc_secure
	  ((size_t) passphrase.length(), sizeof(char));

      if(!s_cipherPassphrase)
	{
	  s_cryptOK = false;
	  s_cipherLength = 0;
	  s_passphraseWasAuthenticated = false;
	  return;
	}

      if(passphrase.isEmpty())
	{
	  s_cipherLength = 256;
	  gcry_create_nonce
	    ((void *) s_cipherPassphrase, s_cipherLength);
	}
      else
	{
	  s_cipherLength = (size_t) passphrase.length();
	  memcpy((void *) s_cipherPassphrase,
		 (const void *) passphrase.toAscii().constData(),
		 (size_t) passphrase.length());
	  s_passphraseWasAuthenticated = true;
	}

      if(setKey() != 0)
	{
	  s_cryptOK = false;
	  gcry_free(s_cipherPassphrase);
	  s_cipherLength = 0;
	  s_cipherPassphrase = 0;
	  s_passphraseWasAuthenticated = false;
	}
      else if(!passphrase.isEmpty())
	/*
	** Save the cipher type.
	*/

	saveCorrectCipherType();
    }
}

QNetworkProxy dmisc::proxyByUrl(const QUrl &url)
{
  QString str(dooble::s_settings.value("settingsWindow/browsingProxySetting",
				       "none").toString());
  QNetworkProxy proxy;

  proxy.setType(QNetworkProxy::NoProxy);

  if(str == "system")
    {
      QNetworkProxyQuery query(url);
      QList<QNetworkProxy> list
	(QNetworkProxyFactory::systemProxyForQuery(query));

      if(!list.isEmpty())
	proxy = list.at(0);

      list.clear();
    }
  else if(str == "manual")
    {
      if(url.scheme().toLower() == "ftp")
	{
	  if(dooble::s_settings.value("settingsWindow/ftpBrowsingProxyEnabled",
				      false).toBool())
	    {
	      proxy.setHostName(dooble::s_settings.value
				("settingsWindow/ftpBrowsingProxyHost",
				 "").toString().trimmed());
	      proxy.setPort
		(static_cast<quint16> (dooble::s_settings.value
				       ("settingsWindow/ftpBrowsingProxyPort",
					21).toInt()));
	    }
	}

      if(url.scheme().toLower() == "http" ||
	 url.scheme().toLower() == "https")
	if(dooble::s_settings.value("settingsWindow/httpBrowsingProxyEnabled",
				    false).toBool())
	  {
	    QString str("");

	    str = dooble::s_settings.value
	      ("settingsWindow/httpBrowsingProxyType",
	       "Socks5").toString().trimmed().
	      toLower();

	    if(str == "socks5")
	      proxy.setType(QNetworkProxy::Socks5Proxy);
	    else if(str == "http")
	      proxy.setType(QNetworkProxy::HttpProxy);
	    else
	      proxy.setType(QNetworkProxy::Socks5Proxy);

	    proxy.setHostName
	      (dooble::s_settings.value("settingsWindow/httpBrowsingProxyHost",
					"").toString().trimmed());
	    proxy.setPort
	      (static_cast<quint16> (dooble::s_settings.value
				     ("settingsWindow/httpBrowsingProxyPort",
				      1080).toInt()));
	    proxy.setUser
	      (dooble::s_settings.value("settingsWindow/httpBrowsingProxyUser",
					"").toString().trimmed());
	    proxy.setPassword(dooble::s_settings.value
			      ("settingsWindow/httpBrowsingProxyPassword",
			       "").toString().trimmed());
	  }
    }

  /*
  ** Override!
  */

  if(url.host().trimmed().toLower().endsWith(".i2p") &&
     dooble::s_settings.value("settingsWindow/i2pBrowsingProxyEnabled",
			      true).toBool())
    {
      QString str("");

      str = dooble::s_settings.value("settingsWindow/i2pBrowsingProxyType",
				     "Http").toString().trimmed().
	toLower();

      if(str == "socks5")
	proxy.setType(QNetworkProxy::Socks5Proxy);
      else if(str == "http")
	proxy.setType(QNetworkProxy::HttpProxy);
      else
	proxy.setType(QNetworkProxy::HttpProxy);

      proxy.setHostName(dooble::s_settings.value
			("settingsWindow/i2pBrowsingProxyHost",
			 "127.0.0.1").toString().trimmed());
      proxy.setPort
	(static_cast<quint16> (dooble::s_settings.value
			       ("settingsWindow/i2pBrowsingProxyPort",
				4444).toInt()));
    }

  return proxy;
}

QNetworkProxy dmisc::proxyByFunctionAndUrl
(const DoobleDownloadType::DoobleDownloadTypeEnum functionType,
 const QUrl &url)
{
  QString str(dooble::s_settings.value("settingsWindow/browsingProxySetting",
				       "none").toString());
  QNetworkProxy proxy;

  proxy.setType(QNetworkProxy::NoProxy);

  switch(functionType)
    {
    case DoobleDownloadType::Ftp:
      {
	if(str == "system")
	  {
	    QNetworkProxyQuery query(url);
	    QList<QNetworkProxy> list
	      (QNetworkProxyFactory::systemProxyForQuery(query));

	    if(!list.isEmpty())
	      proxy = list.at(0);

	    list.clear();
	  }
	else if(str == "manual" &&
		dooble::s_settings.value
		("settingsWindow/ftpDownloadProxyEnabled",
		 false).toBool())
	  {
	    proxy.setHostName(dooble::s_settings.value
			      ("settingsWindow/ftpDownloadProxyHost",
			       "").toString().trimmed());
	    proxy.setPort
	      (static_cast<quint16> (dooble::s_settings.value
				     ("settingsWindow/ftpDownloadProxyPort",
				      21).toInt()));
	  }

	break;
      }
    case DoobleDownloadType::Http:
      {
	if(str == "system")
	  {
	    QNetworkProxyQuery query(url);
	    QList<QNetworkProxy> list
	      (QNetworkProxyFactory::systemProxyForQuery(query));

	    if(!list.isEmpty())
	      proxy = list.at(0);

	    list.clear();
	  }
	else if(str == "manual" &&
		dooble::s_settings.value
		("settingsWindow/httpDownloadProxyEnabled",
		 false).toBool())
	  {
	    QString str("");

	    str = dooble::s_settings.value
	      ("settingsWindow/httpDownloadProxyType",
	       "Socks5").toString().trimmed().
	      toLower();

	    if(str == "socks5")
	      proxy.setType(QNetworkProxy::Socks5Proxy);
	    else if(str == "http")
	      proxy.setType(QNetworkProxy::HttpProxy);
	    else
	      proxy.setType(QNetworkProxy::Socks5Proxy);

	    proxy.setHostName(dooble::s_settings.value
			      ("settingsWindow/httpDownloadProxyHost",
			       "").toString().trimmed());
	    proxy.setPort
	      (static_cast<quint16> (dooble::s_settings.value
				     ("settingsWindow/httpDownloadProxyPort",
				      1080).toInt()));
	    proxy.setUser(dooble::s_settings.value
			  ("settingsWindow/httpDownloadProxyUser",
			   "").toString().trimmed());
	    proxy.setPassword
	      (dooble::s_settings.value
	       ("settingsWindow/httpDownloadProxyPassword",
		"").toString().trimmed());
	  }

	break;
      }
    default:
      break;
    }

  /*
  ** Override!
  */

  if(url.host().trimmed().toLower().endsWith(".i2p") &&
     dooble::s_settings.value("settingsWindow/i2pDownloadProxyEnabled",
			      true).toBool())
    {
      QString str("");

      str = dooble::s_settings.value("settingsWindow/i2pDownloadProxyType",
				     "Http").toString().trimmed().
	toLower();

      if(str == "socks5")
	proxy.setType(QNetworkProxy::Socks5Proxy);
      else if(str == "http")
	proxy.setType(QNetworkProxy::HttpProxy);
      else
	proxy.setType(QNetworkProxy::HttpProxy);

      proxy.setHostName(dooble::s_settings.value
			("settingsWindow/i2pDownloadProxyHost",
			 "127.0.0.1").toString().trimmed());
      proxy.setPort
	(static_cast<quint16> (dooble::s_settings.value
			       ("settingsWindow/i2pDownloadProxyPort",
				4444).toInt()));
    }

  return proxy;
}

QIcon dmisc::iconForUrl(const QUrl &url)
{
  QFileIconProvider iconProvider;

  if(url.scheme().toLower() == "http" ||
     url.scheme().toLower() == "https")
    {
      QIcon icon;

      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE",
						    "history_dmisc");

	db.setDatabaseName(dooble::s_homePath +
			   QDir::separator() + "history.db");

	if(db.open())
	  {
	    QSqlQuery query(db);

	    query.exec
	      (QString
	       ("SELECT icon FROM history WHERE "
		"HEX(url) = "
		"HEX('%1') OR HEX(host) = HEX('%2')").
	       arg(encodedString
		   (url.toString(QUrl::StripTrailingSlash).toAscii(),
		    true).toBase64().data()).
	       arg(encodedString(url.host().toAscii(),
				 true).toBase64().data()));

	    if(query.next())
	      if(!query.isNull(0))
		{
		  QBuffer buffer;
		  QByteArray bytes
		    (QByteArray::fromBase64(query.value(0).
					    toByteArray()));

		  bytes = decodedString(bytes);
		  buffer.setBuffer(&bytes);
		  buffer.open(QIODevice::ReadOnly);

		  QDataStream in(&buffer);

		  in >> icon;
		  buffer.close();
		}
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("history_dmisc");

      if(!icon.isNull())
	return icon;
    }
  else if(url.scheme().toLower() == "ftp")
    return iconProvider.icon(QFileIconProvider::Network);
  else if(url.scheme().toLower() == "file")
    {
      QFileInfo fileInfo(url.toLocalFile());

      if(url.toLocalFile().isEmpty())
	return iconProvider.icon(QFileIconProvider::Folder);
      else if(fileInfo.isFile())
	return iconProvider.icon(QFileIconProvider::File);
      else if(fileInfo.isDir())
	return iconProvider.icon(QFileIconProvider::Folder);
    }

  QSettings settings
    (dooble::s_settings.value("iconSet").toString(), QSettings::IniFormat);

  return QIcon(settings.value("mainWindow/emptyIcon").toString());
}

QString dmisc::saveCorrectCipherType(void)
{
  QString type("");

  if(dooble::s_settings.contains("cipherType"))
    {
      type = dooble::s_settings.value("cipherType").toString();

      if(!isCipherTypeSupported(type))
	{
	  /*
	  ** Overwrite the current cipherType if it is
	  ** not supported.
	  */

	  type = findStrongestCipherType();

	  if(!type.isEmpty())
	    {
	      dooble::s_settings["cipherType"] = type;

	      QSettings settings;

	      settings.setValue("cipherType", type);
	    }
	}
      else if(type != findStrongestCipherType())
	{
	  type = findStrongestCipherType();

	  if(!type.isEmpty())
	    {
	      dooble::s_settings["cipherType"] = type;

	      QSettings settings;

	      settings.setValue("cipherType", type);
	    }
	}
    }
  else
    {
      type = findStrongestCipherType();

      if(!type.isEmpty())
	{
	  dooble::s_settings["cipherType"] = type;

	  QSettings settings;

	  settings.setValue("cipherType", type);
	}
    }

  return type;
}

QByteArray dmisc::decodedString(const QByteArray &byteArray)
{
  if(s_cryptOK && s_cipherPassphrase)
    {
      static size_t blockLength = gcry_cipher_get_algo_blklen(s_algo);
      QByteArray decodedArray(byteArray);
      gcry_error_t err = 0;

      /*
      ** Block ciphers require the length of the buffers
      ** to be multiples of the cipher's block size.
      */

      if(decodedArray.isEmpty())
	decodedArray = decodedArray.leftJustified(blockLength, 0);
      else
	decodedArray = decodedArray.leftJustified
	  (blockLength * qCeil((qreal) decodedArray.size() / (qreal)
			       blockLength), 0);

      gcry_cipher_reset(s_cipherCtx);

      if((err = setInitializationVector()) != 0)
	{
	  qDebug() << "setInitializationVector() failure.";
	  return byteArray;
	}
      else if((err = gcry_cipher_decrypt(s_cipherCtx,
					 (void *) decodedArray.data(),
					 (size_t) decodedArray.size(),
					 (const void *) 0,
					 (size_t) 0)) == 0)
	return decodedArray;
      else
	{
	  qDebug() << "gcry_cipher_decrypt() failure.";
	  return byteArray;
	}
    }
  else
    return byteArray;
}

QByteArray dmisc::encodedString(const QByteArray &byteArray,
				const bool shouldEncode)
{
  if(s_cryptOK && shouldEncode && s_cipherPassphrase)
    {
      static size_t blockLength = gcry_cipher_get_algo_blklen(s_algo);
      QByteArray encodedArray(byteArray);
      gcry_error_t err = 0;

      /*
      ** Block ciphers require the length of the buffers
      ** to be multiples of the cipher's block size.
      */

      if(encodedArray.isEmpty())
	encodedArray = encodedArray.leftJustified(blockLength, 0);
      else
	encodedArray = encodedArray.leftJustified
	  (blockLength * qCeil((qreal) encodedArray.size() / (qreal)
			       blockLength), 0);

      gcry_cipher_reset(s_cipherCtx);

      if((err = setInitializationVector()) != 0)
	{
	  qDebug() << "setInitializationVector() failure.";
	  return byteArray;
	}
      else if((err = gcry_cipher_encrypt(s_cipherCtx,
					 (void *) encodedArray.data(),
					 (size_t) encodedArray.size(),
					 (const void *) 0,
					 (size_t) 0)) == 0)
	return encodedArray;
      else
	{
	  qDebug() << "gcry_cipher_encrypt() failure.";
	  return byteArray;
	}
    }
  else
    return byteArray;
}

QString dmisc::findStrongestHashType(void)
{
  QString type("");
  QStringList types;

  types << "sha512"
	<< "sha384"
	<< "sha256"
	<< "sha224"
	<< "tiger"
	<< "md5";

  for(int i = 0; i < types.size(); i++)
    if(isHashTypeSupported(types.at(i)))
      {
	type = types.at(i);
	break;
      }

  types.clear();
  return type;
}

QString dmisc::findStrongestCipherType(void)
{
  QString type("");
  QStringList types;

  /*
  ** Block ciphers only!
  */

  types << "aes256"
	<< "camellia256"
	<< "serpent256"
	<< "twofish"
	<< "aes192"
	<< "aes128";

  for(int i = 0; i < types.size(); i++)
    if(isCipherTypeSupported(types.at(i)))
      {
	type = types.at(i);
	break;
      }

  types.clear();
  return type;
}

QString dmisc::passphraseHash(const QString &hashType)
{
  if(s_cryptOK && s_cipherPassphrase)
    return passphraseHash(QByteArray(s_cipherPassphrase,
				     s_cipherLength).data(),
			  hashType);
  else
    return "";
}

QString dmisc::passphraseHash(const QString &passphrase,
			      const QString &hashType)
{
  QString hash("");

  if(s_cryptOK)
    {
      int algo = -1;

      if(hashType == "sha512")
	algo = GCRY_MD_SHA512;
      else if(hashType == "sha384")
	algo = GCRY_MD_SHA384;
      else if(hashType == "sha256")
	algo = GCRY_MD_SHA256;
      else if(hashType == "sha224")
	algo = GCRY_MD_SHA224;
      else if(hashType == "tiger")
	algo = GCRY_MD_TIGER;
      else if(hashType == "md5")
	algo = GCRY_MD_MD5;

      if(algo != -1 && isHashTypeSupported(hashType))
	{
	  char buffer[gcry_md_get_algo_dlen(algo) + 1];

	  memset(buffer, 0, sizeof(buffer));
	  gcry_md_hash_buffer(algo,
			      (void *) buffer,
			      (const void *) passphrase.toAscii().constData(),
			      (size_t) passphrase.length());
	  hash = QByteArray(buffer).toHex();
	}
    }

  return hash;
}

bool dmisc::isHashTypeSupported(const QString &hashType)
{
  if(s_cryptOK)
    {
      int algo = gcry_md_map_name(hashType.toAscii().constData());

      if(algo != 0 && gcry_md_test_algo(algo) == 0)
	return true;
      else
	return false;
    }
  else
    return false;
}

bool dmisc::isCipherTypeSupported(const QString &cipherType)
{
  if(s_cryptOK)
    {
      if(gcry_cipher_map_name(cipherType.toAscii().constData()) != 0)
	return true;
      else
	return false;
    }
  else
    return false;
}

gcry_error_t dmisc::setKey(void)
{
  gcry_error_t err = 0;

  if(s_cryptOK && s_cipherPassphrase)
    {
      size_t keyLength = 0;

      if((keyLength = gcry_cipher_get_algo_keylen(s_algo)) != 0)
	err = gcry_cipher_setkey
	  (s_cipherCtx, (const void *) s_cipherPassphrase,
	   keyLength);
    }
  else if(!s_cipherPassphrase)
    err = GPG_ERR_ENOMEM;

  return err;
}

gcry_error_t dmisc::setInitializationVector(void)
{
  gcry_error_t err = 0;

  if(s_cryptOK && s_cipherPassphrase)
    {
      size_t ivLength = 0;

      if((ivLength = gcry_cipher_get_algo_blklen(s_algo)) != 0)
	{
	  char *iv = (char *) gcry_calloc_secure(ivLength, sizeof(char));

	  if(iv)
	    {
	      memcpy
		((void *) iv, (const void *) s_cipherPassphrase,
		 qMin(ivLength, s_cipherLength));
	      err = gcry_cipher_setiv
		(s_cipherCtx, (const void *) iv,
		 ivLength);
	      gcry_free(iv);
	    }
	  else
	    err = GPG_ERR_ENOMEM;
	}
      else
	err = GPG_ERR_INV_LENGTH;
    }
  else if(!s_cipherPassphrase)
    err = GPG_ERR_ENOMEM;

  return err;
}

int dmisc::levenshteinDistance(const QString &str1,
			       const QString &str2)
{
  if(str1.isEmpty())
    return str2.length();
  else if(str2.isEmpty())
    return str1.length();

  int cost = 0;
  int matrix[str1.length() + 1][str2.length() + 1];
  QChar str1_c = 0;
  QChar str2_c = 0;

  for(int i = 0; i <= str1.length(); i++)
    matrix[i][0] = i;

  for(int i = 0; i <= str2.length(); i++)
    matrix[0][i] = i;

  for(int i = 1; i <= str1.length(); i++)
    {
      str1_c = str1.at(i - 1);

      for(int j = 1; j <= str2.length(); j++)
	{
	  str2_c = str2.at(j - 1);

	  if(str1_c == str2_c)
	    cost = 0;
	  else
	    cost = 1;

	  matrix[i][j] = qMin(qMin(matrix[i - 1][j] + 1,
				   matrix[i][j - 1] + 1),
			      matrix[i - 1][j - 1] + cost);
	}
    }

  return matrix[str1.length()][str2.length()];
}

bool dmisc::passphraseWasAuthenticated(void)
{
  return s_passphraseWasAuthenticated;
}

QString dmisc::findUniqueFileName(const QString &fileName,
				  const QDir &path)
{
  QFileInfo fileInfo;

  fileInfo.setFile(path, fileName);

  if(fileInfo.exists())
    {
      int count = 1;

      fileInfo.setFile(path, QString("%1(%2)").arg(fileName).arg(count));

      while(fileInfo.exists())
	{
	  count += 1;
	  fileInfo.setFile(path, QString("%1(%2)").arg(fileName).arg(count));
	}
    }

  return fileInfo.absoluteFilePath();
}
