/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include "dooble.h"
#include "dnetworkaccessmanager.h"

dnetworkaccessmanager::dnetworkaccessmanager(QObject *parent):
  QNetworkAccessManager(parent)
{
  connect(this,
	  SIGNAL(finished(QNetworkReply *)),
	  this,
	  SLOT(slotFinished(QNetworkReply *)));
}

QNetworkReply *dnetworkaccessmanager::createRequest
(Operation op,
 const QNetworkRequest &req, QIODevice *outgoingData)
{
  QNetworkRequest request(req);

  if(!request.url().isEmpty() && request.url().isValid())
    {
      request.setAttribute(QNetworkRequest::HttpPipeliningAllowedAttribute,
			   true);

      if(dooble::s_settings.value("settingsWindow/blockThirdPartyContent",
				  false).toBool())
	{
	  QWebFrame *frame = qobject_cast<QWebFrame *>
	    (request.originatingObject());

	  if(frame && !qobject_cast<dwebpage *> (frame->parent()))
	    {
	      /*
	      ** If the frame's parent is not a dwebpage, we'll
	      ** assume that the request should be blocked unless
	      ** the frame's URL is consistent with the request's
	      ** referer.
	      */

	      QString domain1(request.url().host());

	      if(!dooble::s_adBlockWindow->allowed(domain1))
		{
		  QString domain2(QUrl(request.rawHeader("Referer")).host());
		  QStringList list1(domain1.split("."));
		  QStringList list2(domain2.split("."));

		  if(list1.size() > 1)
		    domain1 = list1.at(list1.size() - 2) + "." +
		      list1.at(list1.size() - 1);

		  if(list2.size() > 1)
		    domain2 = list2.at(list2.size() - 2) + "." +
		      list2.at(list2.size() - 1);

		  if(!(domain1.contains(domain2) || domain2.contains(domain1)))
		    {
		      emit blockThirdPartyHost
			(request.url().host(),
			 QUrl(request.rawHeader("Referer")),
			 QDateTime::currentDateTime());
		      request = QNetworkRequest();
		    }
		}
	    }
	}

      if(dooble::s_settings.value("settingsWindow/suppressHttpReferrer", true).
	 toBool())
	if(!dooble::s_httpReferrerWindow->allowed(request.url().host()))
	  {
	    emit suppressHttpReferrer(request.url().host(),
				      request.url(),
				      QDateTime::currentDateTime());
	    request.setRawHeader("Referer", 0);
	  }
    }

  return QNetworkAccessManager::createRequest(op, request, outgoingData);
}

void dnetworkaccessmanager::slotFinished(QNetworkReply *reply)
{
  if(!reply)
    return;

  if(dooble::s_settings.value("settingsWindow/suppressHttpRedirect",
			      true).toBool())
    {
      QUrl url(reply->attribute(QNetworkRequest::RedirectionTargetAttribute).
	       toUrl());

      /*
      ** The URL may be relative.
      */

      url = reply->url().resolved(url);

      if(url.isEmpty() || !url.isValid() ||
	 url.toString(QUrl::StripTrailingSlash) ==
	 reply->url().toString(QUrl::StripTrailingSlash))
	{
	  /*
	  ** What's up?
	  */
	}
      else
	{
	  /*
	  ** Alright. Here we are. We have a redirect.
	  ** We can either completely stop the current load or
	  ** continue with this redirect if it intends
	  ** to retrieve secondary data for the originating object.
	  */

	  /*
	  ** Is WebKit aware of the redirect before this method is
	  ** reached? How so? The reply includes information containing
	  ** redirection information. WebKit must act upon that information.
	  ** But when? Does it matter? I wish createRequest()'s request
	  ** contained some information that would suggest that it's a
	  ** redirect. If it does, where is it?
	  */

	  /*
	  ** You may have a URL that redirects to some other URL
	  ** just as the originating object is about to fully load.
	  ** Stopping the load then would seem strange.
	  */

	  /*
	  ** Let's keep it simple and abort the current load
	  ** if any prohibited redirect occurs.
	  */

	  if(!dooble::s_httpRedirectWindow->allowed(url.host()))
	    {
	      QWebFrame *frame = qobject_cast<QWebFrame *> 
		 (reply->request().originatingObject());

	      if(frame && frame == frame->page()->mainFrame())
		{
		  frame->page()->triggerAction
		    (QWebPage::StopScheduledPageRefresh);
		  frame->page()->triggerAction(QWebPage::Stop);
		}

	      emit urlRedirectionRequest(url.host(),
					 reply->url(),
					 QDateTime::currentDateTime());
	    }
	}
    }

  reply->deleteLater();
}
