/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Alexis Megas, Bernd Stramm,
** Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include <string>
#include <iostream>

#include <QDir>
#include <QDebug>
#include <QTimer>
#include <QBuffer>
#include <QPrinter>
#include <QProcess>
#include <QLineEdit>
#ifdef Q_WS_MAC
#include <QMacStyle>
#elif defined Q_WS_WIN
#include <QWindowsStyle>
#endif
#include <QSettings>
#include <QSplitter>
#include <QSqlQuery>
#include <QWebFrame>
#include <QClipboard>
#include <QCloseEvent>
#include <QFileDialog>
#include <QMessageBox>
#include <QTranslator>
#include <QApplication>
#include <QPaintEngine>
#include <QPrintDialog>
#include <QSqlDatabase>
#include <QPluginLoader>
#include <QNetworkReply>
#include <QAuthenticator>
#include <QDesktopWidget>
#include <QDesktopServices>
#include <QCoreApplication>
#include <QPrintPreviewDialog>

#include "dmisc.h"
#include "dooble.h"
#include "dhistory.h"
#include "dplugintab.h"
#include "dpagesourcewindow.h"
#include "ui_passwordPrompt.h"
#include "ui_passphrasePrompt.h"
#include "dnetworkaccessmanager.h"

using namespace simpleplugin;

/*
** Global objects.
*/

QString dooble::s_homePath = "";
dcookies *dooble::s_cookies = 0;
dhistory *dooble::s_historyWindow = 0;
dsettings *dooble::s_settingsWindow = 0;
dcookiewindow *dooble::s_cookieWindow = 0;
dhistorymodel *dooble::s_historyModel = 0;
ddownloadwindow *dooble::s_downloadWindow = 0;
dexceptionswindow *dooble::s_popupsWindow = 0;
dexceptionswindow *dooble::s_adBlockWindow = 0;
dexceptionswindow *dooble::s_cookiesBlockWindow = 0;
dexceptionswindow *dooble::s_httpRedirectWindow = 0;
dexceptionswindow *dooble::s_httpReferrerWindow = 0;
dexceptionswindow *dooble::s_javaScriptExceptionsWindow = 0;
QHash<QString, int> dooble::s_mostVisitedHosts;
QHash<QString, QVariant> dooble::s_settings;

/*
** Static (local) objects.
*/

static int doobleCounter = 0;

int main(int argc, char *argv[])
{
  /*
  ** First thing we must do.
  */

  dmisc::initializeCrypt();

  /*
  ** Qt recommends to set the style prior
  ** to defining the QApplication object.
  */

#ifdef Q_WS_MAC
  QApplication::setStyle(new QMacStyle());
#elif defined Q_WS_WIN
  QApplication::setStyle(new QWindowsStyle());
#elif defined Q_WS_X11
  QApplication::setGraphicsSystem("raster");
#endif

  /*
  ** Proceed to Qt initialization.
  */

  QApplication qapp(argc, argv);

  /*
  ** Process command-line arguments.
  */

  QStringList urls;
  QHash<QString, QVariant> argumentsHash;

  if(argc > 1)
    {
      int rc = -1;
      QString usage("");

      usage.append("Usage: Dooble [OPTION]...\n\n");
      usage.append("Options:\n");
      usage.append("--full-screen-mode\n");
      usage.append("--url URL");

      for(int i = 1; i < argc; i++)
	{
	  QString option(argv[i]);

	  if(option == "--full-screen-mode")
	    argumentsHash["full-screen-mode"] = true;
	  else if(option == "--url")
	    {
	      i += 1;

	      if(i < argc)
		{
		  if(QUrl::fromUserInput(argv[i]).isValid())
		    urls.append(QString(argv[i]).trimmed());
		  else
		    std::cerr << "Invalid URL: "
			      << argv[i]
			      << ". Skipping." << std::endl;
		}
	      else
		{
		  std::cerr << "Missing argument to --url." << std::endl;
		  rc = 1;
		  break;
		}
	    }
	  else if(option == "--help")
	    {
	      rc = 0;
	      break;
	    }
	  else
	    {
	      rc = 1;
	      break;
	    }
	}

      if(rc + 1 > 0)
	{
#if QT_VERSION >= 0x040700
	  std::cout << usage.toStdString() << std::endl;
#else
	  std::cout << usage.toAscii().constData() << std::endl;
#endif
	  return rc;
	}
    }

  /*
  ** Configure translations.
  */

  QTranslator qtTranslator;

  qtTranslator.load("qt_" + QLocale::system().name(), "Translations");
  qapp.installTranslator(&qtTranslator);

  QTranslator myappTranslator;

  myappTranslator.load("dooble_" + QLocale::system().name(),
		       "Translations");
  qapp.installTranslator(&myappTranslator);

#ifdef Q_WS_WIN
  char *tmp = getenv("USERNAME");
  QDir homeDir(QDir::current());
  QString username("");
  QFileInfo fileInfo(homeDir.absolutePath());

  if(!(fileInfo.isReadable() && fileInfo.isWritable()))
    {
      homeDir = QDir::home();
      homeDir.mkdir(".dooble");
    }
  else
    {
      if(tmp)
	username = QString(tmp);

      if(username.isEmpty())
	homeDir.mkdir(".dooble");
      else
	homeDir.mkdir(username + QDir::separator() + ".dooble");

      if(username.isEmpty())
	dooble::s_homePath = homeDir.absolutePath() +
	  QDir::separator() + ".dooble";
      else
	dooble::s_homePath = homeDir.absolutePath() + QDir::separator() +
	  username + QDir::separator() + ".dooble";
    }
#else
  QDir homeDir(QDir::home());

  homeDir.mkdir(".dooble");
  dooble::s_homePath = homeDir.absolutePath() +
    QDir::separator() + ".dooble";
#endif

  /*
  ** Disable Web page icons.
  */

  QWebSettings::setIconDatabasePath("");

  /*
  ** Set application attributes.
  */

  QCoreApplication::setApplicationName("Dooble");
  QCoreApplication::setOrganizationName("Dooble");
  QCoreApplication::setOrganizationDomain("dooble.sf.net");
  QCoreApplication::setApplicationVersion(DOOBLE_VERSION_STR);
  QSettings::setPath(QSettings::IniFormat,
		     QSettings::UserScope,
		     dooble::s_homePath);
  QSettings::setDefaultFormat(QSettings::IniFormat);

  /*
  ** Remove old configuration settings.
  */

  QSettings settings;

  settings.remove("popupsWindow");
  settings.remove("adBlockWindow");
  settings.remove("exceptionsWindow");
  settings.remove("httpReferrerWindow");
  settings.remove("mainWindow/pos");
  settings.remove("mainWindow/size");
  settings.remove("mainWindow/state");
  settings.remove("mainWindow/state1");
  settings.remove("mainWindow/searchWidth");
  settings.remove("mainWindow/locationWidth");
  settings.remove("mainWindow/ftpManagerColumnsState");
  settings.remove("mainWindow/fileManagerColumnsState");
  settings.remove("cookieWindow");
  settings.remove("cookiesWindow/size");
  settings.remove("cookiesWindow/tableColumnsState");
  settings.remove("historyWindow/size");
  settings.remove("historyWindow/tableColumnState");
  settings.remove("historyWindow/tableColumnState1");
  settings.remove("historyWindow/tableColumnState2");
  settings.remove("downloadWindow/pos");
  settings.remove("downloadWindow/size");
  settings.remove("settingsWindow/size");
  settings.remove("settingsWindow/url12");
  settings.remove("settingsWindow/saveHistory");

  if(settings.contains("mainWindow/showBookmarksToolBar"))
    /*
    ** Retain previous setting.
    */

    settings.setValue
      ("mainWindow/showFavoritesToolBar",
       settings.value("mainWindow/showBookmarksToolBar").toBool());

  if(settings.contains("settingsWindow/enablePlugins"))
    /*
    ** Retain previous setting.
    */

    settings.setValue
      ("settingsWindow/enableWebPlugins",
       settings.value("settingsWindow/enablePlugins").toBool());

  if(settings.contains("settingsWindow/suppressHttpReferer"))
    /*
    ** Retain previous setting.
    */

    settings.setValue
      ("settingsWindow/suppressHttpReferrer",
       settings.value("settingsWindow/suppressHttpReferer").toBool());

  settings.remove("mainWindow/showBookmarksToolBar");
  settings.remove("settingsWindow/enablePlugins");
  settings.remove("settingsWindow/privateBrowsing");
  settings.remove("settingsWindow/suppressHttpReferer");
  settings.remove("settingsWindow/saveHistoryCheckedId");
  settings.remove("settingsWindow/encryptCookieInformation");
  settings.remove("settingsWindow/encryptHistoryInformation");
  settings.remove("settingsWindow/encryptBookmarkInformation");
  settings.remove("settingsWindow/encryptDownloadInformation");

  if(!settings.contains("containers_cleared_v122") ||
     !settings.contains("containers_cleared_v124"))
    {
      /*
      ** Versions 1.22 and 1.24 required database changes.
      */

      settings.setValue("containers_cleared_v122", true);
      settings.setValue("containers_cleared_v124", true);

      QStringList containers;

      containers << "cookies.db"
		 << "history.db"
		 << "downloads.db";

      while(!containers.isEmpty())
	QFile::remove(dooble::s_homePath +
		      QDir::separator() +
		      containers.takeFirst());
    }

  QList<QString> list(settings.allKeys());
  QMutableListIterator<QString> it(list);

  while(it.hasNext())
    {
      QString str(it.next());

      if(str.startsWith("settingsWindow/backgroundImage"))
	settings.remove(it.value());
      else if(str.contains("proxy"))
	settings.remove(it.value());
#if DOOBLE_VERSION < 0x012301
      else if(str.contains("Proxy") &&
	      !str.contains("Browsing") &&
	      !str.contains("Download"))
	settings.remove(it.value());
#endif
    }

  list.clear();

  /*
  ** We need to set this before creating some of the support windows.
  */

  if(!settings.contains("iconSet"))
    settings.setValue("iconSet",
		      QString("%1/%2").arg(QDir::currentPath()).
		      arg("Icons/nuovext/configuration.cfg"));

  if(!QFileInfo(settings.value("iconSet").toString()).exists())
    settings.setValue("iconSet",
		      QString("%1/%2").arg(QDir::currentPath()).
		      arg("Icons/nuovext/configuration.cfg"));

  if(!settings.contains("settingsWindow/iconSet1"))
    settings.setValue("settingsWindow/iconSet1", settings.value("iconSet"));

  if(!QFileInfo(settings.value("settingsWindow/iconSet1").toString()).exists())
    settings.setValue("settingsWindow/iconSet1",
		      QString("%1/%2").arg(QDir::currentPath()).
		      arg("Icons/nuovext/configuration.cfg"));

  dooble::s_settings.clear();

  for(int i = 0; i < settings.allKeys().size(); i++)
    dooble::s_settings[settings.allKeys().at(i)] =
      settings.value(settings.allKeys().at(i));

  /*
  ** QWebSettings changes have to be performed after dooble::s_settings
  ** gets populated!
  */

  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::JavaEnabled,
     dooble::s_settings.value("settingsWindow/javaEnabled", false).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::JavascriptEnabled,
     dooble::s_settings.value("settingsWindow/javascriptEnabled",
			      true).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::JavascriptCanOpenWindows,
     dooble::s_settings.value("settingsWindow/javascriptEnabled",
			      true).toBool() &&
     dooble::s_settings.value("settingsWindow/javascriptAllowNewWindows",
			      false).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::AutoLoadImages,
     dooble::s_settings.value("settingsWindow/automaticallyLoadImages",
			      true).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::PluginsEnabled,
     dooble::s_settings.value("settingsWindow/enableWebPlugins",
			      true).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::DnsPrefetchEnabled, true);
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::PrivateBrowsingEnabled, true);
#if QT_VERSION >= 0x040700
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::SpatialNavigationEnabled,
     dooble::s_settings.value("settingsWindow/spatialNavigation",
			      false).toBool());
  QWebSettings::globalSettings()->setAttribute
    (QWebSettings::XSSAuditingEnabled,
     dooble::s_settings.value("settingsWindow/xssAuditingEnabled",
			      false).toBool());
#endif
  QWebSettings::globalSettings()->setDefaultTextEncoding
    (dooble::s_settings.value("settingsWindow/characterEncoding",
			      "").toString().toLower());

  QFont font;

  if(!font.fromString
     (dooble::s_settings.value("settingsWindow/standardWebFont", "").
      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Times", 16);
#elif defined Q_WS_WIN
    font = QFont("Serif", 16);
#else
    font = QFont("Serif", 16);
#endif

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::StandardFont, font.family());
  QWebSettings::globalSettings()->setFontSize
    (QWebSettings::DefaultFontSize, font.pointSize());

  if(!font.fromString(dooble::s_settings.value
		      ("settingsWindow/cursiveWebFont", "").
		      toString()) || font.family().isEmpty())
    font = QFont("Serif");

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::CursiveFont,
     font.family());

  if(!font.fromString(dooble::s_settings.value
		      ("settingsWindow/fantasyWebFont", "").
		      toString()) || font.family().isEmpty())
    font = QFont("Serif");

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::FantasyFont,
     font.family());

  if(!font.fromString(dooble::s_settings.value
		      ("settingsWindow/fixedWebFont", "").
		      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Courier", 13);
#elif defined Q_WS_WIN
    font = QFont("Courier New", 16);
#else
    font = QFont("Courier", 10);
#endif

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::FixedFont,
     font.family());
  QWebSettings::globalSettings()->setFontSize
    (QWebSettings::DefaultFixedFontSize,
     font.pointSize());

  if(!font.fromString(dooble::s_settings.value
		      ("settingsWindow/sansSerifWebFont", "").
		      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Helvetica");
#elif defined Q_WS_WIN
    font = QFont("Arial");
#else
    font = QFont("Sans Serif");
#endif

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::SansSerifFont,
     font.family());

  if(!font.fromString(dooble::s_settings.value
		      ("settingsWindow/serifWebFont", "").
		      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Times");
#elif defined Q_WS_WIN
    font = QFont("Times New Roman");
#else
    font = QFont("Serif");
#endif

  font.setWeight(QFont::Normal);
  QWebSettings::globalSettings()->setFontFamily
    (QWebSettings::SerifFont,
     font.family());
  QWebSettings::globalSettings()->setFontSize
    (QWebSettings::MinimumFontSize,
     dooble::s_settings.value("settingsWindow/minimumWebFontSize",
			      9).toInt());

  /*
  ** Initialize static members.
  */

  dooble::s_historyModel = new dhistorymodel();
  dooble::s_popupsWindow = new dexceptionswindow
    (new dexceptionsmodel("popupsexceptions"));
  dooble::s_popupsWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "JavaScript Popups Exceptions"));
  dooble::s_adBlockWindow = new dexceptionswindow
    (new dexceptionsmodel("adblockexceptions"));
  dooble::s_adBlockWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "Third-Party Blocking Exceptions"));
  dooble::s_cookiesBlockWindow = new dexceptionswindow
    (new dexceptionsmodel("cookiesexceptions"));
  dooble::s_cookiesBlockWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "Cookies Exceptions"));
  dooble::s_httpRedirectWindow = new dexceptionswindow
    (new dexceptionsmodel("suppresshttpredirectexceptions"));
  dooble::s_httpRedirectWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "Suppress HTTP Redirect Exceptions"));
  dooble::s_httpReferrerWindow = new dexceptionswindow
    (new dexceptionsmodel("suppresshttpreferrerexceptions"));
  dooble::s_httpReferrerWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "Suppress HTTP Referrer Exceptions"));
  dooble::s_javaScriptExceptionsWindow = new dexceptionswindow
    (new dexceptionsmodel("javascriptexceptions"));
  dooble::s_javaScriptExceptionsWindow->setWindowTitle
    (QObject::tr("Dooble Web Browser: "
		 "JavaScript Exceptions"));
  dfilemanager::treeModel = new QFileSystemModel();
  dfilemanager::tableModel = new dfilesystemmodel();
  dfilemanager::treeModel->setReadOnly(false);
  dfilemanager::tableModel->setReadOnly(false);
  dfilemanager::treeModel->setFilter(QDir::Drives | QDir::AllDirs |
				     QDir::System |
				     QDir::NoDotAndDotDot);
  dfilemanager::treeModel->setRootPath(QDir::rootPath());
  dfilemanager::tableModel->setFilter(QDir::Hidden | QDir::AllEntries |
			       QDir::System | QDir::NoDotAndDotDot);
  dfilemanager::tableModel->setRootPath(QDir::rootPath());

  if(dooble::s_settings.value("mainWindow/showHiddenFiles", true).toBool())
    {
      dfilemanager::treeModel->setFilter
	(dfilemanager::treeModel->filter() | QDir::Hidden);
      dfilemanager::tableModel->setFilter
	(dfilemanager::tableModel->filter() | QDir::Hidden);
    }

  /*
  ** Create singular instances.
  */

  if(!font.fromString(dooble::s_settings.value("settingsWindow/globalFont",
					       "").
		      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Times", 12);
#elif defined Q_WS_WIN
    font = QFont("Arial", 10);
#else
    font = QFont("Serif", 9);
#endif

  font.setWeight(QFont::Normal);
  QApplication::setFont(font);
  dooble::s_historyWindow = new dhistory();
  dooble::s_downloadWindow = new ddownloadwindow();
  dooble::s_settingsWindow = new dsettings();
  dooble::s_cookies = new dcookies(); /*
				      ** This object (dooble::s_cookies)
				      ** must be created after
				      ** dooble::s_cookiesBlockWindow.
				      */
  dooble::s_cookieWindow = new dcookiewindow();
  QObject::connect(dooble::s_cookies,
		   SIGNAL(cookiesChanged(const QList<QNetworkCookie> &)),
		   dooble::s_cookieWindow,
		   SLOT(slotPopulate(const QList<QNetworkCookie> &)));
  QObject::connect(dooble::s_cookies,
		   SIGNAL(domainsRemoved(const QStringList &)),
		   dooble::s_cookieWindow,
		   SLOT(slotDomainsRemoved(const QStringList &)));
  QObject::connect(dooble::s_cookies,
		   SIGNAL(cookiesChangedForUrl(const
					       QList<QNetworkCookie> &,
					       const QUrl &)),
		   dooble::s_cookieWindow,
		   SLOT(slotUpdateView(const
				       QList<QNetworkCookie> &,
				       const QUrl &)));
  QObject::connect(dooble::s_settingsWindow,
		   SIGNAL(cookieTimerChanged(void)),
		   dooble::s_cookies,
		   SLOT(slotCookieTimerChanged(void)));

  QUrl url;
  QLocale locale;
  QString urlText
    (dooble::s_settings.value
     ("settingsWindow/url12_1",
      QString("qrc:/search_%1.html").arg(locale.name().left(2))).toString());

  url = QUrl::fromUserInput(urlText);

  if(!url.isValid())
    urlText = "";

  if(argc > 1)
    {
      urls.append(urlText);
      argumentsHash["urls"] = urls;
      urls.clear();
      Q_UNUSED(new dooble(argumentsHash));
      argumentsHash.clear();
    }
  else
    {
      Q_UNUSED(new dooble(urlText));
    }

  /*
  ** If the last Dooble window is closed, Dooble will exit.
  ** So, it's not necessary to connect QApplication's
  ** lastWindowClosed() signal.
  */

  int rc = qapp.exec();

  dmisc::destroyCrypt();
  return rc;
}

void dooble::init(const bool isJavaScriptWindow)
{
  /*
  ** This method is used whenever a new Dooble window
  ** is about to be created.
  */

  setUrlHandler(this);
  m_isJavaScriptWindow = isJavaScriptWindow;
  showFindFrame = false;
  doobleCounter += 1;
  ui.setupUi(this);
  ui.historyFrame->setLayout(new QHBoxLayout(ui.historyFrame));
  ui.historyFrame->layout()->setContentsMargins(1, 0, 1, 0);
  m_historySideBar = new dhistorysidebar(this);
  ui.historyFrame->layout()->addWidget(m_historySideBar);
  ui.actionShow_HistorySideBar->setChecked
    (s_settings.value("mainWindow/showHistorySideBar", false).
     toBool());
  ui.historyFrame->setVisible(false);
#ifdef Q_WS_MAC
  setAttribute(Qt::WA_MacMetalStyle, true);
#endif
  ui.urlFrame->setParent(0);
  ui.toolBar->setVisible(false);
  ui.toolBar->addWidget(ui.urlFrame);
  ui.toolBar->setVisible(true);
  ui.favoritesFrame->setParent(0);
  ui.favoritesToolBar->setVisible(false);
  ui.favoritesToolBar->addWidget(ui.favoritesFrame);
  ui.favoritesToolBar->setVisible(true);
#ifdef Q_WS_MAC
  ui.actionFull_Screen_Mode->setShortcut
    (QKeySequence(Qt::ControlModifier + Qt::Key_F11));
#endif
  ui.marker->setVisible(false);
  ui.restoreToolButton->setVisible(false);
  sbWidget = new QWidget(this);
  desktopWidget = new ddesktopwidget(this);
  connect(desktopWidget,
	  SIGNAL(viewEntered(void)),
	  this,
	  SLOT(slotViewEntered(void)));
  sb.setupUi(sbWidget);
  sb.progressBar->setVisible(false);
  sb.statusLabel->setVisible(true);
  ui.findFrame->setVisible(false);
  ui.backToolButton->setMenu(new QMenu(this));
  ui.forwardToolButton->setMenu(new QMenu(this));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_historyWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_downloadWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_cookieWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(passphraseWasAuthenticated(const bool)),
	  s_settingsWindow,
	  SLOT(slotPassphraseWasAuthenticated(const bool)));
  connect(s_settingsWindow,
	  SIGNAL(fontChanged(const QFont &)),
	  this,
	  SLOT(slotFontChanged(const QFont &)));
  connect(s_settingsWindow,
	  SIGNAL(iconsChanged(void)),
	  this,
	  SLOT(slotSetIcons(void)));
  connect(s_settingsWindow,
	  SIGNAL(showTabBar(const bool)),
	  this,
	  SLOT(slotSetTabBarVisible(const bool)));
  connect(s_settingsWindow,
	  SIGNAL(showIpAddress(const bool)),
	  this,
	  SLOT(slotShowIpAddress(const bool)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  ui.searchLineEdit,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  ui.locationLineEdit,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  ui.tabWidget,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  m_historySideBar,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_popupsWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_adBlockWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_cookieWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_httpRedirectWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_httpReferrerWindow,
	  SLOT(slotSetIcons(void)));
  connect(this,
	  SIGNAL(iconsChanged(void)),
	  s_javaScriptExceptionsWindow,
	  SLOT(slotSetIcons(void)));
  connect(ui.splitter,
	  SIGNAL(splitterMoved(int, int)),
	  this,
	  SLOT(slotLocationSplitterMoved(int, int)));
  connect(ui.historyAndTabSplitter,
	  SIGNAL(splitterMoved(int, int)),
	  this,
	  SLOT(slotHistoryTabSplitterMoved(int, int)));
  connect(ui.locationLineEdit,
	  SIGNAL(loadPage(const QUrl &)),
	  this,
	  SLOT(slotLoadPage(const QUrl &)));
  connect(desktopWidget,
	  SIGNAL(backgroundImageChanged(void)),
	  this,
	  SLOT(slotChangeDesktopBackgrounds(void)));
  connect(s_downloadWindow, SIGNAL(saveUrl(const QUrl &)), this,
	  SLOT(slotSaveUrl(const QUrl &)));
  connect(m_historySideBar, SIGNAL(open(const QUrl &)),
	  this, SLOT(slotLoadPage(const QUrl &)));
  connect(m_historySideBar, SIGNAL(createTab(const QUrl &)),
	  this, SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(m_historySideBar, SIGNAL(openInNewWindow(const QUrl &)),
	  this, SLOT(slotOpenLinkInNewWindow(const QUrl &)));
  connect(m_historySideBar, SIGNAL(closed(void)),
	  this, SLOT(slotHistorySideBarClosed(void)));
  connect(ui.tabWidget, SIGNAL(currentChanged(int)), this,
	  SLOT(slotTabSelected(int)));
  connect(ui.actionNew_Tab, SIGNAL(triggered(void)), this,
	  SLOT(slotNewTab(void)));
  connect(ui.tabWidget, SIGNAL(createTab(void)), this,
	  SLOT(slotNewTab(void)));
  connect(ui.tabWidget, SIGNAL(openInNewWindow(const int)), this,
	  SLOT(slotOpenPageInNewWindow(const int)));
  connect(ui.actionNew_Window, SIGNAL(triggered(void)), this,
	  SLOT(slotNewWindow(void)));
  connect(ui.actionOpen_URL, SIGNAL(triggered(void)), this,
	  SLOT(slotOpenUrl(void)));
  connect(ui.actionClose_Tab, SIGNAL(triggered(void)), this,
	  SLOT(slotCloseTab(void)));
  connect(ui.tabWidget, SIGNAL(closeTab(const int)), this,
	  SLOT(slotCloseTab(const int)));
  connect(ui.tabWidget, SIGNAL(tabMoved(int, int)), this,
	  SLOT(slotTabMoved(int, int)));
  connect(ui.actionClose_Window, SIGNAL(triggered(void)), this,
	  SLOT(slotClose(void)));
  connect(ui.actionQuit, SIGNAL(triggered(void)), this,
	  SLOT(slotQuit(void)));
  connect(ui.locationLineEdit, SIGNAL(textEdited(const QString &)),
	  this, SLOT(slotTextChanged(const QString &)));
  connect(ui.locationLineEdit, SIGNAL(returnPressed(void)), this,
	  SLOT(slotLoadPage(void)));
  connect(ui.locationLineEdit, SIGNAL(selectionChanged(void)),
	  this, SLOT(slotSelectionChanged(void)));
  connect(ui.locationLineEdit, SIGNAL(openLinkInNewTab(const QUrl &)), this,
	  SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(ui.locationLineEdit, SIGNAL(resetUrl(void)), this,
	  SLOT(slotResetUrl(void)));
  connect(ui.backToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotBack(void)));
  connect(ui.forwardToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotForward(void)));
  connect(ui.reloadToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotReload(void)));
  connect(ui.actionReload, SIGNAL(triggered(void)), this,
	  SLOT(slotReload(void)));
  connect(ui.stopToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotStop(void)));
  connect(ui.actionStop, SIGNAL(triggered(void)), this,
	  SLOT(slotStop(void)));
  connect(ui.homeToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotGoHome(void)));
  connect(ui.searchLineEdit, SIGNAL(returnPressed(void)), this,
	  SLOT(slotSearch(void)));
  connect(ui.searchLineEdit->findButton(), SIGNAL(clicked(void)), this,
	  SLOT(slotSearch(void)));
  connect(ui.searchLineEdit, SIGNAL(selectionChanged(void)),
	  this, SLOT(slotSelectionChanged(void)));
  connect(ui.actionAbout_Dooble, SIGNAL(triggered(void)), this,
	  SLOT(slotAbout(void)));
  connect(ui.actionSave_Page, SIGNAL(triggered(void)), this,
	  SLOT(slotSavePage(void)));
  connect(ui.actionDownloads, SIGNAL(triggered(void)), this,
	  SLOT(slotDisplayDownloadWindow(void)));
  connect(ui.actionFind, SIGNAL(triggered(void)), this,
	  SLOT(slotShowFind(void)));
  connect(ui.actionPrint, SIGNAL(triggered(void)), this,
	  SLOT(slotPrint(void)));
  connect(ui.actionPrint_Preview, SIGNAL(triggered(void)), this,
	  SLOT(slotPrintPreview(void)));
  connect(ui.actionPage_Source, SIGNAL(triggered(void)), this,
	  SLOT(slotShowPageSource(void)));
  connect(ui.actionSettings, SIGNAL(triggered(void)), this,
	  SLOT(slotShowSettingsWindow(void)));
  connect(ui.hideFindToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotHideFind(void)));
  connect(ui.findLineEdit, SIGNAL(returnPressed(void)), this,
	  SLOT(slotFind(void)));
  connect(ui.findLineEdit, SIGNAL(textEdited(const QString &)), this,
	  SLOT(slotFind(void)));
  connect(ui.findLineEdit, SIGNAL(returnPressed(void)), this,
	  SLOT(slotFindNext(void)));
  connect(ui.findLineEdit, SIGNAL(textEdited(const QString &)), this,
	  SLOT(slotFindNext(void)));
  connect(ui.highlightAllCheckBox, SIGNAL(clicked(bool)), this,
	  SLOT(slotFind(void)));
  connect(ui.nextToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotFindNext(void)));
  connect(ui.previousToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotFindPrevious(void)));
  connect(ui.desktopToolButton, SIGNAL(clicked(void)), this,
	  SLOT(slotShowDesktopTab(void)));
  connect(ui.actionStatusbar, SIGNAL(triggered(void)), this,
	  SLOT(slotStatusBarDisplay(void)));
  connect(ui.actionShow_Hidden_Files, SIGNAL(triggered(void)), this,
	  SLOT(slotShowHiddenFiles(void)));
  connect(ui.actionOpen_Directory, SIGNAL(triggered(void)), this,
	  SLOT(slotOpenDirectory(void)));
  connect(ui.actionZoom_In, SIGNAL(triggered(void)), this,
	  SLOT(slotViewZoomIn(void)));
  connect(ui.actionZoom_Out, SIGNAL(triggered(void)), this,
	  SLOT(slotViewZoomOut(void)));
  connect(ui.actionReset_Zoom, SIGNAL(triggered(void)), this,
	  SLOT(slotViewResetZoom(void)));
  connect(ui.actionZoom_Text_Only, SIGNAL(toggled(bool)), this,
	  SLOT(slotViewZoomTextOnly(bool)));
  connect(ui.actionMy_Retrieved_Files, SIGNAL(triggered(void)), this,
	  SLOT(slotOpenMyRetrievedFiles(void)));
  connect(ui.actionP2P_Email, SIGNAL(triggered(void)), this,
	  SLOT(slotOpenP2PEmail(void)));
  connect(ui.actionCopy, SIGNAL(triggered(void)), this,
	  SLOT(slotCopy(void)));
  connect(ui.actionPaste, SIGNAL(triggered(void)), this,
	  SLOT(slotPaste(void)));
  connect(ui.actionSelect_All_Content, SIGNAL(triggered(void)), this,
	  SLOT(slotSelectAllContent(void)));
  connect(ui.actionFull_Screen_Mode, SIGNAL(triggered(void)), this,
	  SLOT(slotFullScreenMode(void)));
  connect(ui.actionApplication_Cookies, SIGNAL(triggered(void)), this,
	  SLOT(slotShowApplicationCookies(void)));
  connect(ui.marker, SIGNAL(markerEntered(void)), this,
	  SLOT(slotMarkerEntered(void)));
  connect(ui.restoreToolButton, SIGNAL(clicked(void)),
	  this, SLOT(slotFullScreenMode(void)));
  connect(ui.actionShow_FavoritesToolBar, SIGNAL(toggled(bool)),
	  this, SLOT(slotShowFavoritesToolBar(bool)));
  connect(ui.actionShow_HistorySideBar, SIGNAL(toggled(bool)),
	  this, SLOT(slotShowHistorySideBar(bool)));
  connect(ui.action_Authenticate, SIGNAL(triggered(void)),
	  this, SLOT(slotAuthenticate(void)));
  connect(ui.actionRefreshPlugins, SIGNAL(triggered()),
          this, SLOT(slotRefreshPlugins()));
  connect(ui.menu_Plugins, SIGNAL (triggered(QAction*)),
          this, SLOT(slotPluginAction(QAction*)));
  connect(ui.tabWidget, SIGNAL(urlsReceivedViaDrop(const QList<QUrl> &)),
	  this, SLOT(slotOpenUrlsFromDrop(const QList<QUrl> &)));
  connect(ui.tabWidget, SIGNAL(openLinkInNewTab(const QUrl &)),
	  this, SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(ui.action_Cookies, SIGNAL(triggered(void)),
	  s_cookiesBlockWindow, SLOT(slotShow(void)));
  connect(ui.action_HTTP_Redirect, SIGNAL(triggered(void)),
	  s_httpRedirectWindow, SLOT(slotShow(void)));
  connect(ui.action_HTTP_Referrer, SIGNAL(triggered(void)),
	  s_httpReferrerWindow, SLOT(slotShow(void)));
  connect(ui.action_JavaScript, SIGNAL(triggered(void)),
	  s_javaScriptExceptionsWindow, SLOT(slotShow(void)));
  connect(ui.action_JavaScript_Popups, SIGNAL(triggered(void)),
	  s_popupsWindow, SLOT(slotShow(void)));
  connect(ui.action_Third_Party_Frame_Content, SIGNAL(triggered(void)),
	  s_adBlockWindow, SLOT(slotShow(void)));
#if QT_VERSION >= 0x040700
  connect(ui.favoritesToolBar,
	  SIGNAL(visibilityChanged(bool)),
	  ui.actionShow_FavoritesToolBar,
	  SLOT(setChecked(bool)));
#endif
  ui.actionZoom_Text_Only->setCheckable(true);
  ui.actionZoom_Text_Only->setChecked(false);
  ui.viewMenu->insertSeparator(ui.zoomMenu->menuAction());

  if(ui.historyMenu->actions().isEmpty())
    {
      QSettings settings
	(s_settings.value("iconSet").toString(), QSettings::IniFormat);

      QAction *action = ui.historyMenu->addAction
	(QIcon(settings.value("mainWindow/historyMenu").toString()),
	 tr("&Clear History"));

      action->setEnabled(false);
      connect(action, SIGNAL(triggered(void)), this,
	      SLOT(slotClearHistory(void)));
      action = ui.historyMenu->addAction
	(QIcon(settings.value("mainWindow/viewHistory").toString()),
	 tr("Show &History"));
      action->setShortcut(QKeySequence(Qt::ControlModifier + Qt::Key_H));
      action->setEnabled(true);
      connect(action, SIGNAL(triggered(void)), this,
	      SLOT(slotShowHistory(void)));
      ui.historyMenu->addSeparator();
      action = ui.historyMenu->addAction("&Recently-Closed Tabs");
      action->setEnabled(false);
    }

  ui.editMenu->setEnabled(false);
  ui.viewMenu->setEnabled(false);
  ui.historyMenu->setEnabled(false);
  ui.actionPrint->setEnabled(false);
  ui.actionPrint_Preview->setEnabled(false);
  ui.actionOpen_URL->setEnabled(false);
  ui.actionShow_Hidden_Files->setEnabled(false);
  ui.actionSave_Page->setEnabled(false);

  for(int i = 0; i < ui.tabWidget->count(); i++)
    if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(i)))
      {
	ui.desktopToolButton->setEnabled(false);
	break;
      }

  /*
  ** According to Qt documentation, resize() followed by move() should
  ** be used to restore a window's geometry. Actual results may vary.
  */

  /*
  ** I don't know what to do about JavaScript windows that do not resize
  ** themselves.
  */

  if(s_settings.contains("mainWindow/geometry"))
    {
      if(!m_isJavaScriptWindow)
	{
	  QRect rect
	    (s_settings.value("mainWindow/geometry").
	     toRect());

	  if(!rect.isValid())
	    rect = QRect(100, 100, 800, 600);

	  if(rect.width() >
	     QApplication::desktop()->screenGeometry().width())
	    rect.setWidth
	      (0.85 * QApplication::desktop()->screenGeometry().width());

	  if(rect.height() >
	     QApplication::desktop()->screenGeometry().height())
	    rect.setWidth
	      (0.85 * QApplication::desktop()->screenGeometry().height());

	  setGeometry(rect);
	}
      else
	{
	  if(m_parentWindow)
	    setGeometry(m_parentWindow->geometry().x(),
			m_parentWindow->geometry().y(),
			800, 600);
	  else
	    resize(800, 600);
	}
    }
  else if(QApplication::desktop()->screenGeometry().isValid())
    {
      if(!m_isJavaScriptWindow)
	resize(0.85 * QApplication::desktop()->screenGeometry().width(),
	       0.85 * QApplication::desktop()->screenGeometry().height());
      else if(m_parentWindow)
	setGeometry(m_parentWindow->geometry().x(),
		    m_parentWindow->geometry().y(),
		    800, 600);
      else
	/*
	** Instead of sizeHint(), use a fixed value. sizeHint() may
	** return an invalid size.
	*/

	resize(800, 600);
    }
  else if(m_parentWindow)
    setGeometry(m_parentWindow->geometry().x(),
		m_parentWindow->geometry().y(),
		800, 600);
  else
    /*
    ** Please read the above comment.
    */

    resize(800, 600);

  if(!m_isJavaScriptWindow)
    restoreState(s_settings["mainWindow/state2"].toByteArray());

  if(s_settings.contains("mainWindow/splitterState"))
    ui.splitter->restoreState
      (s_settings.value("mainWindow/splitterState", "").toByteArray());

  if(s_settings.contains("mainWindow/historyTabSplitterState"))
    {
      if(!ui.historyAndTabSplitter->restoreState
	 (s_settings.value("mainWindow/historyTabSplitterState",
			   "").toByteArray()))
	{
	  ui.historyAndTabSplitter->setStretchFactor(0, 0);
	  ui.historyAndTabSplitter->setStretchFactor(1, 100);
	}
    }
  else
    {
      ui.historyAndTabSplitter->setStretchFactor(0, 0);
      ui.historyAndTabSplitter->setStretchFactor(1, 100);
    }

  statusBar()->addPermanentWidget(sbWidget, 100);
  statusBar()->setStyleSheet("QStatusBar::item {"
			     "border: none; "
			     "}");
  statusBar()->setMaximumHeight(sbWidget->height());
  statusBar()->setVisible
    (s_settings.value("mainWindow/statusbarDisplay", true).toBool());
  ui.actionStatusbar->setChecked(!statusBar()->isHidden());
  ui.actionShow_Hidden_Files->setChecked
    (s_settings.value("mainWindow/showHiddenFiles", true).toBool());
  ui.favoritesToolBar->setVisible
    (s_settings.value("mainWindow/showFavoritesToolBar", false).
     toBool());
  ui.actionShow_FavoritesToolBar->setChecked
    (s_settings.value("mainWindow/showFavoritesToolBar", false).
     toBool());
  ui.action_Authenticate->setEnabled
    (!dmisc::passphraseWasAuthenticated() &&
     !s_settings.value("settingsWindow/hashType",
		       "").toString().isEmpty() &&
     !s_settings.value("settingsWindow/passphraseHash",
		       "").toString().isEmpty() &&
     dmisc::isHashTypeSupported
     (s_settings.value("settingsWindow/hashType").
      toString()));
  slotSetIcons();
#ifdef Q_WS_MAC
  foreach(QToolButton *toolButton, findChildren<QToolButton *> ())
    if(toolButton == ui.backToolButton ||
       toolButton == ui.forwardToolButton)
      toolButton->setStyleSheet
	("QToolButton {border: none; padding-right: 5px}"
	 "QToolButton::menu-button {border: none;}");
    else if(toolButton == ui.stopToolButton ||
	    toolButton == ui.reloadToolButton ||
	    toolButton == ui.homeToolButton ||
	    toolButton == ui.desktopToolButton ||
	    toolButton == ui.restoreToolButton)
      toolButton->setStyleSheet
	("QToolButton {border: none;}"
	 "QToolButton::menu-button {border: none;}");

  int margins[4];

  centralWidget()->layout()->getContentsMargins(&margins[0], &margins[1],
						&margins[2], &margins[3]);
  centralWidget()->layout()->setContentsMargins(margins[0], 10,
						margins[2], margins[3]);
#endif

  QFont font;

  if(!font.fromString(s_settings.value("settingsWindow/globalFont", "").
		      toString()) || font.family().isEmpty())
#ifdef Q_WS_MAC
    font = QFont("Times", 12);
#elif defined Q_WS_WIN
    font = QFont("Arial", 10);
#else
    font = QFont("Serif", 9);
#endif

  font.setWeight(QFont::Normal);
  QApplication::setFont(font);

  foreach(QWidget *widget, findChildren<QWidget *> ())
    if(!qobject_cast<dview *> (widget))
      widget->setFont(font);

  menuBar()->setFont(font);

  foreach(QMenu *menu, menuBar()->findChildren<QMenu *> ())
    {
      menu->setFont(font);

      foreach(QAction *action, menu->actions())
	action->setFont(font);
    }

#ifdef Q_WS_X11
  setWindowRole("browser");
#endif

  slotRefreshPlugins();
}

dooble::dooble(const bool isJavaScriptWindow,
	       QMainWindow *parentWindow):QMainWindow(),
					  dest_url_("")
{
  m_parentWindow = parentWindow;
  init(isJavaScriptWindow);
  copyDooble();
  prepareMostVisited();

  if(!isJavaScriptWindow)
    if(s_settings.value("settingsWindow/displayDesktopCheckBox",
			false).toBool())
      slotShowDesktopTab(false);

  show();
  ui.historyFrame->setVisible
    (ui.actionShow_HistorySideBar->isChecked());
}

dooble::dooble(const QUrl &url):
  QMainWindow(),
  dest_url_(url.toString(QUrl::StripTrailingSlash))
{
  m_parentWindow = 0;
  init(false);
  ui.tabWidget->setVisible(false);

  if(promptForPassphrase())
    {
      s_cookies->populate();
      s_historyWindow->populate();
      s_downloadWindow->populate();
      s_historyModel->populate();
      s_popupsWindow->populate();
      s_adBlockWindow->populate();
      s_cookiesBlockWindow->populate();
      s_httpRedirectWindow->populate();
      s_httpReferrerWindow->populate();
      s_javaScriptExceptionsWindow->populate();
    }

  copyDooble();
  prepareMostVisited();
  ui.tabWidget->setVisible(true);
  newTab(url);

  if(s_settings.value("settingsWindow/displayDesktopCheckBox",
		      false).toBool())
    slotShowDesktopTab(false);

  show();
  ui.tabWidget->update();
  ui.historyFrame->setVisible
    (ui.actionShow_HistorySideBar->isChecked());
}

dooble::dooble(dview *p, const QString &tabTitle):
  QMainWindow(),
  dest_url_(p ? p->url().toString(QUrl::StripTrailingSlash) : "")
{
  /*
  ** This method is called when the user has decided to
  ** change a tab into a window.
  */

  m_parentWindow = 0;
  init(false);
  copyDooble();
  prepareMostVisited();
  newTab(p, tabTitle);

  if(s_settings.value("settingsWindow/displayDesktopCheckBox",
		      false).toBool())
    slotShowDesktopTab(false);

  show();
  ui.tabWidget->update();
  ui.historyFrame->setVisible
    (ui.actionShow_HistorySideBar->isChecked());
}

dooble::dooble(const QHash<QString, QVariant> &hash):
  QMainWindow(),
  dest_url_("")
{
  m_parentWindow = 0;
  init(false);
  ui.tabWidget->setVisible(false);

  if(promptForPassphrase())
    {
      s_cookies->populate();
      s_historyWindow->populate();
      s_downloadWindow->populate();
      s_historyModel->populate();
      s_popupsWindow->populate();
      s_adBlockWindow->populate();
      s_cookiesBlockWindow->populate();
      s_httpRedirectWindow->populate();
      s_httpReferrerWindow->populate();
      s_javaScriptExceptionsWindow->populate();
    }

  copyDooble();
  prepareMostVisited();
  ui.tabWidget->setVisible(true);

  QStringList urls(hash["urls"].toStringList());

  while(!urls.isEmpty())
    {
      QUrl url(QUrl::fromUserInput(urls.takeFirst()));

      if(url.isValid())
	newTab(url);
    }

  if(s_settings.value("settingsWindow/displayDesktopCheckBox",
		      false).toBool())
    slotShowDesktopTab(false);

  if(hash.contains("full-screen-mode"))
    slotFullScreenMode();

  show();
  ui.tabWidget->update();
  ui.historyFrame->setVisible
    (ui.actionShow_HistorySideBar->isChecked());
}

dooble::~dooble()
{
  doobleCounter -= 1;

  while(!m_closedTabs.isEmpty())
    delete m_closedTabs.takeFirst();
}

void dooble::slotSetIcons(void)
{
  QSettings settings(s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);

  ui.backToolButton->setIcon
    (QIcon(settings.value("mainWindow/backToolButton").toString()));
  ui.forwardToolButton->setIcon
    (QIcon(settings.value("mainWindow/forwardToolButton").toString()));
  ui.reloadToolButton->setIcon
    (QIcon(settings.value("mainWindow/reloadToolButton").toString()));
  ui.stopToolButton->setIcon
    (QIcon(settings.value("mainWindow/stopToolButton").toString()));
  ui.homeToolButton->setIcon
    (QIcon(settings.value("mainWindow/homeToolButton").toString()));
  ui.hideFindToolButton->setIcon
    (QIcon(settings.value("mainWindow/hideFindToolButton").toString()));
  ui.actionNew_Tab->setIcon
    (QIcon(settings.value("mainWindow/actionNew_Tab").toString()));
  ui.actionNew_Window->setIcon
    (QIcon(settings.value("mainWindow/actionNew_Window").toString()));
  ui.actionOpen_URL->setIcon
    (QIcon(settings.value("mainWindow/actionOpen_URL").toString()));
  ui.actionClose_Tab->setIcon
    (QIcon(settings.value("mainWindow/actionClose_Tab").toString()));
  ui.actionClose_Window->setIcon
    (QIcon(settings.value("mainWindow/actionClose_Window").toString()));
  ui.actionSave_Page->setIcon
    (QIcon(settings.value("mainWindow/actionSave_Page").toString()));
  ui.actionPrint->setIcon
    (QIcon(settings.value("mainWindow/actionPrint").toString()));
  ui.actionPrint_Preview->setIcon
    (QIcon(settings.value("mainWindow/actionPrint_Preview").toString()));
  ui.actionQuit->setIcon
    (QIcon(settings.value("mainWindow/actionQuit").toString()));
  ui.actionFind->setIcon
    (QIcon(settings.value("mainWindow/actionFind").toString()));
  ui.actionReload->setIcon
    (QIcon(settings.value("mainWindow/actionReload").toString()));
  ui.actionStop->setIcon
    (QIcon(settings.value("mainWindow/actionStop").toString()));
  ui.actionPage_Source->setIcon
    (QIcon(settings.value("mainWindow/actionPage_Source").toString()));
  ui.actionDownloads->
    setIcon(QIcon(settings.value("mainWindow/actionDownloads").toString()));
  ui.actionSettings->setIcon
    (QIcon(settings.value("mainWindow/actionSettings").toString()));
  ui.actionAbout_Dooble->
    setIcon(QIcon(settings.value("mainWindow/actionAbout_Dooble").toString()));
  ui.desktopToolButton->
    setIcon(QIcon(settings.value("mainWindow/desktopToolButton").toString()));
  ui.actionOpen_Directory->setIcon
    (QIcon(settings.value("mainWindow/actionOpen_Directory").toString()));
  ui.actionMy_Retrieved_Files->setIcon
    (QIcon(settings.value("mainWindow/actionMy_Retrieved_Files").toString()));
  ui.restoreToolButton->setIcon
    (QIcon(settings.value("mainWindow/restoreToolButton").toString()));
  ui.actionCopy->setIcon
    (QIcon(settings.value("mainWindow/actionCopy").toString()));
  ui.actionPaste->setIcon
    (QIcon(settings.value("mainWindow/actionPaste").toString()));
  ui.actionSelect_All_Content->setIcon
    (QIcon(settings.value("mainWindow/actionSelect_All_Content").toString()));
  ui.zoomMenu->setIcon
    (QIcon(settings.value("mainWindow/zoomMenu").toString()));
  ui.actionFull_Screen_Mode->setIcon
    (QIcon(settings.value("mainWindow/actionFull_Screen_Mode").toString()));
  ui.nextToolButton->setIcon
    (QIcon(settings.value("mainWindow/nextToolButton").toString()));
  ui.previousToolButton->setIcon
    (QIcon(settings.value("mainWindow/previousToolButton").toString()));
  ui.actionApplication_Cookies->setIcon
    (QIcon(settings.value
	   ("mainWindow/actionApplication_Cookies").toString()));
  ui.actionP2P_Email->setIcon
    (QIcon(settings.value("mainWindow/actionP2P_Email").toString()));
  ui.action_Authenticate->setIcon
    (QIcon(settings.value("mainWindow/authenticate_Action").toString()));
  ui.actionRefreshPlugins->setIcon
    (QIcon(settings.value("mainWindow/actionRefreshPlugins").toString()));

  if(ui.historyMenu->actions().size() > 0)
    ui.historyMenu->actions()[0]->setIcon
      (QIcon(settings.value("mainWindow/historyMenu").toString()));

  if(ui.historyMenu->actions().size() > 1)
    ui.historyMenu->actions()[1]->setIcon
      (QIcon(settings.value("mainWindow/viewHistory").toString()));

  QMessageBox *mb = findChild<QMessageBox *> ("about");

  if(mb)
    {
      for(int i = 0; i < mb->buttons().size(); i++)
	if(mb->buttonRole(mb->buttons().at(i)) == QMessageBox::AcceptRole)
	  mb->buttons().at(i)->setIcon
	    (QIcon(settings.value("okButtonIcon").toString()));

      mb->setWindowIcon
	(QIcon(settings.value("mainWindow/windowIcon").toString()));
    }

  /*
  ** The 0th tab is not necessarily the Desktop.
  */

  for(int i = 0; i < ui.tabWidget->count(); i++)
    if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(i)))
      {
	ui.tabWidget->setTabIcon
	  (i, QIcon(settings.value("mainWindow/tabWidget").toString()));
	break;
      }

  setWindowIcon(QIcon(settings.value("mainWindow/windowIcon").toString()));
  emit iconsChanged();
}

void dooble::newTabInit(dview *p, const QString &tabTitle)
{
  if(!p)
    return;

  connect(p, SIGNAL(destroyed(QObject *)),
	  this, SLOT(slotObjectDestroyed(QObject *)));
  connect(p, SIGNAL(urlChanged(const QUrl &)), this,
	  SLOT(slotUrlChanged(const QUrl &)));
  connect(p, SIGNAL(titleChanged(const QString &)), this,
	  SLOT(slotTitleChanged(const QString &)));
  connect(p, SIGNAL(loadFinished(bool)), this,
	  SLOT(slotLoadFinished(bool)));
  connect(p, SIGNAL(loadProgress(int)), this,
	  SLOT(slotLoadProgress(int)));
  connect(p, SIGNAL(iconChanged(void)), this,
	  SLOT(slotIconChanged(void)));
  connect(p, SIGNAL(loadStarted(void)), this,
	  SLOT(slotLoadStarted(void)));
  connect(p, SIGNAL(openLinkInNewTab(const QUrl &)), this,
	  SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(p, SIGNAL(openLinkInNewWindow(const QUrl &)), this,
	  SLOT(slotOpenLinkInNewWindow(const QUrl &)));
  connect(p, SIGNAL(copyLink(const QUrl &)), this,
	  SLOT(slotCopyLink(const QUrl &)));
  connect(p, SIGNAL(saveUrl(const QUrl &)), this,
	  SLOT(slotSaveUrl(const QUrl &)));
  connect(p, SIGNAL(saveFile(const QString &, const QUrl &)), this,
	  SLOT(slotSaveFile(const QString &, const QUrl &)));
  connect(p, SIGNAL(viewImage(const QUrl &)), this,
	  SLOT(slotLoadPage(const QUrl &)));
  connect(p, SIGNAL(ipAddressChanged(const QString &)), this,
	  SLOT(slotIpAddressChanged(const QString &)));
  connect(p->page(),
	  SIGNAL(linkHovered(const QString &, const QString &,
			     const QString &)),
	  this,
	  SLOT(slotLinkHovered(const QString &, const QString &,
			       const QString &)));
  connect(p->page()->networkAccessManager(),
	  SIGNAL(authenticationRequired(QNetworkReply *, QAuthenticator *)),
	  this,
	  SLOT(slotAuthenticationRequired(QNetworkReply *, QAuthenticator *)));
  connect(p->page()->networkAccessManager(),
	  SIGNAL(proxyAuthenticationRequired(const QNetworkProxy &,
					     QAuthenticator *)),
	  this,
	  SLOT(slotProxyAuthenticationRequired(const QNetworkProxy &,
					       QAuthenticator *)));
  connect(p->page(), SIGNAL(windowCloseRequested(void)),
	  this, SLOT(slotCloseWindow(void)));
  connect(p->page(), SIGNAL(menuBarVisibilityChangeRequested(bool)),
	  this, SLOT(slotHideMenuBar(void)));
  connect(p->page(), SIGNAL(statusBarVisibilityChangeRequested(bool)),
	  this, SLOT(slotHideStatusBar(void)));
  connect(p->page(), SIGNAL(toolBarVisibilityChangeRequested(bool)),
	  this, SLOT(slotHideToolBar(void)));
  connect(p->page(), SIGNAL(printRequested(QWebFrame *)),
	  this, SLOT(slotPrintRequested(QWebFrame *)));
  connect(p->page(), SIGNAL(geometryChangeRequested(const QRect &)),
	  this, SLOT(slotGeometryChangeRequested(const QRect &)));
  connect(p->page(), SIGNAL(repaintRequested(const QRect &)),
	  this, SLOT(slotRepaintRequested(const QRect &)));
  connect(p->page(), SIGNAL(statusBarMessage(const QString &)),
	  this, SLOT(slotStatusBarMessage(const QString &)));
  connect(this,
	  SIGNAL(clearHistory(void)),
	  p,
	  SLOT(slotClearHistory(void)));
  connect(p, SIGNAL(viewEntered(void)), this,
	  SLOT(slotViewEntered(void)));
  connect(p, SIGNAL(selectionChanged(const QString &)),
	  this, SLOT(slotSelectionChanged(const QString &)));
  connect(p, SIGNAL(viewPageSource(void)),
	  this, SLOT(slotShowPageSource(void)));
  connect(p, SIGNAL(urlsReceivedViaDrop(const QList<QUrl> &)),
	  this, SLOT(slotOpenUrlsFromDrop(const QList<QUrl> &)));
  connect(p, SIGNAL(goBack(void)),
	  this, SLOT(slotBack(void)));
  connect(p, SIGNAL(goReload(void)),
	  this, SLOT(slotReload(void)));
  connect(p, SIGNAL(goForward(void)),
	  this, SLOT(slotForward(void)));

  int index = 0;

  if(s_settings.value("settingsWindow/appendNewTabs",
		      false).toBool())
    {
      ui.tabWidget->addTab(p, tabTitle);
      index = ui.tabWidget->count() - 1;
    }
  else
    {
      index = ui.tabWidget->currentIndex() + 1;
      ui.tabWidget->insertTab(index, p, tabTitle);
    }

  ui.tabWidget->setTabButton(index);

  /*
  ** If p came from another window, its animation icon will
  ** appear distorted because of the new tab
  ** having a "selected" background color (stylesheet).
  */

  ui.tabWidget->animateIndex(index, !p->isLoaded(), p->icon());
  sb.statusLabel->clear();
  ui.tabWidget->setTabsClosable(ui.tabWidget->count() > 1);

  if(ui.tabWidget->count() == 1)
    ui.tabWidget->setBarVisible
      (s_settings.value("settingsWindow/alwaysShowTabBar",
			true).toBool());
  else
    ui.tabWidget->setBarVisible(true);
}

dview *dooble::newTab(dview *p, const QString &tabTitle)
{
  if(p)
    {
      p->setParent(this);

      QString title(tabTitle);

      if(title.isEmpty())
	title = p->url().toString(QUrl::StripTrailingSlash);

      if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	title = title.mid
	  (0,
	   MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() + "...";

      if(title.contains("&"))
	title = title.replace("&", "&&");

      if(title.isEmpty())
	title = tr("Dooble Web Browser");

      newTabInit(p, title);

      if(p->url().isEmpty() || !p->url().isValid())
	slotOpenUrl();

      QAction *action = new QAction(p->icon(), title, this);

      action->setData(p->url());
      connect(action, SIGNAL(triggered(void)), this,
	      SLOT(slotLinkActionTriggered(void)));
      p->setTabAction(action);
      prepareTabsMenu();
    }

  return p;
}

dview *dooble::newTab(const QUrl &url)
{
  dview *p = new dview(this);

  newTabInit(p, tr("(Untitled)"));

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      ui.locationLineEdit->setText
	(url.toString(QUrl::StripTrailingSlash));

      if(url.isEmpty() || !url.isValid())
	slotOpenUrl();
    }

  QAction *action = new QAction(p->icon(), tr("(Untitled)"), this);

  action->setData(url);
  connect(action, SIGNAL(triggered(void)), this,
	  SLOT(slotLinkActionTriggered(void)));
  p->setTabAction(action);
  p->load(url);
  prepareTabsMenu();
  prepareWidgetsBasedOnView(p);
  return p;
}

dview *dooble::newTab(dwebhistory *history)
{
  QUrl url;
  QIcon icon;
  dview *p = new dview(this, history);

  newTabInit(p, tr("(Untitled)"));

  if(history)
    {
      url = history->currentItem().url();
      icon = history->currentItem().icon();
    }

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      ui.locationLineEdit->setText
	(url.toString(QUrl::StripTrailingSlash));

      if(url.isEmpty() || !url.isValid())
	slotOpenUrl();
    }

  QAction *action = new QAction(icon, tr("(Untitled)"), this);

  action->setData(url);
  connect(action, SIGNAL(triggered(void)), this,
	  SLOT(slotLinkActionTriggered(void)));
  p->setTabAction(action);
  p->load(url);
  prepareTabsMenu();
  prepareWidgetsBasedOnView(p);
  return p;
}

void dooble::slotQuit(void)
{
  /*
  ** closeEvent() may call this slot.
  */

  if(sender() != 0)
    {
      if(!warnAboutDownloads())
	return;

      if(!warnAboutTabs(QApplication::instance()))
	return;
    }

  s_historyWindow->deleteLater();
  s_downloadWindow->deleteLater();
  s_cookies->deleteLater();
  s_cookieWindow->deleteLater();
  s_historyModel->deleteLater();
  s_popupsWindow->deleteLater();
  s_adBlockWindow->deleteLater();
  s_cookiesBlockWindow->deleteLater();
  s_httpRedirectWindow->deleteLater();
  s_httpReferrerWindow->deleteLater();
  s_javaScriptExceptionsWindow->deleteLater();
  saveSettings();
  QApplication::instance()->exit(0);
}

void dooble::saveSettings(void)
{
  if(!m_isJavaScriptWindow && !isFullScreen())
    {
      QSettings settings;

      settings.setValue("mainWindow/state2", saveState());
      settings.setValue("mainWindow/geometry", geometry());
      settings.setValue
	("mainWindow/ftpManagerColumnsState1",
	 s_settings.value
	 ("mainWindow/ftpManagerColumnsState1").toByteArray());
      settings.setValue
	("mainWindow/fileManagerColumnsState1",
	 s_settings.value
	 ("mainWindow/fileManagerColumnsState1").toByteArray());
      s_settings["mainWindow/state2"] = saveState();
      s_settings["mainWindow/geometry"] = geometry();
    }
}

void dooble::loadPage(const QUrl &url)
{
  if(url.isEmpty() || !url.isValid())
    {
      /*
      ** Prevent the user from loading an invalid URL.
      */

      slotOpenUrl();
      return;
    }

  sb.statusLabel->clear();

  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      QString str(url.toString(QUrl::StripTrailingSlash).trimmed());

      ui.locationLineEdit->setText(str);
      sb.progressBar->setVisible(true);
      ui.reloadStopWidget->setCurrentIndex(0);
      ui.stopToolButton->setEnabled(true);
      ui.actionStop->setEnabled(true);
      ui.tabWidget->setTabText(ui.tabWidget->indexOf(p), tr("Loading..."));
      ui.tabWidget->setTabToolTip(ui.tabWidget->indexOf(p), tr("Loading..."));
      p->load(url);
      ui.zoomMenu->setEnabled(!p->isDir() && !p->isFtp());
      ui.actionSelect_All_Content->setEnabled
	(!p->isDir() && !p->isFtp());
      ui.locationLineEdit->setIcon(p->icon());
      ui.actionFind->setEnabled(!p->isDir() && !p->isFtp());

      if(p->isDir() || p->isFtp())
	ui.findFrame->setVisible(false);
      else
	ui.findFrame->setVisible(showFindFrame);
    }
  else
    {
      newTab(url);
      ui.tabWidget->update();
    }
}

void dooble::slotLoadPage(void)
{
  loadPage(QUrl::fromUserInput(ui.locationLineEdit->text().trimmed()));
  ui.tabWidget->currentWidget()->setFocus();
}

void dooble::slotLoadPage(const QUrl &url)
{
  loadPage(url);
  ui.tabWidget->currentWidget()->setFocus();
}

void dooble::slotNewTab(void)
{
  QUrl url;
  dview *p = 0;
  QLocale locale;

  url = QUrl::fromUserInput
    (s_settings.value
     ("settingsWindow/url12_1",
      QString("qrc:/search_%1.html").arg(locale.name().left(2))).toString());

  if(!url.isValid())
    url = QUrl();

  p = newTab(url);

  if(p)
    {
      ui.tabWidget->setCurrentWidget(p);
      ui.tabWidget->update();

      if(url.isEmpty() || !url.isValid())
	/*
	** p's url may be empty at this point.
	*/

	slotOpenUrl();
    }
}

void dooble::slotNewWindow(void)
{
  QUrl url;
  QLocale locale;

  url = QUrl::fromUserInput
    (s_settings.value
     ("settingsWindow/url12_1",
      QString("qrc:/search_%1.html").arg(locale.name().left(2))).toString());

  if(!url.isValid())
    url = QUrl();

  if(s_settings.value("settingsWindow/openUserWindowsInNewProcesses",
		      false).toBool())
    launchDooble(url);
  else
    {
      Q_UNUSED(new dooble(url));
    }
}

void dooble::slotClose(void)
{
  close();
}

bool dooble::warnAboutTabs(QObject *object)
{
  /*
  ** Returns true if the warning was accepted or
  ** if the setting is disabled.
  */

  if(s_settings.value("settingsWindow/warnBeforeClosingModifiedTab", false).
     toBool())
    {
      bool ask = false;

      if(qobject_cast<QApplication *> (object))
	{
	  foreach(QWidget *widget, QApplication::allWidgets())
	    {
	      dview *p = qobject_cast<dview *> (widget);

	      if(p && p->isModified())
		{
		  ask = true;
		  break;
		}
	    }
	}
      else
	{
	  dooble *d = qobject_cast<dooble *> (object);

	  for(int i = 0; i < d->ui.tabWidget->count(); i++)
	    {
	      dview *p = qobject_cast<dview *> (d->ui.tabWidget->widget(i));

	      if(p && p->isModified())
		{
		  ask = true;
		  break;
		}
	    }
	}

      if(ask)
	{
	  QMessageBox mb(this);

	  mb.setIcon(QMessageBox::Question);
	  mb.setWindowTitle(tr("Dooble Web Browser: Confirmation"));
	  mb.setWindowModality(Qt::ApplicationModal);
	  mb.setStandardButtons(QMessageBox::No | QMessageBox::Yes);

	  if(doobleCounter <= 1 || qobject_cast<QApplication *> (object))
	    mb.setText(tr("You have tabs with modified content. "
			  "Are you sure that you wish to exit?"));
	  else
	    mb.setText(tr("You have tabs with modified content. "
			  "Are you sure that you wish to close?"));

	  QSettings settings(s_settings.value("iconSet").toString(),
			     QSettings::IniFormat);

	  for(int i = 0; i < mb.buttons().size(); i++)
	    if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
	      mb.buttons().at(i)->setIcon
		(QIcon(settings.value("okButtonIcon").toString()));
	    else
	      mb.buttons().at(i)->setIcon
		(QIcon(settings.value("cancelButtonIcon").toString()));

	  mb.setWindowIcon
	    (QIcon(settings.value("mainWindow/windowIcon").toString()));

	  if(mb.exec() == QMessageBox::No)
	    return false;
	}
    }

  return true;
}

bool dooble::warnAboutDownloads(void)
{
  /*
  ** Returns true if the warning was accepted.
  */

  if(s_downloadWindow && s_downloadWindow->isActive())
    {
      QMessageBox mb(this);

      mb.setIcon(QMessageBox::Question);
      mb.setWindowTitle(tr("Dooble Web Browser: Confirmation"));
      mb.setWindowModality(Qt::ApplicationModal);
      mb.setStandardButtons(QMessageBox::No | QMessageBox::Yes);
      mb.setText(tr("Exiting Dooble "
		    "will cause existing downloads to be interrupted. "
		    "Are you sure that you wish to continue?"));

      QSettings settings(s_settings.value("iconSet").toString(),
			 QSettings::IniFormat);

      for(int i = 0; i < mb.buttons().size(); i++)
	if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("okButtonIcon").toString()));
	else
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("cancelButtonIcon").toString()));

      mb.setWindowIcon
	(QIcon(settings.value("mainWindow/windowIcon").toString()));

      if(mb.exec() == QMessageBox::No)
	return false;
    }

  return true;
}

void dooble::closeEvent(QCloseEvent *event)
{
  if(doobleCounter <= 1 && !warnAboutDownloads())
    {
      if(event)
	event->ignore();

      return;
    }

  if(!warnAboutTabs(this))
    {
      if(event)
	event->ignore();

      return;
    }

  unsetUrlHandler();
  deleteLater();
  QMainWindow::closeEvent(event);

  if(doobleCounter <= 1)
    slotQuit();
}

void dooble::slotStop(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    p->stop();

  sb.progressBar->setVisible(false);
  ui.reloadStopWidget->setCurrentIndex(1);
  ui.stopToolButton->setEnabled(false);
  ui.actionStop->setEnabled(false);
}

void dooble::slotReload(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      if(p->isModified() && !warnAboutLeavingModifiedTab())
	return;

      sb.progressBar->setVisible(true);
      ui.reloadStopWidget->setCurrentIndex(0);
      ui.stopToolButton->setEnabled(true);
      ui.actionStop->setEnabled(true);
      p->reload();
    }
}

void dooble::slotBack(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      if(p->isModified() && !warnAboutLeavingModifiedTab())
	return;

      ui.forwardToolButton->setEnabled(p->canGoForward());
      sb.progressBar->setVisible(true);
      ui.reloadStopWidget->setCurrentIndex(0);
      ui.stopToolButton->setEnabled(true);
      ui.actionStop->setEnabled(true);
      p->back();
    }
}

void dooble::slotForward(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      if(p->isModified() && !warnAboutLeavingModifiedTab())
	return;

      ui.backToolButton->setEnabled(p->canGoBack());
      sb.progressBar->setVisible(true);
      ui.reloadStopWidget->setCurrentIndex(0);
      ui.stopToolButton->setEnabled(true);
      ui.actionStop->setEnabled(true);
      p->forward();
    }
}

void dooble::slotTabSelected(const int index)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->widget(index));

  if(p)
    {
      ui.homeToolButton->setEnabled(true);
      ui.searchLineEdit->setVisible(!m_isJavaScriptWindow);
      ui.locationLineEdit->setVisible(true);
      ui.reloadToolButton->setEnabled(true);
      ui.searchLineEdit->setEnabled(true);
      ui.editMenu->setEnabled(true);
      ui.viewMenu->setEnabled(true);
      ui.historyMenu->setEnabled(true);
      ui.actionPrint->setEnabled(true);
      ui.actionPrint_Preview->setEnabled(true);
      ui.actionOpen_URL->setEnabled(true);
      ui.actionSave_Page->setEnabled(true);

      if(p->title().isEmpty())
	{
	  if(p->ipAddress().isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(tr("Dooble Web Browser"));
	  else
	    setWindowTitle(QString(tr("Dooble Web Browser (%1)")).
			   arg(p->ipAddress()));
	}
      else
	{
	  if(p->ipAddress().isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(p->title() + tr(" - Dooble Web Browser"));
	  else
            setWindowTitle(p->title() + QString(" (%1)").
			   arg(p->ipAddress()) + tr(" - Dooble Web Browser"));
	}

      sb.progressBar->setVisible(!p->isLoaded());
      sb.progressBar->setValue(p->progress());
      ui.reloadStopWidget->setCurrentIndex(p->isLoaded());
      ui.stopToolButton->setEnabled(!p->isLoaded());
      ui.actionStop->setEnabled(!p->isLoaded());
      ui.backToolButton->setEnabled(p->canGoBack());
      ui.forwardToolButton->setEnabled(p->canGoForward());
      ui.locationLineEdit->setIcon(p->icon());
      ui.locationLineEdit->setText
	(p->url().toString(QUrl::StripTrailingSlash));
      ui.locationLineEdit->setSecureColor(p->hasSecureSonnection());
      prepareNavigationButtonMenus(p);
      sb.statusLabel->clear();
      statusBar()->setVisible(ui.actionStatusbar->isChecked());
      prepareWidgetsBasedOnView(p);

      if(p->url().isEmpty() || !p->url().isValid())
	slotOpenUrl();
      else
	p->setFocus();

      p->update();
    }
  else
    {
      setWindowTitle(tr("Dooble Web Browser"));
      ui.forwardToolButton->setEnabled(false);
      ui.backToolButton->setEnabled(false);
      ui.stopToolButton->setEnabled(false);
      ui.reloadToolButton->setEnabled(false);
      ui.homeToolButton->setEnabled(false);
      ui.searchLineEdit->setVisible(false);
      ui.locationLineEdit->setVisible(false);
      ui.editMenu->setEnabled(false);
      ui.viewMenu->setEnabled(false);
      ui.historyMenu->setEnabled(false);
      ui.actionPrint->setEnabled(false);
      ui.actionPrint_Preview->setEnabled(false);
      ui.actionOpen_URL->setEnabled(false);
      ui.actionSave_Page->setEnabled(false);
      ui.searchLineEdit->clear();
      ui.findFrame->setVisible(false);
      sb.statusLabel->clear();
      sb.progressBar->setValue(0);
      sb.progressBar->setVisible(false);
      statusBar()->setVisible(false);

      if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(index)))
	desktopWidget->setFocus();
    }
}

void dooble::slotTextChanged(const QString &text)
{
  dest_url_ = text;

  QIcon icon;
  QSettings settings
    (s_settings.value("iconSet").toString(), QSettings::IniFormat);

  icon = QIcon(settings.value("urlWidget/emptyIcon").toString());
  ui.locationLineEdit->setIcon(icon);
}

void dooble::slotUrlChanged(const QUrl &url)
{
  if(url.isEmpty() || !url.isValid())
    return;

  QString urlText(url.toString(QUrl::StripTrailingSlash));

  ui.locationLineEdit->addItem(urlText);

  dview *p = qobject_cast<dview *> (sender());

  prepareWidgetsBasedOnView(p);

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      /*
      ** Set the URL widget's icon to the empty icon whenever
      ** the URL for the current Web view has changed. When the
      ** URL changes, dwebview will attempt to download the favicon
      ** associated with the URL.
      */

      ui.locationLineEdit->setIcon(dmisc::iconForUrl(QUrl()));
      ui.backToolButton->setEnabled(p->canGoBack());
      ui.forwardToolButton->setEnabled(p->canGoForward());
    }

  /*
  ** The URL widget's text should not be changed if the
  ** user has edited its contents. However, how will
  ** the widget reflect the current URL?
  */

  if(!ui.locationLineEdit->hasFocus() ||
     !ui.locationLineEdit->isModified())
    if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
      ui.locationLineEdit->setText(urlText);

  if(p && p->tabAction())
    p->tabAction()->setData(url);
}

void dooble::slotTitleChanged(const QString &titleArg)
{
  dview *p = qobject_cast<dview *> (sender());
  QString title(titleArg.trimmed());

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      if(title.isEmpty())
	{
	  if(p->ipAddress().isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(tr("Dooble Web Browser"));
	  else
	    setWindowTitle(QString(tr("Dooble Web Browser (%1)")).
			   arg(p->ipAddress()));
	}
      else
	{
	  if(p->ipAddress().isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(title + tr(" - Dooble Web Browser"));
	  else
	    setWindowTitle(title + QString(" (%1)").arg(p->ipAddress()) +
			   tr(" - Dooble Web Browser"));
	}
    }

  if(p)
    {
      if(title.isEmpty())
	/*
	** The tab's title will be set to the URL.
	*/

	title = p->url().toString(QUrl::StripTrailingSlash);

      ui.tabWidget->setTabText(ui.tabWidget->indexOf(p),
			       title.replace("&", "&&"));
      ui.tabWidget->setTabToolTip(ui.tabWidget->indexOf(p),
				  title);

      if(p->tabAction())
	{
	  title = titleArg.trimmed();

	  if(title.isEmpty())
	    title = p->url().toString(QUrl::StripTrailingSlash);

	  if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	    title = title.mid
	      (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
	      "...";

	  if(title.contains("&"))
	    title = title.replace("&", "&&");

	  if(title.isEmpty())
	    title = tr("Dooble Web Browser");

	  p->tabAction()->setText(title);
	}
    }
}

void dooble::slotIconChanged(void)
{
  QIcon icon;
  dview *p = qobject_cast<dview *> (sender());

  if(p)
    icon = p->icon();

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    ui.locationLineEdit->setIcon(icon);

  if(p)
    {
      int index = ui.locationLineEdit->findText
	(p->url().toString(QUrl::StripTrailingSlash));

      if(index > -1)
	ui.locationLineEdit->setItemIcon(index, icon);

      if(p->tabAction())
	p->tabAction()->setIcon(icon);

      QList<QAction *> actions = ui.historyMenu->actions();

      for(int i = 5; i < actions.size(); i++)
	if(actions.at(i)->data().toString() ==
	   p->url().toString(QUrl::StripTrailingSlash))
	  {
	    actions.at(i)->setIcon(icon);
	    break;
	  }

      foreach(QToolButton *toolButton,
	      ui.favoritesToolButtonsFrame->findChildren<QToolButton *> ())
	if(toolButton->property("url").toUrl().
	   toString(QUrl::StripTrailingSlash) ==
	   p->url().toString(QUrl::StripTrailingSlash))
	  {
	    toolButton->setIcon(icon);
	    break;
	  }
    }
}

void dooble::slotLoadProgress(int progress)
{
  dview *p = qobject_cast<dview *> (sender());

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      if(progress < 0)
	{
	  sb.progressBar->setMaximum(0);
	  sb.progressBar->setMinimum(0);
	}
      else
	{
	  sb.progressBar->setMaximum(100);
	  sb.progressBar->setValue(progress >= 0 ? progress : 0);
	}
    }
}

void dooble::slotLoadFinished(bool ok)
{
  dview *p = qobject_cast<dview *> (sender());

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      sb.progressBar->setVisible(false);
      ui.reloadStopWidget->setCurrentIndex(1);
      ui.stopToolButton->setEnabled(false);
      ui.actionStop->setEnabled(false);
      ui.backToolButton->setEnabled(p->canGoBack());
      ui.forwardToolButton->setEnabled(p->canGoForward());
      ui.locationLineEdit->setSecureColor(p->hasSecureSonnection());
    }

  if(!ok)
    {
      if(p && ui.tabWidget->indexOf(p) > -1)
	{
	  int index = ui.tabWidget->indexOf(p);

	  if(ui.tabWidget->tabText(index) == tr("Loading..."))
	    {
	      ui.tabWidget->setTabText
		(ui.tabWidget->indexOf(p), tr("(Untitled)"));
	      ui.tabWidget->setTabToolTip
		(ui.tabWidget->indexOf(p), tr("(Untitled)"));
	    }

	  ui.tabWidget->animateIndex(index, false, p->icon());
	}

      return;
    }

  QIcon icon;

  if(p)
    {
      icon = p->icon();

      int index = ui.locationLineEdit->findText
	(p->url().toString(QUrl::StripTrailingSlash));

      if(index > -1)
	ui.locationLineEdit->setItemIcon(index, icon);

      ui.tabWidget->animateIndex(ui.tabWidget->indexOf(p), false, icon);
      saveHistoryItem(p->currentItem());

      if(!p->url().host().isEmpty())
	{
	  s_mostVisitedHosts[p->url().host()] += 1;
	  prepareMostVisited();
	}
    }

  /*
  ** Add items to the history menu.
  */

  if(s_settings.value("settingsWindow/rememberHistory",
		      true).toBool() && p)
    {
      bool actionExists = false;
      QList<QAction *> actions = ui.historyMenu->actions();

      for(int i = 5; i < actions.size(); i++)
	if(actions.at(i)->data().toString() ==
	   p->url().toString(QUrl::StripTrailingSlash))
	  {
	    actions.at(i)->setIcon(icon);
	    actionExists = true;
	    break;
	  }

      actions.clear();

      if(!actionExists)
	{
	  if(ui.historyMenu->actions().size() == 4)
	    {
	      ui.historyMenu->addSeparator();
	      ui.historyMenu->actions()[0]->setEnabled
		(ui.historyMenu->isEnabled());
	      ui.historyMenu->actions()[1]->setEnabled(true);
	    }

	  QString title(p->title());

	  if(title.isEmpty())
	    title = p->url().toString(QUrl::StripTrailingSlash);

	  if(title.contains("&"))
	    title = title.replace("&", "&&");

	  if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	    title = title.mid
	      (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
	      "...";

	  if(title.isEmpty())
	    title = tr("Dooble Web Browser");

	  /*
	  ** 5 = Clear History, Show History, a separator,
	  **     Recently-Closed Tabs, and a separator.
	  */

	  QAction *action = 0;

	  if(ui.historyMenu->actions().size() >= 5 + MAX_HISTORY_ITEMS)
	    {
	      action = ui.historyMenu->actions().at(5);

	      for(int i = ui.historyMenu->actions().size() - 1; i >= 5; i--)
		{
		  ui.historyMenu->actions().at(i)->setData
		    (ui.historyMenu->actions().at(i - 1)->data());
		  ui.historyMenu->actions().at(i)->setIcon
		    (ui.historyMenu->actions().at(i - 1)->icon());
		  ui.historyMenu->actions().at(i)->setText
		    (ui.historyMenu->actions().at(i - 1)->text());
		}
	    }
	  else
	    {
	      action = new QAction(this);
	      connect(action, SIGNAL(triggered(void)), this,
		      SLOT(slotLinkActionTriggered(void)));

	      if(ui.historyMenu->actions().size() == 5)
		ui.historyMenu->addAction(action);
	      else
		ui.historyMenu->insertAction
		  (ui.historyMenu->actions().at(5), action);
	    }

	  if(action)
	    {
	      action->setEnabled(ui.historyMenu->isEnabled());
	      action->setData
		(QVariant(p->url().toString(QUrl::StripTrailingSlash)));
	      action->setIcon(icon);
	      action->setText(title);
	    }
	}
    }

  prepareNavigationButtonMenus(p);

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      if(!ui.locationLineEdit->hasFocus() &&
	 !ui.locationLineEdit->isModified() &&
	 !p->url().isEmpty() && p->url().isValid())
	p->setFocus();
    }
}

void dooble::prepareNavigationButtonMenus(dview *p)
{
  if(!p)
    return;

  if(!s_settings.value("settingsWindow/rememberHistory",
		       true).toBool())
    return;

  while(!ui.backToolButton->menu()->actions().isEmpty())
    {
      QAction *action = ui.backToolButton->menu()->actions().first();

      ui.backToolButton->menu()->removeAction(action);
      action->deleteLater();
    }

  QList<dwebhistoryitem> list(p->backItems(MAX_HISTORY_ITEMS));

  for(int i = 0; i < list.size(); i++)
    {
      QString title(list.at(i).title());

      if(title.isEmpty())
	title = list.at(i).url().toString(QUrl::StripTrailingSlash);

      if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	title = title.mid
	  (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
	  "...";

      if(title.contains("&"))
	title = title.replace("&", "&&");

      if(title.isEmpty())
	title = tr("Dooble Web Browser");

      QIcon icon(list.at(i).icon());

      if(icon.isNull())
	icon = dmisc::iconForUrl(list.at(i).url());

      QAction *action = ui.backToolButton->menu()->addAction
	(icon, title, this, SLOT(slotLinkActionTriggered(void)));

      if(action)
	action->setData
	  (QVariant(list.at(i).url().toString(QUrl::StripTrailingSlash)));
    }

  list.clear();

  while(!ui.forwardToolButton->menu()->actions().isEmpty())
    {
      QAction *action = ui.forwardToolButton->menu()->actions().first();

      ui.forwardToolButton->menu()->removeAction(action);
      action->deleteLater();
    }

  /*
  ** Forward items must be reversed for display purposes.
  */

  list = p->forwardItems(MAX_HISTORY_ITEMS);

  for(int i = list.size() - 1; i >= 0; i--)
    {
      QString title(list.at(i).title());

      if(title.isEmpty())
	title = list.at(i).url().toString(QUrl::StripTrailingSlash);

      if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	title = title.mid
	  (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() + "...";

      if(title.contains("&"))
	title = title.replace("&", "&&");

      if(title.isEmpty())
	title = tr("Dooble Web Browser");

      QIcon icon(list.at(i).icon());

      if(icon.isNull())
	icon = dmisc::iconForUrl(list.at(i).url());

      QAction *action = ui.forwardToolButton->menu()->addAction
	(icon, title, this, SLOT(slotLinkActionTriggered(void)));

      if(action)
	action->setData
	  (QVariant(list.at(i).url().toString(QUrl::StripTrailingSlash)));
    }

  list.clear();
}

void dooble::slotGoHome(void)
{
  QStringList allKeys = s_settings.keys();

  for(int i = 0; i < allKeys.size(); i++)
    {
      if(!(allKeys[i].startsWith("settingsWindow/url") ||
	   allKeys[i].startsWith("settingsWindow/myRetrievedFiles")))
	continue;

      QUrl url(QUrl::fromUserInput(s_settings.value
				   (allKeys[i],
				    "").toString().trimmed()));

      if(!url.isEmpty() && url.isValid())
	{
	  bool urlFound = false;

	  for(int j = 0; j < ui.tabWidget->count(); j++)
	    {
	      dview *p = qobject_cast<dview *> (ui.tabWidget->widget(j));

	      if(p && p->url().toString(QUrl::StripTrailingSlash) ==
		 url.toString(QUrl::StripTrailingSlash))
		{
		  urlFound = true;
		  break;
		}
	    }

	  if(!urlFound)
	    {
	      newTab(url);
	      ui.tabWidget->update();
	    }
	}
    }

  allKeys.clear();
}

void dooble::closeTab(const int index)
{
  int count = ui.tabWidget->count();

  if(count == 1 && doobleCounter <= 1)
    {
      slotQuit();
      return;
    }
  else if(count == 1)
    {
      close();
      return;
    }

  if(index < 0)
    return;

  dview *p = qobject_cast<dview *> (ui.tabWidget->widget(index));

  if(p && p->isModified() &&
     s_settings.value("settingsWindow/warnBeforeClosingModifiedTab", false).
     toBool())
    {
      ui.tabWidget->setCurrentWidget(p);

      QMessageBox mb(this);

      mb.setIcon(QMessageBox::Question);
      mb.setWindowTitle(tr("Dooble Web Browser: Confirmation"));
      mb.setWindowModality(Qt::ApplicationModal);
      mb.setStandardButtons(QMessageBox::No | QMessageBox::Yes);
      mb.setText(tr("Are you sure that you wish to close this modified "
		    "tab?"));

      QSettings settings(s_settings.value("iconSet").toString(),
			 QSettings::IniFormat);

      for(int i = 0; i < mb.buttons().size(); i++)
	if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("okButtonIcon").toString()));
	else
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("cancelButtonIcon").toString()));

      mb.setWindowIcon
	(QIcon(settings.value("mainWindow/windowIcon").toString()));

      if(mb.exec() == QMessageBox::No)
	return;
    }

  if(count > 1)
    {
      if(p)
	{
	  if(!p->url().isEmpty() && p->url().isValid() &&
	     s_settings.value("settingsWindow/rememberClosedTabs",
			      false).toBool())
	    {
	      int number = s_settings.value
		("settingsWindow/rememberClosedTabsCount", 1).toInt();

	      if(number < 1 || number > MAX_HISTORY_ITEMS)
		number = 1;

	      dwebhistory *history = new dwebhistory(p->history());

	      m_closedTabs.prepend(history);

	      QString title(p->title());

	      if(title.isEmpty())
		title = p->url().toString(QUrl::StripTrailingSlash);

	      if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
		title = title.mid
		  (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
		  "...";

	      if(title.contains("&"))
		title = title.replace("&", "&&");

	      if(title.isEmpty())
		title = tr("Dooble Web Browser");

	      QAction *action = new QAction(p->icon(), title, this);

	      connect(action, SIGNAL(triggered(void)), this,
		      SLOT(slotReopenClosedTab(void)));

	      if(ui.historyMenu->actions().size() >= 4 &&
		 !ui.historyMenu->actions()[3]->menu())
		ui.historyMenu->actions()[3]->setMenu(new QMenu(this));

	      if(ui.historyMenu->actions().size() >= 4)
		ui.historyMenu->actions()[3]->menu()->insertAction
		  (ui.historyMenu->actions()[3]->menu()->actions().value(0),
		   action);

	      if(ui.historyMenu->actions().size() >= 1)
		ui.historyMenu->actions()[0]->setEnabled(true);

	      if(ui.historyMenu->actions().size() >= 4)
		ui.historyMenu->actions()[3]->setEnabled(true);

	      for(int i = m_closedTabs.size() - 1; i >= number; i--)
		{
		  delete m_closedTabs.takeAt(i);

		  if(ui.historyMenu->actions().size() >= 4)
		    ui.historyMenu->actions()[3]->menu()->actions()[i]->
		      deleteLater();
		}
	    }

	  p->deleteLater();

	  /*
	  ** The status bar should be reset if the user closed
	  ** a tab while the mouse cursor was on a link.
	  */

	  sb.statusLabel->clear();
	}
      else if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(index)))
	closeDesktop();
      else
	{
	  /*
	  ** Close a plugin tab?
	  */
	}

      count -= 1;
    }

  ui.tabWidget->setTabsClosable(count > 1);

  if(count == 1)
    ui.tabWidget->setBarVisible
      (s_settings.value("settingsWindow/alwaysShowTabBar",
			true).toBool());
  else
    ui.tabWidget->setBarVisible(true);

  ui.tabWidget->update();
}

void dooble::slotCloseTab(void)
{
  closeTab(ui.tabWidget->currentIndex());
}

void dooble::slotCloseTab(const int index)
{
  closeTab(index);
}

void dooble::slotOpenUrl(void)
{
  if(isFullScreen())
    slotMarkerEntered();

  ui.locationLineEdit->setFocus();
  ui.locationLineEdit->selectAll();
  update();
}

void dooble::slotLinkActionTriggered(void)
{
  QAction *action = qobject_cast<QAction *> (sender());

  if(action)
    {
      bool doLoad = true;
      QList<QWidget *> list(action->associatedWidgets());

      for(int i = 0; i < list.size(); i++)
	if(qobject_cast<dview *> (list.at(i)))
	  {
	    doLoad = false;
	    ui.tabWidget->setCurrentWidget(list.at(i));
	    break;
	  }

      list.clear();

      if(doLoad)
	{
	  loadPage(action->data().toUrl());
	  ui.tabWidget->currentWidget()->setFocus();
	}
    }
}

void dooble::slotLoadStarted(void)
{
  dview *p = qobject_cast<dview *> (sender());

  if(p)
    ui.tabWidget->animateIndex(ui.tabWidget->indexOf(p), true, QIcon());

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      sb.statusLabel->clear();
      sb.progressBar->setVisible(true);
      ui.reloadStopWidget->setCurrentIndex(0);
      ui.stopToolButton->setEnabled(true);
      ui.actionStop->setEnabled(true);
      ui.backToolButton->setEnabled(p->canGoBack());
      ui.forwardToolButton->setEnabled(p->canGoForward());
      ui.locationLineEdit->setSecureColor(false);
    }
}

void dooble::slotSearch(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      bool post = false;
      QString urlText("");
      QString postText("");
      QString searchName(ui.searchLineEdit->type());

      if(searchName == "bing")
	urlText = QString("http://www.bing.com/search?q=%1").arg
	  (ui.searchLineEdit->text().trimmed().replace(" ", "+"));
      else if(searchName == "dogpile")
	urlText = QString("http://www.dogpile.com/dogpile_other/ws/results/"
			  "Web/%1/1/417/TopNavigation/Relevance/iq=true/"
			  "zoom=off/_iceUrlFlag=7?_IceUrl=true").arg
	  (ui.searchLineEdit->text().trimmed());
      else if(searchName == "duckduckgo")
	{
	  post = true;

	  /*
	  ** kd=1: Redirect (Prevent Information Sharing)
	  ** kg=p: Post
	  ** kh=1: HTTPS
	  ** kp=1: Safe Search
	  */

	  urlText = "https://duckduckgo.com?kd=1&kg=p&kh=1&kp=1";
	  postText = QString("q=%1").arg
	    (ui.searchLineEdit->text().trimmed());
	}
      else if(searchName == "google")
        urlText = QString("https://www.google.com/search?q=%1").arg
	  (ui.searchLineEdit->text().trimmed().replace(" ", "+"));
      else if(searchName == "ixquick")
	{
	  post = true;
	  urlText = "https://ixquick.com/do/metasearch.pl";
	  postText = QString("query=%1").
	    arg(ui.searchLineEdit->text().trimmed());
	}
      else if(searchName == "history")
	{
	  ui.actionShow_HistorySideBar->setChecked(true);
	  m_historySideBar->search(ui.searchLineEdit->text().trimmed().
				   toLower());
	}
      else if(searchName == "wikipedia")
	{
	  QLocale locale;

	  urlText = QString
            ("https://%1.wikipedia.org/wiki/Special:"
	     "Search?search=%2").
	    arg(locale.name().left(2)).
	    arg(ui.searchLineEdit->text().trimmed().replace(" ", "+"));
	}
      else if(searchName == "wikinews")
	{
	  QLocale locale;

	  urlText = QString
	    ("https://secure.wikimedia.org/wikinews/%1/w/index.php?"
	     "search=%2").
	    arg(locale.name().left(2)).
	    arg(ui.searchLineEdit->text().trimmed().replace(" ", "+"));
	}
      else if(searchName == "wolframalpha")
	urlText = QString("http://www.wolframalpha.com/input/?i=%1").
	  arg(ui.searchLineEdit->text().trimmed().replace(" ", "+"));
      else if(searchName == "yacy")
        urlText = QString
	  ("http://localhost:8080/yacysearch.html?display=2&count=100"
	   "&resource=global&query=%1").arg
	  (ui.searchLineEdit->text().trimmed().replace(" ", "+"));
      else
	/*
	** Unknown!
	*/

	return;

      ui.tabWidget->setTabText(ui.tabWidget->indexOf(p),
			       tr("Loading..."));
      ui.tabWidget->setTabToolTip(ui.tabWidget->indexOf(p),
				  tr("Loading..."));

      if(post)
	p->post(QUrl::fromUserInput(urlText), postText);
      else
	p->load(QUrl::fromUserInput(urlText));
    }
}

void dooble::slotCopyLink(const QUrl &url)
{
  QApplication::clipboard()->setText(url.toString(QUrl::StripTrailingSlash));
}

void dooble::slotOpenLinkInNewTab(const QUrl &url)
{
  dview *p = newTab(url);

  if(p)
    {
      if(s_settings.value("settingsWindow/proceedToNewTab",
			  true).toBool())
	ui.tabWidget->setCurrentWidget(p);

      ui.tabWidget->update();
    }
}

void dooble::slotOpenLinkInNewWindow(const QUrl &url)
{
  if(s_settings.value("settingsWindow/openUserWindowsInNewProcesses",
		      false).toBool())
    launchDooble(url);
  else
    {
      Q_UNUSED(new dooble(url));
    }
}

void dooble::slotOpenPageInNewWindow(const int index)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->widget(index));

  if(p)
    {
      int count = ui.tabWidget->count() - 1;

      if(s_settings.value
	 ("settingsWindow/openUserWindowsInNewProcesses",
	  false).toBool())
	{
	  launchDooble(p->url());
	  p->deleteLater();
	}
      else
	{
	  Q_UNUSED(new dooble(p, ui.tabWidget->tabText(index)));
	}

      ui.tabWidget->setTabsClosable(count > 1);
    }
}

void dooble::slotLinkHovered(const QString &link, const QString &title,
			     const QString &textContent)
{
  Q_UNUSED(title);
  Q_UNUSED(textContent);

  if(link.isEmpty())
    sb.statusLabel->clear();
  else
    sb.statusLabel->setText(link.trimmed().mid(0, 100));
}

void dooble::slotAbout(void)
{
  QMessageBox *mb = findChild<QMessageBox *> ("about");

  if(!mb)
    {
      mb = new QMessageBox(this);
      mb->setObjectName("about");
    }

  mb->setWindowModality(Qt::NonModal);
  mb->setStyleSheet("QMessageBox {background: white;}");
  mb->setWindowTitle(tr("Dooble Web Browser: About"));
  mb->setTextFormat(Qt::RichText);
  mb->setText
    (QString("<html>"
             "Dooble Web Browser<br><br>"
             "Version %1 Swedish Cabbage.<br>"
	     "Copyright (c) 2008 - 2011 Mad Hatter<sup>10/6</sup>.<br>"
	     "Qt version %3."
	     "<hr>"
	     "Please visit <a href='http://dooble.sf.net'>"
	     "http://dooble.sf.net</a> for more information."
	     "</html>").
     arg(DOOBLE_VERSION_STR).
     arg(QT_VERSION_STR));
  mb->setStandardButtons(QMessageBox::Ok);

  QSettings settings
    (s_settings.value("iconSet").toString(), QSettings::IniFormat);

  mb->setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));

  for(int i = 0; i < mb->buttons().size(); i++)
    if(mb->buttonRole(mb->buttons().at(i)) == QMessageBox::AcceptRole)
      mb->buttons().at(i)->setIcon
	(QIcon(settings.value("okButtonIcon").toString()));

  mb->setIconPixmap(QPixmap("Icons/AxB/dooble.png"));
  mb->show();
}

void dooble::slotSavePage(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      if(p->isDir() || p->isFtp() ||
	 p->url().scheme().toLower() == "qrc")
	saveHandler(p->url(), p->html(), "");
      else
	saveHandler(p->url(), "", "");
    }
}

void dooble::slotSaveUrl(const QUrl &url)
{
  saveHandler(url, "", "");
}

void dooble::slotSaveFile(const QString &fileName, const QUrl &url)
{
  saveHandler(url, "", fileName);
}

void dooble::saveHandler(const QUrl &url, const QString &html,
			 const QString &fileName)
{
  QString path
    (s_settings.value("settingsWindow/myRetrievedFiles", "").
     toString());
  QFileInfo fileInfo(path);

  if(!(fileInfo.isReadable() && fileInfo.isWritable()))
    {
      path = QDesktopServices::storageLocation
	(QDesktopServices::DesktopLocation);

      if(path.isEmpty() || !QFile::exists(path) ||
	 !QFileInfo(path).isWritable())
	path = QDir::homePath();

      if(path.isEmpty() || !QFile::exists(path) ||
	 !QFileInfo(path).isWritable())
	path = QDir::tempPath();
    }

  QWidget *parent = qobject_cast<QWidget *> (sender());

  if(!parent)
    parent = this;

  QString l_fileName(fileName);

  /*
  ** This fileDialog business is troublesome. At times,
  ** when the dialog is closed, Dooble will terminate.
  */

  QFileDialog *fileDialog = new QFileDialog(parent);

#ifdef Q_WS_MAC
  fileDialog->setAttribute(Qt::WA_MacMetalStyle, true);
#endif
  fileDialog->setWindowTitle(tr("Dooble Web Browser: Save As"));
  fileDialog->setFileMode(QFileDialog::AnyFile);
  fileDialog->setDirectory(path);
  fileDialog->setLabelText(QFileDialog::Accept, tr("&Save"));
  fileDialog->setAcceptMode(QFileDialog::AcceptSave);

#ifdef Q_WS_MAC
  /*
  ** Attempt to find the QLineEdit widget.
  */

  QLineEdit *lineEdit = fileDialog->findChild<QLineEdit *> ("fileNameEdit");

  if(!lineEdit)
    lineEdit = fileDialog->findChild<QLineEdit *> ();
#endif

  if(url.path().isEmpty() || url.path() == "/")
#ifdef Q_WS_MAC
    {
      if(lineEdit)
	{
	  if(l_fileName.isEmpty())
	    l_fileName = "dooble.download";

	  l_fileName = dmisc::findUniqueFileName
	    (l_fileName, fileDialog->directory());
	  lineEdit->setText(l_fileName);
	}
    }
#else
    {
      if(l_fileName.isEmpty())
	l_fileName = "dooble.download";

      l_fileName = dmisc::findUniqueFileName
	(l_fileName, fileDialog->directory());
      fileDialog->selectFile(l_fileName);
    }
#endif
  else if(url.path().contains("/"))
    {
      QString l_fileName(QFileInfo(url.path()).fileName());

      if(l_fileName.isEmpty())
	l_fileName = "dooble.download";

      l_fileName = dmisc::findUniqueFileName
	(l_fileName, fileDialog->directory());

#ifdef Q_WS_MAC
      if(lineEdit)
	lineEdit->setText(l_fileName);
#else
      fileDialog->selectFile(l_fileName);
#endif
    }

  /*
  ** We sometimes terminate on this very spot.
  ** Are we not allowed to start an event loop
  ** from a function that is called by a slot?
  */

  /*
  ** Now for a workaround.
  */

  connect(fileDialog,
	  SIGNAL(finished(int)),
	  this,
	  SLOT(slotHandleQuirkySaveDialog(int)));
  m_saveDialogHackContainer.url = url;
  m_saveDialogHackContainer.html = html;
  fileDialog->setWindowModality(Qt::WindowModal);
  fileDialog->show();
}

void dooble::slotHandleQuirkySaveDialog(int result)
{
  QFileDialog *fileDialog = qobject_cast<QFileDialog *> (sender());

  if(!fileDialog)
    return;

  /*
  ** You should not be here when their parents are out!
  */

  if(result == QDialog::Accepted)
    {
      QStringList list(fileDialog->selectedFiles());

      if(!list.isEmpty())
	{
	  s_downloadWindow->show();

	  QString fileName(QFileInfo(list[0]).absoluteFilePath());

	  if(!m_saveDialogHackContainer.html.isEmpty())
	    s_downloadWindow->addHtmlItem
	      (m_saveDialogHackContainer.html, fileName);
	  else if(!m_saveDialogHackContainer.url.scheme().toLower().
		  startsWith("file"))
	    s_downloadWindow->addUrlItem
	      (m_saveDialogHackContainer.url, fileName);
	  else
	    s_downloadWindow->addFileItem
	      (m_saveDialogHackContainer.url.toLocalFile(), fileName);

	  list.clear();
	}
    }

  fileDialog->deleteLater();
}

void dooble::slotDisplayDownloadWindow(void)
{
  s_downloadWindow->show();
}

void dooble::slotShowFind(void)
{
  if(qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

      if(p && (p->isDir() || p->isFtp()))
	ui.findFrame->setVisible(false);
      else
	{
	  showFindFrame = true;
	  ui.findFrame->setVisible(true);
	}

      if(p)
	ui.actionFind->setEnabled(!p->isDir() && !p->isFtp());

      if(ui.tabWidget->isVisible())
	{
	  ui.findLineEdit->setFocus();
	  ui.findLineEdit->selectAll();
	  update();
	}
    }
}

void dooble::slotHideFind(void)
{
  showFindFrame = false;
  ui.findFrame->setVisible(false);
}

void dooble::slotFindNext(void)
{
  find(QWebPage::FindWrapsAroundDocument);
}

void dooble::slotFindPrevious(void)
{
  find(QWebPage::FindBackward | QWebPage::FindWrapsAroundDocument);
}

void dooble::slotFind(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      p->page()->findText(QString(), QWebPage::HighlightAllOccurrences);

      if(ui.highlightAllCheckBox->isChecked())
	find(QWebPage::HighlightAllOccurrences);
    }
}

void dooble::find(QWebPage::FindFlags flags)
{
  QString text(ui.findLineEdit->text().trimmed());
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());
  static QPalette lineEditPalette(ui.findLineEdit->palette());

  if(p && !p->page()->findText(text, flags))
    {
      if(!text.isEmpty())
	{
	  QColor color(240, 128, 128); // Light Coral
	  QPalette palette(ui.findLineEdit->palette());

	  palette.setColor(ui.findLineEdit->backgroundRole(), color);
	  ui.findLineEdit->setPalette(palette);
	}
      else
	ui.findLineEdit->setPalette(lineEditPalette);
    }
  else if(p)
    ui.findLineEdit->setPalette(lineEditPalette);
}

void dooble::keyPressEvent(QKeyEvent *event)
{
  if(event)
    {
      bool processed = false;

      if(event->key() == Qt::Key_Escape && ui.findFrame->isVisible())
	{
	  processed = true;
	  showFindFrame = false;
	  ui.findFrame->setVisible(false);
	}
      else if(event->key() == Qt::Key_Escape)
	{
	  processed = true;
	  slotStop();

	  if(isFullScreen())
	    slotViewEntered();
	}
      else if(event->key() == Qt::Key_X &&
	      event->modifiers() == Qt::AltModifier)
	{
	  processed = true;
	  showMinimized();
	}
      else if(QKeySequence(event->modifiers() + event->key()) ==
	      QKeySequence(Qt::ControlModifier + Qt::ShiftModifier +
			   Qt::Key_Plus))
	{
	  processed = true;
	  ui.actionZoom_In->trigger();
	}

      if(!processed && (isFullScreen() || m_isJavaScriptWindow))
	{
	  /*
	  ** OK, this is going to be tricky.
	  ** Grab all the actions.
	  ** For every action that is enabled, determine
	  ** its shortcuts. If a shortcut matches the event's key
	  ** and the event's modifier, issue the action's trigger() method.
	  */

#ifdef Q_WS_X11
	  if(QKeySequence(event->modifiers() + event->key()) ==
	     QKeySequence(Qt::ControlModifier + Qt::Key_Equal) ||
	     QKeySequence(event->modifiers() + event->key()) ==
	     QKeySequence(Qt::ControlModifier + Qt::ShiftModifier +
			  Qt::Key_Plus))
	    {
	      ui.actionZoom_In->trigger();
	      return;
	    }
#endif

	  QKeySequence shortcut(event->modifiers() + event->key());

	  foreach(QAction *action, findChildren<QAction *> ())
	    if(action->isEnabled() && !action->shortcuts().isEmpty())
	      {
		QList<QKeySequence> shortcuts(action->shortcuts());

		for(int i = 0; i < shortcuts.size(); i++)
		  if(shortcut == shortcuts.at(i))
		    {
		      action->trigger();
		      break;
		    }

		shortcuts.clear();
	      }
	}
    }

  QMainWindow::keyPressEvent(event);
}

void dooble::slotPrint(void)
{
  QPrinter printer;
  QPrintDialog printDialog(&printer, this);

#ifdef Q_WS_MAC
  printDialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif
  printDialog.setSizeGripEnabled(true);

  if(printDialog.exec() == QDialog::Accepted)
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

      if(p)
	p->print(&printer);
    }
}

void dooble::slotPrintRequested(QWebFrame *frame)
{
  if(frame)
    {
      QPrinter printer;
      QPrintDialog printDialog(&printer, this);

#ifdef Q_WS_MAC
      printDialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif
      printDialog.setSizeGripEnabled(true);

      if(printDialog.exec() == QDialog::Accepted)
	frame->print(&printer);
    }
}

void dooble::slotPrintPreview(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      QPrinter printer;
      QPrintPreviewDialog printDialog(&printer, this);

#ifdef Q_WS_MAC
      printDialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif
      printDialog.setWindowModality(Qt::WindowModal);
      printDialog.setSizeGripEnabled(true);
      connect(&printDialog,
	      SIGNAL(paintRequested(QPrinter *)),
	      p->currentFrame(),
	      SLOT(print(QPrinter *)));

      if(printDialog.exec() == QDialog::Accepted)
	p->print(&printer);
    }
}

void dooble::slotShowDesktopTab(const bool state)
{
  bool desktopFound = false;

  for(int i = 0; i < ui.tabWidget->count(); i++)
    if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(i)))
      {
	desktopFound = true;

	if(state)
	  ui.tabWidget->setCurrentIndex(i);

	break;
      }

  if(!desktopFound)
    {
      QSettings settings
	(s_settings.value("iconSet").toString(), QSettings::IniFormat);

      settings.beginGroup("mainWindow");

      if(s_settings.value("settingsWindow/appendNewTabs",
			  false).toBool())
	ui.tabWidget->addTab(desktopWidget,
			     QIcon(settings.value("tabWidget").toString()),
			     tr("Dooble Desktop"));
      else
	ui.tabWidget->insertTab
	  (ui.tabWidget->currentIndex() + 1,
	   desktopWidget,
	   QIcon(settings.value("tabWidget").toString()),
	   tr("Dooble Desktop"));

      ui.tabWidget->setTabToolTip(ui.tabWidget->count() - 1,
				  tr("Dooble Desktop"));
      ui.desktopToolButton->setEnabled(false);
      ui.tabWidget->setTabsClosable(ui.tabWidget->count() > 1);

      QAction *action = new QAction(tr("Dooble Desktop"), this);

      action->setData("Desktop");
      action->setIcon
	(QIcon(settings.value("actionDesktop").toString()));
      connect(action, SIGNAL(triggered(void)), this,
	      SLOT(slotShowDesktopTab(void)));
      desktopWidget->setTabAction(action);
      prepareTabsMenu();

      if(state)
	ui.tabWidget->setCurrentWidget(desktopWidget);

      if(ui.tabWidget->count() == 1)
	ui.tabWidget->setBarVisible
	  (s_settings.value("settingsWindow/alwaysShowTabBar",
			    true).toBool());
      else
	ui.tabWidget->setBarVisible(true);
    }
}

void dooble::slotShowPageSource(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p && p->currentFrame())
    {
      dpagesourcewindow *ps = new dpagesourcewindow
	(p->url(),
	 p->currentFrame()->toHtml());

      connect(this,
	      SIGNAL(iconsChanged(void)),
	      ps,
	      SLOT(slotSetIcons(void)));

      /*
      ** The dpagesourcewindow object automatically destroys
      ** itself whenever it is closed.
      */
    }
}

void dooble::slotClearHistory(void)
{
  /*
  ** Clear histories.
  */

  foreach(QWidget *widget, QApplication::topLevelWidgets())
    {
      dooble *d = qobject_cast<dooble *> (widget);

      if(!d)
	continue;

      d->ui.locationLineEdit->clearHistory();
      emit d->clearHistory();

      while(!d->m_closedTabs.isEmpty())
	delete d->m_closedTabs.takeFirst();

      if(d->ui.historyMenu->actions().size() >= 4 &&
	 d->ui.historyMenu->actions()[3]->menu())
	{
	  while(!d->ui.historyMenu->actions()[3]->menu()->actions().
		isEmpty())
	    {
	      QAction *action = d->ui.historyMenu->actions()[3]->menu()->
		actions().last();

	      d->ui.historyMenu->actions()[3]->menu()->removeAction(action);
	      action->deleteLater();
	    }

	  d->ui.historyMenu->actions()[3]->menu()->deleteLater();
	}

      while(d->ui.historyMenu->actions().size() > 4)
	{
	  QAction *action = d->ui.historyMenu->actions().last();

	  d->ui.historyMenu->removeAction(action);
	  action->deleteLater();
	}

      if(!d->ui.historyMenu->actions().isEmpty())
	{
	  d->ui.historyMenu->actions().at(0)->setEnabled(false);
	  d->ui.historyMenu->actions().at(1)->setEnabled(true);
	  d->ui.historyMenu->actions().at(3)->setEnabled(false);
	}

      while(!d->ui.backToolButton->menu()->actions().isEmpty())
	{
	  QAction *action = d->ui.backToolButton->menu()->actions().first();

	  d->ui.backToolButton->menu()->removeAction(action);
	  action->deleteLater();
	}

      while(!d->ui.forwardToolButton->menu()->actions().isEmpty())
	{
	  QAction *action = d->ui.forwardToolButton->menu()->
	    actions().first();

	  d->ui.forwardToolButton->menu()->removeAction(action);
	  action->deleteLater();
	}

      d->ui.backToolButton->setEnabled(false);
      d->ui.forwardToolButton->setEnabled(false);
    }

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history");

    db.setDatabaseName(s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	QString hashType
	  (s_settings.value("settingsWindow/hashType").
	   toString());
	QSqlQuery query(db);

	query.exec("PRAGMA synchronous = OFF");
	query.exec
	  (QString("DELETE FROM history WHERE "
		   "passphrase_hash = '%1'").
	   arg(dmisc::passphraseHash(hashType)));
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history");
}

QList<QVariantList> dooble::tabUrls(void) const
{
  QList<QVariantList> urls;

  for(int i = 0; i < ui.tabWidget->count(); i++)
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->widget(i));

      if(p)
	{
	  QString title(p->title());

	  if(title.isEmpty())
	    title = p->url().toString(QUrl::StripTrailingSlash);

	  if(title.contains("&"))
	    title = title.replace("&", "&&");

	  if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
	    title = title.mid
	      (0, MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
	      "...";

	  if(title.isEmpty())
	    title = tr("Dooble Web Browser");

	  QVariantList list;

	  list.append(title);
	  list.append(p->url());
	  list.append(p->icon());
	  urls.append(list);
	}
    }

  return urls;
}

void dooble::slotShowSettingsWindow(void)
{
  s_settingsWindow->exec(this);
}

void dooble::slotViewZoomIn(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      p->view()->setZoomFactor(p->view()->zoomFactor() + 0.1);
      update();
    }
}

void dooble::slotViewZoomOut(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());
	
  if(p)
    {
      p->view()->setZoomFactor(p->view()->zoomFactor() - 0.1);
      update();
    }
}

void dooble::slotViewResetZoom(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p && p->page()->currentFrame())
    {
      p->page()->currentFrame()->setZoomFactor(1.0);
      update();
    }
}

void dooble::slotViewZoomTextOnly(bool enable)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p && p->page()->currentFrame())
    {
      p->page()->currentFrame()->page()->settings()->setAttribute
	(QWebSettings::ZoomTextOnly, enable);
      update();
    }
}

void dooble::slotStatusBarDisplay(void)
{
  if(isFullScreen())
    return;

  statusBar()->setVisible(ui.actionStatusbar->isChecked());

  QSettings settings;

  settings.setValue("mainWindow/statusbarDisplay", statusBar()->isVisible());
  s_settings["mainWindow/statusbarDisplay"] =
    statusBar()->isVisible();
}

void dooble::slotShowHiddenFiles(void)
{
  QSettings settings;

  settings.setValue("mainWindow/showHiddenFiles",
		    ui.actionShow_Hidden_Files->isChecked());
  s_settings["mainWindow/showHiddenFiles"] =
    ui.actionShow_Hidden_Files->isChecked();

  if(s_settings.value("mainWindow/showHiddenFiles", true).toBool())
    {
      dfilemanager::treeModel->setFilter
	(dfilemanager::treeModel->filter() | QDir::Hidden);
      dfilemanager::tableModel->setFilter
	(dfilemanager::tableModel->filter() | QDir::Hidden);
    }
  else
    {
      dfilemanager::treeModel->setFilter
	(dfilemanager::treeModel->filter() & ~QDir::Hidden);
      dfilemanager::tableModel->setFilter
	(dfilemanager::tableModel->filter() & ~QDir::Hidden);
    }
}

void dooble::closeDesktop(void)
{
  /*
  ** Destroy the widget that is housing the Desktop.
  */

  for(int i = 0; i < ui.tabWidget->count(); i++)
    if(qobject_cast<ddesktopwidget *> (ui.tabWidget->widget(i)))
      {
	ui.tabWidget->removeTab(i);
	ui.desktopToolButton->setEnabled(true);
	break;
      }

  for(int i = 0; i < ui.menu_Tabs->actions().size(); i++)
    if(ui.menu_Tabs->actions().at(i)->data().toString() == "Desktop")
      {
	ui.menu_Tabs->removeAction(ui.menu_Tabs->actions().at(i));
	break;
      }

  ui.tabWidget->setTabsClosable(ui.tabWidget->count() > 1);
}

void dooble::slotOpenDirectory(void)
{
  QFileDialog fileDialog(this);

#ifdef Q_WS_MAC
  fileDialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif
  fileDialog.setWindowModality(Qt::WindowModal);
  fileDialog.setWindowTitle(tr("Dooble Web Browser: Open Directory"));
  fileDialog.setFileMode(QFileDialog::Directory);
  fileDialog.setDirectory(QDir::homePath());
  fileDialog.setLabelText(QFileDialog::Accept, tr("&Open"));
  fileDialog.setAcceptMode(QFileDialog::AcceptOpen);

  if(fileDialog.exec() == QDialog::Accepted)
    {
      QStringList list(fileDialog.selectedFiles());

      if(!list.isEmpty())
	{
	  if(qobject_cast<ddesktopwidget *> (ui.tabWidget->currentWidget()))
	    desktopWidget->showFileManagerWindow
	      (QUrl::fromLocalFile(list[0]));
	  else
	    {
	      loadPage(QUrl::fromLocalFile(list[0]));
	      ui.tabWidget->currentWidget()->setFocus();
	    }

	  list.clear();
	}
    }
}

void dooble::slotOpenMyRetrievedFiles(void)
{
  if(qobject_cast<ddesktopwidget *> (ui.tabWidget->currentWidget()))
    desktopWidget->showFileManagerWindow();
  else
    {
      QString path
	(s_settings.value
	 ("settingsWindow/myRetrievedFiles", "").toString());
      QFileInfo fileInfo(path);

      if(!(fileInfo.isReadable() && fileInfo.isWritable()))
	{
	  QString path = QDesktopServices::storageLocation
	    (QDesktopServices::DesktopLocation);

	  if(path.isEmpty() || !QFile::exists(path))
	    path = QDir::homePath();
	}

      loadPage(QUrl::fromLocalFile(path));
      ui.tabWidget->currentWidget()->setFocus();
    }
}

void dooble::slotChangeDesktopBackgrounds(void)
{
  foreach(QWidget *widget, QApplication::topLevelWidgets())
    if(qobject_cast<dooble *> (widget))
      {
	dooble *d = qobject_cast<dooble *> (widget);

	emit d->updateDesktopBackground();
      }
}

void dooble::saveHistoryItem(const dwebhistoryitem &item)
{
  if(!item.isValid())
    return;

  if(!s_settings.value("settingsWindow/rememberHistory",
		       true).toBool())
    return;

  {
    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history");

    db.setDatabaseName(s_homePath +
		       QDir::separator() + "history.db");

    if(db.open())
      {
	int count = 1;
	QSqlQuery query(db);

	query.exec("CREATE TABLE IF NOT EXISTS history ("
		   "title TEXT, "
		   "url TEXT PRIMARY KEY NOT NULL, "
		   "host TEXT NOT NULL, "
		   "last_visited TEXT NOT NULL, "
		   "icon BLOB DEFAULT NULL, "
		   "visits TEXT NOT NULL, "
		   "passphrase_hash TEXT NOT NULL)");

	if(query.exec(QString("SELECT visits "
			      "FROM history WHERE HEX(url) = "
			      "HEX('%1')").
		      arg(dmisc::encodedString
			  (item.url().toString(QUrl::StripTrailingSlash).
			   toAscii(),
			   true).toBase64().data())))
	  {
	    if(query.next())
	      count = dmisc::decodedString
		(QByteArray::fromBase64
		 (query.value(0).toString().toAscii())).toInt() + 1;
	  }

	query.exec("PRAGMA synchronous = OFF");
	query.prepare
	  ("INSERT OR REPLACE INTO history "
	   "(url, host, icon, last_visited, title, visits, passphrase_hash) "
	   "VALUES (?, ?, ?, ?, ?, ?, ?)");
	query.bindValue
	  (0,
	   dmisc::encodedString
	   (item.url().toString(QUrl::StripTrailingSlash).toAscii(),
	    true).toBase64());
	query.bindValue
	  (1,
	   dmisc::encodedString(item.url().host().toAscii(),
				true).toBase64());

	QBuffer buffer;
	QByteArray bytes;

	buffer.setBuffer(&bytes);
	buffer.open(QIODevice::WriteOnly);

	QDataStream out(&buffer);

	out << item.icon();
	query.bindValue
	  (2,
	   dmisc::encodedString(bytes,
				true).toBase64());
	buffer.close();
	query.bindValue
	  (3,
	   dmisc::encodedString(item.visitDate().toString(Qt::ISODate).
				toAscii(),
				true).toBase64());

	if(item.title().trimmed().isEmpty())
	  query.bindValue
	    (4,
	     dmisc::encodedString(item.url().
				  toString(QUrl::StripTrailingSlash).
				  toAscii(),
				  true).toBase64());
	else
	  query.bindValue
	    (4,
	     dmisc::encodedString(item.title().toAscii(),
				  true).toBase64());

	query.bindValue
	  (5,
	   dmisc::encodedString(QString::number(count).toAscii(),
				true).toBase64());

	QString hashType(s_settings.value("settingsWindow/hashType").
			 toString());

	query.bindValue
	  (6,
	   dmisc::passphraseHash(hashType));
	query.exec();
      }

    db.close();
  }

  QSqlDatabase::removeDatabase("history");
}

void dooble::slotCopy(void)
{
  if(ui.searchLineEdit->hasFocus() &&
     !ui.searchLineEdit->selectedText().isEmpty())
    ui.searchLineEdit->copy();
  else if(ui.locationLineEdit->hasFocus() &&
	  !ui.locationLineEdit->selectedText().isEmpty())
    ui.locationLineEdit->copy();
  else
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

      if(p && !p->isDir() && !p->isFtp())
	p->page()->triggerAction(QWebPage::Copy);
    }
}

void dooble::slotPaste(void)
{
  if(ui.searchLineEdit->hasFocus())
    ui.searchLineEdit->paste();
  else if(ui.locationLineEdit->hasFocus())
    ui.locationLineEdit->paste();
  else
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

      if(p && !p->isDir() && !p->isFtp())
	p->page()->triggerAction(QWebPage::Paste);
    }
}

void dooble::slotSelectAllContent(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    p->page()->triggerAction(QWebPage::SelectAll);
}

void dooble::slotAuthenticationRequired(QNetworkReply *reply,
					QAuthenticator *authenticator)
{
  if(!authenticator)
    return;

  /*
  ** We should inspect the reply object. However, based on the
  ** name of the signal authenticationRequired and the capabilities of
  ** QAuthenticator, we can conclude that credentials are sought.
  */

  Q_UNUSED(reply);
  QDialog dialog(this);
  Ui_passwordDialog ui_p;

  ui_p.setupUi(&dialog);
#ifdef Q_WS_MAC
  dialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif

  QSettings settings(s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);

  dialog.setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));

  for(int i = 0; i < ui_p.buttonBox->buttons().size(); i++)
    if(ui_p.buttonBox->buttonRole(ui_p.buttonBox->buttons().at(i)) ==
       QDialogButtonBox::YesRole)
      {
	qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
	  setDefault(true);
	ui_p.buttonBox->buttons().at(i)->setIcon
	  (QIcon(settings.value("okButtonIcon").toString()));
      }
    else
      {
	qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
	  setDefault(false);
	ui_p.buttonBox->buttons().at(i)->setIcon
	  (QIcon(settings.value("cancelButtonIcon").toString()));
      }

  if(dialog.exec() == QDialog::Accepted)
    {
      authenticator->setUser(ui_p.usernameLineEdit->text());
      authenticator->setPassword(ui_p.passwordLineEdit->text());
    }
  else
    {
      /*
      ** Abort the load if the user has rejected the dialog.
      */

      dnetworkaccessmanager *networkAccessManager =
	qobject_cast<dnetworkaccessmanager *> (sender());

      if(networkAccessManager)
	{
	  QWebPage *webPage = qobject_cast<QWebPage *>
	    (networkAccessManager->parent());

	  if(webPage)
	    {
	      dwebview *webView =
		qobject_cast<dwebview *> (webPage->parent());

	      if(webView)
		webView->stop();
	    }
	}
    }
}
void dooble::slotProxyAuthenticationRequired(const QNetworkProxy &proxy,
					     QAuthenticator *authenticator)
{
  if(!authenticator)
    return;

  Q_UNUSED(proxy);
  QDialog dialog(this);
  Ui_passwordDialog ui_p;

  ui_p.setupUi(&dialog);
#ifdef Q_WS_MAC
  dialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif

  QSettings settings(s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);

  dialog.setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));

  for(int i = 0; i < ui_p.buttonBox->buttons().size(); i++)
    if(ui_p.buttonBox->buttonRole(ui_p.buttonBox->buttons().at(i)) ==
       QDialogButtonBox::YesRole)
      {
	qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
	  setDefault(true);
	ui_p.buttonBox->buttons().at(i)->setIcon
	  (QIcon(settings.value("okButtonIcon").toString()));
      }
    else
      {
	qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
	  setDefault(false);
	ui_p.buttonBox->buttons().at(i)->setIcon
	  (QIcon(settings.value("cancelButtonIcon").toString()));
      }

  if(dialog.exec() == QDialog::Accepted)
    {
      authenticator->setUser(ui_p.usernameLineEdit->text());
      authenticator->setPassword(ui_p.passwordLineEdit->text());
    }
}

void dooble::slotFullScreenMode(void)
{
  bool setVisible = false;
  static QSize m_size = size();

  if(ui.actionFull_Screen_Mode->text().contains("Full"))
    {
      setVisible = false;
      ui.actionFull_Screen_Mode->setText("&Normal Screen Mode");
    }
  else
    {
      setVisible = true;
      ui.actionFull_Screen_Mode->setText("&Full Screen Tablet Mode");
    }

  if(!setVisible)
    {
      m_size = size();
      showFullScreen();
    }
  else
    {
      showNormal();
      resize(m_size);
    }

  if(m_isJavaScriptWindow)
    ui.toolBar->setVisible(setVisible);
  else
    {
      menuBar()->setVisible(setVisible);

      if(qobject_cast<dview *> (ui.tabWidget->currentWidget()))
	statusBar()->setVisible(ui.actionStatusbar->isChecked());
      else
	statusBar()->setVisible(false);

      ui.toolBar->setVisible(setVisible);
    }

  ui.marker->setVisible(!setVisible);
  ui.tabWidget->setBarVisible
    (setVisible &&
     s_settings.value("settingsWindow/alwaysShowTabBar",
		      true).toBool());
  ui.restoreToolButton->setVisible(!setVisible);
  activateWindow();
}

void dooble::slotMarkerEntered(void)
{
  ui.marker->setVisible(false);
  ui.toolBar->setVisible(true);
  ui.tabWidget->setBarVisible(true);
  ui.restoreToolButton->setVisible(true);
}

void dooble::slotViewEntered(void)
{
  if(isFullScreen())
    {
      ui.toolBar->setVisible(false);
      ui.tabWidget->setBarVisible(false);
      ui.restoreToolButton->setVisible(false);
      ui.marker->setVisible(true);
    }
}

void dooble::slotFontChanged(const QFont &font)
{
  QApplication::setFont(font);

  foreach(QWidget *widget, QApplication::allWidgets())
    if(!qobject_cast<dview *> (widget))
      widget->setFont(font);

  menuBar()->setFont(font);

  foreach(QMenu *menu, menuBar()->findChildren<QMenu *> ())
    {
      menu->setFont(font);

      foreach(QAction *action, menu->actions())
	action->setFont(font);
    }
}

void dooble::slotSelectionChanged(const QString &text)
{
  Q_UNUSED(text);
}

void dooble::slotSelectionChanged(void)
{
}

void dooble::slotEnablePaste(void)
{
}

void dooble::slotShowHistory(void)
{
  /*
  ** The current window should be the artificial parent.
  */

  disconnect(s_historyWindow, SIGNAL(open(const QUrl &)),
	     this, SLOT(slotLoadPage(const QUrl &)));
  disconnect(s_historyWindow, SIGNAL(createTab(const QUrl &)),
	     this, SLOT(slotOpenLinkInNewTab(const QUrl &)));
  disconnect(s_historyWindow, SIGNAL(openInNewWindow(const QUrl &)),
	     this, SLOT(slotOpenLinkInNewWindow(const QUrl &)));
  connect(s_historyWindow, SIGNAL(open(const QUrl &)),
	  this, SLOT(slotLoadPage(const QUrl &)));
  connect(s_historyWindow, SIGNAL(createTab(const QUrl &)),
	  this, SLOT(slotOpenLinkInNewTab(const QUrl &)));
  connect(s_historyWindow, SIGNAL(openInNewWindow(const QUrl &)),
	  this, SLOT(slotOpenLinkInNewWindow(const QUrl &)));
  s_historyWindow->show(this);
}

void dooble::slotResetUrl(void)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(p)
    {
      QString title(p->title());

      if(title.isEmpty())
	/*
	** The tab's title will be set to the URL.
	*/

	title = p->url().toString(QUrl::StripTrailingSlash);

      if(title.isEmpty())
	title = tr("(Untitled)");

      ui.tabWidget->setTabText(ui.tabWidget->indexOf(p),
			       title.replace("&", "&&"));
      ui.tabWidget->setTabToolTip(ui.tabWidget->indexOf(p),
				  title);
      ui.tabWidget->animateIndex
	(ui.tabWidget->indexOf(p), false, p->icon());
      ui.locationLineEdit->setText
	(p->url().toString(QUrl::StripTrailingSlash));
      ui.locationLineEdit->setIcon(p->icon());
      slotOpenUrl();
    }
}

void dooble::slotCloseWindow(void)
{
  if(sender() && sender()->parent() && ui.tabWidget->count() > 1)
    {
      dview *p = qobject_cast<dview *> (sender()->parent()->parent());

      if(p)
	closeTab(ui.tabWidget->indexOf(p));
    }
  else
    close();
}

void dooble::launchDooble(const QUrl &url)
{
  QProcess *process = new QProcess();
  QStringList arguments;

  if(!url.isEmpty() && url.isValid())
    arguments << "--url"
	      << url.toString(QUrl::StripTrailingSlash);

  process->start(QCoreApplication::applicationDirPath() +
		 QDir::separator() + QCoreApplication::applicationName(),
		 arguments);
  arguments.clear();
}

void dooble::slotTabMoved(int from, int to)
{
  if(from < 0 || to < 0)
    return;

  if(from < ui.menu_Tabs->actions().size() &&
     to < ui.menu_Tabs->actions().size())
    {
      QAction *toAction = ui.menu_Tabs->actions().at(to);
      QAction *fromAction = ui.menu_Tabs->actions().at(from);

      if(to < from)
	{
	  ui.menu_Tabs->removeAction(fromAction);
	  ui.menu_Tabs->insertAction(toAction, fromAction);
	}
      else
	{
	  ui.menu_Tabs->removeAction(toAction);
	  ui.menu_Tabs->insertAction(fromAction, toAction);
	}
    }
}

void dooble::slotGeometryChangeRequested(const QRect &geometry)
{
  if(geometry.isValid() && m_isJavaScriptWindow)
    if(s_settings.value("settingsWindow/javascriptEnabled",
			true).toBool() &&
       s_settings.value
       ("settingsWindow/javascriptAcceptGeometryChangeRequests",
	false).toBool())
      setGeometry(geometry);
}

void dooble::slotHideMenuBar(void)
{
  if(m_isJavaScriptWindow)
    if(s_settings.value("settingsWindow/javascriptEnabled",
			true).toBool() &&
       s_settings.value("settingsWindow/javascriptAllowMenuBarHiding",
			false).toBool())
      {
	menuBar()->setVisible(false);
	ui.historyFrame->setVisible(false);
      }
}

void dooble::slotHideStatusBar(void)
{
  if(m_isJavaScriptWindow)
    if(s_settings.value("settingsWindow/javascriptEnabled",
			true).toBool() &&
       s_settings.value
       ("settingsWindow/javascriptAllowStatusBarHiding",
	false).toBool() &&
       !isFullScreen())
      statusBar()->setVisible(false);
}

void dooble::slotHideToolBar(void)
{
  if(m_isJavaScriptWindow)
    if(s_settings.value("settingsWindow/javascriptEnabled",
			true).toBool() &&
       s_settings.value("settingsWindow/javascriptAllowToolBarHiding",
			false).toBool())
      {
	ui.backToolButton->setVisible(false);
	ui.forwardToolButton->setVisible(false);
	ui.homeToolButton->setVisible(false);
	ui.reloadStopWidget->setVisible(false);
	ui.desktopToolButton->setVisible(false);
	ui.searchLineEdit->setVisible(false);
      }
}

void dooble::slotSetTabBarVisible(const bool state)
{
  if(ui.tabWidget->count() == 1)
    ui.tabWidget->setBarVisible(state);
}

void dooble::slotShowApplicationCookies(void)
{
  s_cookieWindow->show(this);
}

void dooble::prepareWidgetsBasedOnView(dview *p)
{
  if(!p)
    return;

  ui.zoomMenu->setEnabled(!p->isDir() && !p->isFtp());
  ui.actionFind->setEnabled(!p->isDir() && !p->isFtp());
  ui.actionSelect_All_Content->setEnabled(!p->isDir() && !p->isFtp());
  ui.actionShow_Hidden_Files->setEnabled(p->isDir());

  if(p->isDir() || p->isFtp())
    ui.findFrame->setVisible(false);
  else
    ui.findFrame->setVisible(showFindFrame);
}

void dooble::slotLocationSplitterMoved(int pos, int index)
{
  Q_UNUSED(pos);
  Q_UNUSED(index);

  if(!m_isJavaScriptWindow)
    {
      QSettings settings;

      settings.setValue("mainWindow/splitterState",
			ui.splitter->saveState());
      s_settings["mainWindow/splitterState"] =
	ui.splitter->saveState();
    }
}

void dooble::slotHistoryTabSplitterMoved(int pos, int index)
{
  Q_UNUSED(pos);
  Q_UNUSED(index);

  if(!m_isJavaScriptWindow)
    {
      QSettings settings;

      settings.setValue("mainWindow/historyTabSplitterState",
			ui.historyAndTabSplitter->saveState());
      s_settings["mainWindow/historyTabSplitterState"] =
	ui.historyAndTabSplitter->saveState();
    }
}

void dooble::setCurrentPage(dview *p)
{
  if(p)
    ui.tabWidget->setCurrentWidget(p);
}

bool dooble::promptForPassphrase(const bool override)
{
  /*
  ** This methods returns true if it established
  ** a passphrase.
  */

  if(dmisc::passphraseWasAuthenticated())
    return false;

  QString hashType(s_settings.value("settingsWindow/hashType", "").
		   toString());

  if((override || doobleCounter == 1) &&
     !s_settings.value("settingsWindow/hashType",
		       "").toString().isEmpty() &&
     !s_settings.value("settingsWindow/passphraseHash",
		       "").toString().isEmpty() &&
     dmisc::isHashTypeSupported(hashType))
    {
      show();

      do
	{
	  QDialog dialog(this);
	  Ui_passphrasePrompt ui_p;

	  ui_p.setupUi(&dialog);
#ifdef Q_WS_MAC
	  dialog.setAttribute(Qt::WA_MacMetalStyle, true);
#endif
#ifdef Q_WS_WIN
	  ui_p.passphraseLineEdit->setStyleSheet
	    ("QLineEdit { "
	     "padding: 1px; "
	     "border-style: solid; "
	     "border: 2px solid gray; "
	     "border-radius: 8px; "
	     "}");

	  foreach(QPushButton *pushButton,
		  dialog.findChildren<QPushButton *> ())
	    pushButton->setStyleSheet
	    ("QPushButton { "
	     "color: black; "
	     "background-color: "
	     "QLinearGradient(x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #eef, "
	     "stop: 1 #ccf); "
	     "border-width: 1px; "
	     "border-color: #339; "
	     "border-style: solid; "
	     "border-radius: 7px; "
	     "padding: 3px; "
	     "font-size: 12px; "
	     "padding-left: 5px; "
	     "padding-right: 5px; "
	     "min-width: 50px; "
	     "max-width: 70px; "
	     "min-height: 16px; "
	     "max-height: 16px; "
	     "}");
#endif

	  QSettings settings(s_settings.value("iconSet").toString(),
			     QSettings::IniFormat);

	  ui_p.label->setPixmap
	    (QPixmap(settings.value("settingsWindow/safeButtonIcon").
		     toString()));
	  dialog.setWindowIcon
	    (QIcon(settings.value("mainWindow/windowIcon").toString()));

	  for(int i = 0; i < ui_p.buttonBox->buttons().size(); i++)
	    if(ui_p.buttonBox->buttonRole(ui_p.buttonBox->buttons().at(i)) ==
	       QDialogButtonBox::AcceptRole)
	      {
		qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
		  setDefault(true);
		ui_p.buttonBox->buttons().at(i)->setIcon
		  (QIcon(settings.value("okButtonIcon").toString()));
	      }
	    else
	      {
		qobject_cast<QPushButton *> (ui_p.buttonBox->buttons().at(i))->
		  setDefault(false);
		ui_p.buttonBox->buttons().at(i)->setIcon
		  (QIcon(settings.value("cancelButtonIcon").toString()));
	      }

	  if(dialog.exec() == QDialog::Accepted)
	    {
	      QString hash("");

	      /*
	      ** Validate the passphrase.
	      */

	      hash = dmisc::passphraseHash(ui_p.passphraseLineEdit->text(),
					   hashType);

	      if(hash == s_settings.value("settingsWindow/"
					  "passphraseHash").toString())
		{
		  dmisc::setCipherPassphrase(ui_p.passphraseLineEdit->text());
		  ui_p.passphraseLineEdit->clear();
		  emit passphraseWasAuthenticated
		    (dmisc::passphraseWasAuthenticated());

		  foreach(QWidget *widget, QApplication::topLevelWidgets())
		    if(qobject_cast<dooble *> (widget))
		      {
			dooble *d = qobject_cast<dooble *> (widget);

			if(d)
			  d->ui.action_Authenticate->setEnabled
			    (!dmisc::passphraseWasAuthenticated());
		      }

		  break;
		}
	      else
		ui_p.passphraseLineEdit->clear();
	    }
	  else
	    {
	      ui_p.passphraseLineEdit->clear();

	      /*
	      ** Create a random super-secret key.
	      */

	      dmisc::setCipherPassphrase();
	      break;
	    }
	}
      while(true);

      return true;
    }
  else if(!override && doobleCounter == 1)
    {
      dmisc::setCipherPassphrase();
      return true;
    }

  return false;
}

void dooble::initializeHistoryMenu(void)
{
  /*
  ** Add history menu actions, but only if the user wishes
  ** to have a history. The entries will be date-ordered.
  */

  if(s_settings.value("settingsWindow/rememberHistory",
		      true).toBool())
    {
      {
	QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", "history");

	db.setDatabaseName(s_homePath +
			   QDir::separator() + "history.db");

	if(db.open())
	  {
	    QString hashType
	      (s_settings.value("settingsWindow/hashType").
	       toString());
	    QSqlQuery query(db);
	    QPair<QString, QIcon> pair;
	    QMap<QDateTime, QString> titles; // This is really inconvenient.
	    QMap<QDateTime, QPair<QString, QIcon> > items;

	    query.exec
	      (QString("SELECT url, title, icon, "
		       "last_visited, host, visits "
		       "FROM history WHERE "
		       "passphrase_hash = '%1'").
	       arg(dmisc::passphraseHash(hashType)));

	    while(query.next())
	      {
		QUrl url;
		QString urlStr(query.value(0).toString());

		/*
		** Simple test to determine if the URL
		** can be decoded. If it can't be,
		** ignore the history entry.
		*/

		if(!(url = QUrl::fromEncoded
		     (dmisc::decodedString
		      (QByteArray::fromBase64
		       (urlStr.toAscii())).data(),
		      QUrl::StrictMode)).isValid())
		  {
		    QSqlQuery deleteQuery(db);

		    deleteQuery.exec
		      ("PRAGMA synchronous = OFF");
		    deleteQuery.exec(QString("DELETE FROM history WHERE "
					     "HEX(url) = HEX('%1')").
				     arg(query.value(0).toString()));
		    continue;
		  }

		if(!(url.scheme().toLower() == "file" ||
		     url.scheme().toLower() == "ftp" ||
		     url.scheme().toLower() == "http" ||
		     url.scheme().toLower() == "https" ||
		     url.scheme().toLower() == "qrc"))
		  continue;

		if(ui.historyMenu->actions().size() == 4)
		  {
		    ui.historyMenu->actions()[0]->setEnabled(true);
		    ui.historyMenu->actions()[1]->setEnabled(true);
		    ui.historyMenu->addSeparator();
		  }

		QIcon icon;

		if(!query.isNull(2))
		  {
		    QBuffer buffer;
		    QByteArray bytes
		      (QByteArray::fromBase64(query.value(2).toByteArray()));

		    bytes = dmisc::decodedString(bytes);
		    buffer.setBuffer(&bytes);
		    buffer.open(QIODevice::ReadOnly);

		    QDataStream in(&buffer);

		    in >> icon;
		    buffer.close();
		  }

		QString title("");
		QDateTime dateTime
		  (QDateTime::fromString
		   (dmisc::decodedString
		    (QByteArray::fromBase64
		     (query.value(3).toString().toAscii())).data(),
		    Qt::ISODate));

		title = dmisc::decodedString
		  (QByteArray::fromBase64
		   (query.value(1).toString().toAscii())).data();

		if(title.isEmpty())
		  title = url.toString(QUrl::StripTrailingSlash);

		if(title.length() > MAX_NUMBER_OF_MENU_TITLE_CHARACTERS)
		  title = title.mid
		    (0,
		     MAX_NUMBER_OF_MENU_TITLE_CHARACTERS - 3).trimmed() +
		    "...";

		if(title.contains("&"))
		  title = title.replace("&", "&&");

		if(title.isEmpty())
		  title = tr("Dooble Web Browser");

		titles[dateTime] = title;
		pair.first = url.toString(QUrl::StripTrailingSlash);
		pair.second = icon;
		items[dateTime] = pair;

		QString host
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(4).toString().toAscii())).data());
		QString visits
		  (dmisc::decodedString
		   (QByteArray::fromBase64
		    (query.value(5).toString().toAscii())).data());

		if(!host.isEmpty())
		  s_mostVisitedHosts[host] += visits.toInt();
	      }

	    for(int i = items.size() - 1; i >= 0; i--)
	      {
		QAction *action = new QAction(items.values().at(i).second,
					      titles.values().at(i), this);

		action->setData(items.values().at(i).first);
		connect(action, SIGNAL(triggered(void)), this,
			SLOT(slotLinkActionTriggered(void)));
		ui.historyMenu->addAction(action);

		if(ui.historyMenu->actions().size() >= 5 + MAX_HISTORY_ITEMS)
		  break;
	      }

	    titles.clear();

	    /*
	    ** The history menu should have at least three entries
	    ** at this point.
	    */

	    ui.locationLineEdit->appendItems(items);
	    items.clear();
	  }

	db.close();
      }

      QSqlDatabase::removeDatabase("history");
    }
}

void dooble::slotOpenP2PEmail(void)
{
  QUrl url(QUrl::fromUserInput
	   (s_settings.value("settingsWindow/p2pUrl",
			     "about: blank").toString()));

  loadPage(url);
  ui.tabWidget->currentWidget()->setFocus();
}

void dooble::slotRepaintRequested(const QRect &dirtyRect)
{
  dwebpage *p = qobject_cast<dwebpage *> (sender());

  if(p && p->mainFrame())
    p->mainFrame()->render(0, dirtyRect);
}

void dooble::slotShowFavoritesToolBar(bool checked)
{
  ui.favoritesToolBar->blockSignals(true);
  ui.favoritesToolBar->setVisible(checked);
  ui.favoritesToolBar->blockSignals(false);

  QSettings settings;

  settings.setValue("mainWindow/showFavoritesToolBar", checked);
  s_settings["mainWindow/showFavoritesToolBar"] = checked;
}

void dooble::prepareMostVisited(void)
{
  foreach(QToolButton *toolButton,
	  ui.favoritesToolButtonsFrame->findChildren<QToolButton *> ())
    toolButton->deleteLater();

  QList<int> list(s_mostVisitedHosts.values());
  QStringList addedHosts;

  qSort(list);

  for(int i = list.size() - 1, j = 0; i >= 0; i--)
    {
      QList<QString> keys(s_mostVisitedHosts.keys(list.at(i)));

      /*
      ** We should have something like:
      ** (Count (C1 >= C2 >= ... >= CN): Host
      ** (H1 >= H2 >= ... >= HI >= ... >= HJ))
      ** C1: H1
      ** C1: H2
      ** ...
      ** C1: HI
      ** ...
      ** CN: HJ
      */

      qSort(keys);

      /*
      ** Multiple hosts may enjoy identical counts.
      */

      for(int k = 0; k < keys.size(); k++)
	if(j < MAX_MOST_VISITED_ITEMS)
	  {
	    if(!keys.at(k).isEmpty() &&
	       !addedHosts.contains(keys.at(k)))
	      {
		QUrl url(QUrl::fromUserInput(keys.at(k)));
		QIcon icon(dmisc::iconForUrl(url));
		QToolButton *toolButton = new QToolButton
		  (ui.favoritesToolButtonsFrame);

		toolButton->setIcon(icon);
		toolButton->setToolTip
		  (url.toString(QUrl::StripTrailingSlash));
		toolButton->setIconSize(QSize(16, 16));
#ifdef Q_WS_MAC
		toolButton->setStyleSheet
		  ("QToolButton {border: none;}"
		   "QToolButton::menu-button {border: none;}");
#else
		toolButton->setAutoRaise(true);
#endif
		ui.favoritesToolButtonsFrame->layout()->
		  addWidget(toolButton);
		toolButton->setProperty("url", url);
		connect(toolButton,
			SIGNAL(clicked(void)),
			this,
			SLOT(slotFavoritesToolButtonClicked(void)));
		j += 1;
		addedHosts.append(keys.at(k));
	      }
	  }
	else
	  break;

      if(j >= MAX_MOST_VISITED_ITEMS)
	break;
    }

  list.clear();
  addedHosts.clear();
}

void dooble::slotFavoritesToolButtonClicked(void)
{
  QToolButton *toolButton = qobject_cast<QToolButton *> (sender());

  if(toolButton)
    {
      dview *p = newTab(toolButton->property("url").toUrl());

      if(p)
	{
	  ui.tabWidget->setCurrentWidget(p);
	  ui.tabWidget->update();
	}
    }
}

QMenu *dooble::createPopupMenu(void)
{
  return 0;
}

void dooble::setCustomStyle(void)
{
  /*
  ** Empty.
  */
}

void dooble::slotShowIpAddress(const bool state)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());

  if(!p)
    return;

  if(state)
    p->findIpAddress();
  else
    {
      if(p->title().isEmpty())
	setWindowTitle(tr("Dooble Web Browser"));
      else
	setWindowTitle(p->title() + tr(" - Dooble Web Browser"));
    }
}

void dooble::slotIpAddressChanged(const QString &ipAddress)
{
  dview *p = qobject_cast<dview *> (sender());

  if(p == qobject_cast<dview *> (ui.tabWidget->currentWidget()))
    {
      if(p->title().isEmpty())
	{
	  if(ipAddress.isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(tr("Dooble Web Browser"));
	  else
	    setWindowTitle(QString(tr("Dooble Web Browser (%1)")).
			   arg(ipAddress));
	}
      else
	{
	  if(ipAddress.isEmpty() ||
	     !s_settings.value("settingsWindow/displayIpAddress",
			       false).toBool())
	    setWindowTitle(p->title() + tr(" - Dooble Web Browser"));
	  else
	    setWindowTitle(p->title() + QString(" (%1)").arg(ipAddress) +
			   tr(" - Dooble Web Browser"));
	}
    }
}

void dooble::slotShowHistorySideBar(bool state)
{
  ui.historyFrame->setVisible(state);
  dooble::s_settings["mainWindow/showHistorySideBar"] = state;

  QSettings settings;

  settings.setValue("mainWindow/showHistorySideBar", state);
}

void dooble::slotHistorySideBarClosed(void)
{
  ui.actionShow_HistorySideBar->setChecked(false);
  slotShowHistorySideBar(false);
}

void dooble::slotOpenUrlsFromDrop(const QList<QUrl> &list)
{
  if(list.size() > 5)
    {
      QApplication::restoreOverrideCursor();

      QMessageBox mb(this);

      mb.setIcon(QMessageBox::Question);
      mb.setWindowTitle(tr("Dooble Web Browser: Confirmation"));
      mb.setWindowModality(Qt::ApplicationModal);
      mb.setStandardButtons(QMessageBox::No | QMessageBox::Yes);
      mb.setText(QString(tr("Are you sure that you wish to open "
			    "%1 pages?")).arg(list.size()));

      QSettings settings(s_settings.value("iconSet").toString(),
			 QSettings::IniFormat);

      for(int i = 0; i < mb.buttons().size(); i++)
	if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("okButtonIcon").toString()));
	else
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("cancelButtonIcon").toString()));

      mb.setWindowIcon
	(QIcon(settings.value("mainWindow/windowIcon").toString()));

      if(mb.exec() == QMessageBox::No)
	return;
    }

  for(int i = 0; i < list.size(); i++)
    if(i == 0)
      loadPage(list.at(i));
    else
      newTab(list.at(i));

  ui.tabWidget->update();
  ui.tabWidget->currentWidget()->setFocus();
}

void dooble::copyDooble(const bool override)
{
  /*
  ** We'll bundle methods that are particularly slow
  ** within this little routine.
  */

  if(override || doobleCounter == 1)
    initializeHistoryMenu();
  else
    {
      /*
      ** Find a Dooble window.
      */

      dooble *d = 0;

      foreach(QWidget *widget, QApplication::topLevelWidgets())
	if(this != widget &&
	   qobject_cast<dooble *> (widget))
	  {
	    d = qobject_cast<dooble *> (widget);
	    break;
	  }

      if(d)
	{
	  for(int i = 0; i < d->ui.historyMenu->actions().size(); i++)
	    if(i == 0 || i == 1)
	      ui.historyMenu->actions().at(i)->setEnabled
		(d->ui.historyMenu->actions().at(i)->isEnabled());
	    else if(i == 2)
	      ui.historyMenu->addSeparator();
	    else if(i == 3 || i == 4)
	      continue;
	    else
	      {
		QAction *a = new QAction
		  (d->ui.historyMenu->actions().at(i)->icon(),
		   d->ui.historyMenu->actions().at(i)->text(),
		   this);

		a->setData
		  (d->ui.historyMenu->actions().at(i)->data().toUrl().
		   toString(QUrl::StripTrailingSlash));
		connect(a, SIGNAL(triggered(void)), this,
			SLOT(slotLinkActionTriggered(void)));
		ui.historyMenu->addAction(a);
	      }

	  ui.locationLineEdit->copyContentsOf(d->ui.locationLineEdit);
	}
    }
}

void dooble::slotStatusBarMessage(const QString &text)
{
  dview *p = qobject_cast<dview *> (ui.tabWidget->currentWidget());
  dwebpage *d = qobject_cast<dwebpage *> (sender());

  if(p && d == p->page())
    {
      sb.statusLabel->setText(text);
      QTimer::singleShot(1000, sb.statusLabel, SLOT(clear(void)));
    }
}

void dooble::slotObjectDestroyed(QObject *object)
{
  QWidget *w = qobject_cast<QWidget *> (object);

  if(w)
    {
      if(ui.tabWidget->indexOf(w) != -1)
	ui.tabWidget->removeTab(ui.tabWidget->indexOf(w));

      if(ui.tabWidget->count() == 1)
	{
	  ui.tabWidget->setTabsClosable(false);
	  ui.tabWidget->setBarVisible
	    (s_settings.value("settingsWindow/alwaysShowTabBar",
			      true).toBool());
	}
    }
}

bool dooble::warnAboutLeavingModifiedTab(void)
{
  /*
  ** Returns true if the warning was accepted or
  ** if the setting is disabled.
  */

  if(s_settings.value("settingsWindow/warnBeforeLeavingModifiedTab",
		      false).toBool())
    {
      QMessageBox mb(this);

      mb.setIcon(QMessageBox::Question);
      mb.setWindowTitle(tr("Dooble Web Browser: Confirmation"));
      mb.setWindowModality(Qt::ApplicationModal);
      mb.setStandardButtons(QMessageBox::Cancel |
			    QMessageBox::No | QMessageBox::Yes);
      mb.setText(tr("Are you sure that you wish to leave the modified "
		    "page?"));

      QSettings settings(s_settings.value("iconSet").toString(),
			 QSettings::IniFormat);

      for(int i = 0; i < mb.buttons().size(); i++)
	if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("okButtonIcon").toString()));
	else
	  mb.buttons().at(i)->setIcon
	    (QIcon(settings.value("cancelButtonIcon").toString()));

      mb.setWindowIcon
	(QIcon(settings.value("mainWindow/windowIcon").toString()));

      if(mb.exec() == QMessageBox::No)
	return false;
    }

  return true;
}

void dooble::slotAuthenticate(void)
{
  if(promptForPassphrase(true))
    {
      s_cookies->populate();
      copyDooble(true);
      prepareMostVisited();
      s_historyWindow->populate();
      s_downloadWindow->populate();
      s_historyModel->populate();
      s_popupsWindow->populate();
      s_adBlockWindow->populate();
      s_cookiesBlockWindow->populate();
      s_httpRedirectWindow->populate();
      s_httpReferrerWindow->populate();
      s_javaScriptExceptionsWindow->populate();
    }
}

void dooble::setUrlHandler(dooble *d)
{
  QDesktopServices::setUrlHandler("file",
				  d,
				  "slotOpenLinkInNewTab");
  QDesktopServices::setUrlHandler("ftp",
				  d,
				  "slotOpenLinkInNewTab");
  QDesktopServices::setUrlHandler("http",
				  d,
				  "slotOpenLinkInNewTab");
  QDesktopServices::setUrlHandler("https",
				  d,
				  "slotOpenLinkInNewTab");
  QDesktopServices::setUrlHandler("qrc",
				  d,
				  "slotOpenLinkInNewTab");
}

void dooble::unsetUrlHandler(void)
{
  QDesktopServices::unsetUrlHandler("file");
  QDesktopServices::unsetUrlHandler("ftp");
  QDesktopServices::unsetUrlHandler("http");
  QDesktopServices::unsetUrlHandler("https");
  QDesktopServices::unsetUrlHandler("qrc");

  foreach(QWidget *widget, QApplication::topLevelWidgets())
    if(qobject_cast<dooble *> (widget))
      {
	dooble *d = qobject_cast<dooble *> (widget);

	if(d && d != this)
	  {
	    setUrlHandler(d);
	    break;
	  }
      }
}

void dooble::slotReopenClosedTab(void)
{
  QAction *action = qobject_cast<QAction *> (sender());

  if(action)
    {
      int index = -1;

      /*
      ** First, delete the correct action from the Recently-Closed Tabs
      ** menu.
      */

      if(ui.historyMenu->actions().size() >= 4)
	for(int i = 0;
	    i < ui.historyMenu->actions()[3]->menu()->actions().size(); i++)
	  if(action == ui.historyMenu->actions()[3]->menu()->actions()[i])
	    {
	      index = i;
	      action->deleteLater();
	      break;
	    }

      /*
      ** Create a view having the old view's history.
      */

      for(int i = 0; i < m_closedTabs.count(); i++)
	if(i == index)
	  {
	    dview *p = 0;
	    dwebhistory *history = m_closedTabs.at(i);

	    if((p = newTab(history)))
	      {
		ui.tabWidget->setCurrentWidget(p);
		ui.tabWidget->update();
	      }

	    delete m_closedTabs.takeAt(i);
	    prepareTabsMenu();
	    break;
	  }

      if(m_closedTabs.isEmpty() &&
	 ui.historyMenu->actions().size() >= 4)
	{
	  ui.historyMenu->actions()[3]->setEnabled(false);
	  ui.historyMenu->actions()[3]->menu()->deleteLater();
	}
    }
}

void dooble::prepareTabsMenu(void)
{
  /*
  ** This method assumes that the views (and desktop)
  ** have valid actions.
  */

  ui.menu_Tabs->clear();

  for(int i = 0; i < ui.tabWidget->count(); i++)
    {
      dview *p = qobject_cast<dview *> (ui.tabWidget->widget(i));

      if(p)
	{
	  if(p->tabAction())
	    ui.menu_Tabs->addAction(p->tabAction());
	}
      else
	{
	  if(desktopWidget->tabAction())
	    ui.menu_Tabs->addAction(desktopWidget->tabAction());
	}
    }

  ui.menu_Tabs->setEnabled(!ui.menu_Tabs->actions().isEmpty());
}

QString dooble::pluginPath ()
{
  return QCoreApplication::applicationDirPath() + "/plugins";
}

void dooble::slotRefreshPlugins ()
{
  QList<QAction*> actions = ui.menu_Plugins->actions();
  for (int a=0; a<actions.count(); a++) 
    {
      if (actions.at(a)->objectName() != "actionRefreshPlugins"
          && ! actions.at(a)->isSeparator()) 
        {
          QAction * dead = actions.at(a);
          std::cerr << " removing action " << dead->objectName().toStdString()
                    << std::endl;
          ui.menu_Plugins->removeAction (dead);
          dead->deleteLater ();
        }
    }
  QDir plugDir (pluginPath());
  QStringList fileList (plugDir.entryList (QDir::Files  
                    | QDir::Readable 
                    | QDir::NoDotAndDotDot,
                      QDir::IgnoreCase));
  for (int p=0; p<fileList.count(); p++) 
    {
      QString fileName (fileList.at(p));
      if (!pluginLoadCount.contains(fileName))
        {
          pluginLoadCount[fileName] = 0;
        }
      if (pluginLoadCount[fileName] < 1
         && fileName != "README") 
        {
          addPluginAction (fileName);
        }
    }
}

void dooble::addPluginAction (const QString & pluginFileName)
{
  QAction * plugAction = new QAction (this);
  QStringList ops ("start");
  ops.append (pluginFileName);
  plugAction->setData (ops);
  QString title (pluginFileName);
  if (title.startsWith ("lib")) 
    {
      title.remove (0,3); 
    }
  title = title.section (".",0,0);
  title[0] = title[0].toUpper();
  plugAction->setText (tr ("Start %1").arg(title));
  ui.menu_Plugins->addAction (plugAction);
}

void dooble::slotPluginAction (QAction *plugAction)
{
  if (!plugAction) 
    {
      return;
    }
  QVariant plugData = plugAction->data();

  if (plugData.type() != QVariant::StringList) 
    {
      return;
    }
  QStringList ops = plugData.toStringList ();
  if (ops.count() != 2) 
    {
       return;
    }
  if (ops.at(0) == "start") 
    {
      startPlugin (ops.at(1));
    }
}

Extension * dooble::loadPlugin (const QString & plugName)
{
  QPluginLoader loader;
  QString filename (plugName);
  Extension * loadedPlugin (0);
  QString fullName (pluginPath() +QString ("/") + filename);
  loader.setFileName (fullName);
  bool ok = loader.load ();
  if (ok)
    {
      loadedPlugin = qobject_cast <Extension*> (loader.instance());
    }
  else
    {
      QMessageBox::information (this, "Addon Load failed",
                    loader.errorString());
    }

  return loadedPlugin;
}

void dooble::startPlugin (const QString & plugName)
{
  Extension * plugin = loadPlugin (plugName);
  if (plugin)
    {
      dplugintab * plugTab = new dplugintab (plugin, this);
      pluginLoadCount[plugName] += 1;
      pluginMap[plugin->widget()] = PluginRec (plugTab, plugName);
      if (plugTab) 
        {
          ui.tabWidget->addTab (plugTab,plugin->icon(),plugin->title());
          ui.tabWidget->setCurrentWidget (plugTab);
          SignalAgent * agent = plugin->agent();
          if (agent) 
            {
              connect (agent, SIGNAL(exiting(Extension*, int)),
                       this, SLOT (slotPluginExiting(Extension*, int)));
              connect (agent, SIGNAL(iconChange(Extension*)),
                       plugTab, SLOT (slotIconChange(Extension*)));
              connect (agent, SIGNAL(titleChange(Extension*)),
                       plugTab, SLOT (slotTitleChange(Extension*)));
              connect (plugTab, SIGNAL (iconChange (QWidget *, const QIcon &)),
                       ui.tabWidget, SLOT (slotIconChange(QWidget*, const QIcon &)));
              connect (plugTab, SIGNAL (titleChange (QWidget*, const QString &)),
                       ui.tabWidget, SLOT (slotTitleChange (QWidget*, const QString &)));
            }
        }
      slotRefreshPlugins ();
      plugin->run ();
    }
}

void dooble::slotPluginExiting (Extension *plugin, int status)
{
  Q_UNUSED (status)

  if (plugin) 
    {
      QWidget * plugWidget = plugin->widget();
      dplugintab * plugTab = pluginMap[plugWidget].tab;

      pluginLoadCount[pluginMap[plugWidget].fileName] -= 1;
      if (plugTab)
        {
           pluginMap.remove (plugWidget);
           int index = ui.tabWidget->indexOf (plugTab);

           if (index >= 0)
             {
               ui.tabWidget->removeTab (index);
             }
        }
      slotRefreshPlugins ();
    }
}
