/****************************************************************************
** Dooble - The Secure Internet Web Browser
**
** Copyright (c) 2008, 2009, 2010, 2011 Ronald Landheer-Cieslak,
** Alexis Megas, Gunther van Dooble, and the Dooble Team.
** All rights reserved.
**
** License: GPL2 only:
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; version 2 of the License only.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
** or see here: http://www.gnu.org/licenses/gpl.html
**
** For the WebKit library, please see: http://webkit.org.
**
** THE CODE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
** ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS APPLICATION, EVEN IF ADVISED
** OF THE POSSIBILITY OF SUCH DAMAGE.
**
** Please report all praise, requests, bugs, and problems to the project
** team and administrators: http://sf.net/projects/dooble.
**
** You can find us listed at our project page. New team members are welcome.
** The name of the authors should not be used to endorse or promote products
** derived from Dooble without specific prior written permission.
** If you use this code for other projects, please let us know.
**
** Web sites:
**   http://sf.net/projects/dooble
**   http://dooble.sf.net
****************************************************************************/

#include "dooble.h"
#include "dwebpage.h"
#include "dexceptionswindow.h"
#include "dnetworkaccessmanager.h"

#include <QUrl>
#include <QFile>
#include <QtDebug>
#include <QSettings>
#include <QWebFrame>
#include <QMessageBox>
#include <QInputDialog>

dwebpage::dwebpage(QObject *parent):QWebPage(parent),
				    never_show_error_page_(false),
				    never_emit_error_signal_(false)
{
  if(networkAccessManager())
    networkAccessManager()->deleteLater();

  dnetworkaccessmanager *networkAccessManager =
    new dnetworkaccessmanager(this);

  connect(this,
	  SIGNAL(popupRequested(const QString &,
				const QUrl &,
				const QDateTime &)),
	  dooble::s_popupsWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));
  connect(networkAccessManager,
	  SIGNAL(blockThirdPartyHost(const QString &,
				     const QUrl &,
				     const QDateTime &)),
	  dooble::s_adBlockWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));
  connect(networkAccessManager,
	  SIGNAL(suppressHttpReferrer(const QString &,
				      const QUrl &,
				      const QDateTime &)),
	  dooble::s_httpReferrerWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));
  connect(networkAccessManager,
	  SIGNAL(urlRedirectionRequest(const QString &,
				       const QUrl &,
				       const QDateTime &)),
	  dooble::s_httpRedirectWindow,
	  SLOT(slotAdd(const QString &,
		       const QUrl &,
		       const QDateTime &)));
  setNetworkAccessManager(networkAccessManager);
  setForwardUnsupportedContent(true);
}

void dwebpage::javaScriptAlert(QWebFrame *frame, const QString &msg)
{
  /*
  ** The default implementation of javaScriptAlert() causes
  ** segmentation faults when viewing certain Web pages.
  */

  QMessageBox *mb = new QMessageBox(frame->page()->view());

  mb->setIcon(QMessageBox::Information);
  mb->setWindowTitle(tr("JavaScript Alert"));
  mb->setStandardButtons(QMessageBox::Ok);
  mb->setText(msg);

  QSettings settings(dooble::s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);

  for(int i = 0; i < mb->buttons().size(); i++)
    if(mb->buttonRole(mb->buttons().at(i)) == QMessageBox::AcceptRole)
      mb->buttons().at(i)->setIcon
	(QIcon(settings.value("okButtonIcon").toString()));

  mb->setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));
  connect(mb,
	  SIGNAL(finished(int)),
	  mb,
	  SLOT(deleteLater(void)));
  mb->show();
}

bool dwebpage::javaScriptConfirm(QWebFrame *frame, const QString &msg)
{
  QMessageBox mb(frame->page()->view());

  mb.setIcon(QMessageBox::Question);
  mb.setWindowTitle(tr("JavaScript Confirm"));
  mb.setWindowModality(Qt::ApplicationModal);
  mb.setStandardButtons(QMessageBox::No | QMessageBox::Yes);
  mb.setText(msg);

  QSettings settings(dooble::s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);

  for(int i = 0; i < mb.buttons().size(); i++)
    if(mb.buttonRole(mb.buttons().at(i)) == QMessageBox::YesRole)
      mb.buttons().at(i)->setIcon
	(QIcon(settings.value("okButtonIcon").toString()));
    else
      mb.buttons().at(i)->setIcon
	(QIcon(settings.value("cancelButtonIcon").toString()));

  mb.setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));

  if(mb.exec() == QMessageBox::Yes)
    return true;
  else
    return false;
}

bool dwebpage::javaScriptPrompt(QWebFrame *frame, const QString &msg,
				const QString &defaultValue, QString *result)
{
  QSettings settings(dooble::s_settings.value("iconSet").toString(),
		     QSettings::IniFormat);
  QInputDialog id(frame->page()->view());

  id.setInputMode(QInputDialog::TextInput);
  id.setLabelText(msg);
  id.setWindowTitle(tr("JavaScript Prompt"));
  id.setWindowModality(Qt::ApplicationModal);
  id.setTextValue(defaultValue);
  id.setOkButtonText(tr("OK"));
  id.setCancelButtonText(tr("Cancel"));
  id.setWindowIcon
    (QIcon(settings.value("mainWindow/windowIcon").toString()));

  foreach(QPushButton *button, id.findChildren<QPushButton *> ())
    if(button->text() == tr("OK"))
      button->setIcon
	(QIcon(settings.value("okButtonIcon").toString()));
    else
      button->setIcon
	(QIcon(settings.value("cancelButtonIcon").toString()));

  if(id.exec() == QDialog::Accepted)
    {
      *result = id.textValue();
      return true;
    }
  else
    {
      *result = "";
      return false;
    }
}

QWebPage *dwebpage::createWindow(WebWindowType type)
{
  Q_UNUSED(type);

  if(dooble::s_settings.value("settingsWindow/blockPopups", false).toBool())
    {
      /*
      ** This is a difficult situation. The createWindow() method
      ** doesn't provide us with a URL.
      */

      bool allowPopup = false;
      dwebview *v = qobject_cast<dwebview *> (parent());

      if(v)
	allowPopup = v->checkAndClearPopup();

      if(!allowPopup)
	{
	  if(!dooble::s_popupsWindow->allowed(m_requestedUrl.host()))
	    {
	      emit popupRequested(m_requestedUrl.host(),
				  m_requestedUrl,
				  QDateTime::currentDateTime());
	      return 0;
	    }
	  else
	    allowPopup = true;
	}

      if(!allowPopup)
	return 0;
    }

  if(dooble::s_settings.value("settingsWindow/javascriptEnabled",
			      true).toBool())
    {
      if(dooble::s_settings.value("settingsWindow/javascriptAllowNewWindows",
				  false).toBool())
	{
	  dview *v = 0;
	  dooble *dbl = new dooble(true, findDooble());

	  v = dbl->newTab(m_requestedUrl);
	  return v->page();
	}
      else
	return 0;
    }

  /*
  ** A new window will not be created.
  */

  return 0;
}

dooble *dwebpage::findDooble(void)
{
  dooble *dbl(0);
  QObject *prnt(this);

  do
    {
      prnt = prnt->parent();

#ifdef DOOBLE_X11_FLASH_BYPASS
      if(qobject_cast<dwebview *> (prnt))
	prnt = qobject_cast<dwebview *> (prnt)->scene();
#endif

      dbl = qobject_cast<dooble *> (prnt);
    }
  while (prnt != 0 && dbl == 0);

  return dbl;
}

QString dwebpage::userAgentForUrl(const QUrl &url) const
{
  QString agent(QWebPage::userAgentForUrl(url));

  if(url.host().contains("ebay.com"))
    {
      agent.remove("(KHTML, like Gecko) ");
      agent.remove(" Safari/533.3");
      agent.remove(" Apple/533.3");
    }
  else if(url.host().contains("gmail.com") ||
	  url.host().contains("google.com"))
    {
#ifdef Q_OS_MAC
      QSysInfo::MacVersion version = QSysInfo::MacintoshVersion;

      if(version == QSysInfo::MV_10_3)
	agent.replace("Mac OS X", "Mac OS X 10_3");
      else if(version == QSysInfo::MV_10_4)
	agent.replace("Mac OS X", "Mac OS X 10_4");
      else if(version == QSysInfo::MV_10_5)
	agent.replace("Mac OS X", "Mac OS X 10_5");
      else if(version == QSysInfo::MV_10_6)
	agent.replace("Mac OS X", "Mac OS X 10_6");
      else
	agent.replace("Mac OS X", "Mac OS X 10");
#endif
    }

  if(!url.host().contains("ebay.com"))
    agent.replace("(KHTML, like Gecko)", "(KHTML, like Gecko) Version/5.0");

  return agent;
}

dwebpage::~dwebpage()
{
}

bool dwebpage::acceptNavigationRequest(QWebFrame *frame,
				       const QNetworkRequest &request,
				       NavigationType type)
{
  if(!request.url().isValid())
    return false;

  m_requestedUrl = request.url();

  if(type == QWebPage::NavigationTypeLinkClicked)
    {
      if(!frame)
	{
	  if(dooble::s_settings.value("settingsWindow/openInNewTab",
				      true).toBool())
	    {
	      dview *new_page = findDooble()->newTab(request.url());

	      if(dooble::s_settings.value("settingsWindow/proceedToNewTab",
					  true).toBool())
		findDooble()->setCurrentPage(new_page);

	      return false;
	    }
	  else
	    {
	      dooble *dbl = new dooble(false, 0);

	      dbl->newTab(request.url());
	      return false;
	    }
	}
      else
	{
	  if(QApplication::keyboardModifiers() == Qt::ControlModifier)
	    {
	      dview *new_page = findDooble()->newTab(request.url());

	      if(dooble::s_settings.value("settingsWindow/proceedToNewTab",
					  true).toBool())
		findDooble()->setCurrentPage(new_page);

	      return false;
	    }
	}
    }

  if(frame)
    /*
    ** Prevent loading both the current frame and a new tab (window)
    ** with the same URL.
    */

    if(request.url().scheme().toLower() == "ftp")
      emit linkClicked(request.url());

  return QWebPage::acceptNavigationRequest(frame, request, type);
}
